#include <baseInternal.h>
#include <lst.h>
#include <field.h>
#include <pregex.h>

#include <memory.h>
#include <assert.h>

static boolean check_nil(const Field*F, unsigned long v){
	USED(F);
	USED(v);
	assert(0);
}

boolean field_check_okay(const Field* F, unsigned long v){
	USED(F);
	USED(v);
	return true;
}

boolean field_check_exact(const Field* F, unsigned long v){
	int i;

	v &= F->mask;

	switch(F->exact_bit){
	default:
		assert(0);
		break;
	case FExactZero:
		return true;
	case FExactEqual:
		for(i=0; i<F->n_exact; i++){	
			if (v == F->exact[i]){
				return true;
			}
		}
		break;
	case FExactAnd:
		for(i=0; i<F->n_exact; i++){
			if (F->exact[i] == (v & F->exact[i])){
				return true;
			}
		}
		break;
	case FExactBitsSet:
		for(i=0; i<F->n_exact; i++){
			if (v & F->exact[i]){
				return true;
			}
		}
		break;
	case FExactXor:
		for(i=0; i<F->n_exact; i++){
			if (v ^ F->exact[i]){
				return true;
			}
		}
		break;
	}

	return false;
}

boolean field_check_range(const Field* F, unsigned long v){
	if (F->has_min && v<F->min){
		return false;
	}
	if (F->has_max && v>F->max){
		return false;
	}
	return true;
}

boolean field_check_all(const Field*F, unsigned long v){
	return field_check_range(F,v) && field_check_exact(F,v);
}

boolean field_check_re(const Field*F, ul lR){
	return regexMatchRange(F->re_stuff,(Range*)lR);
}

Field* fieldCreate(void){
	Field* F;

	F=xalloc(sizeof(Field));
	F->base = 0;
	F->nth=~0UL;
	F->check = field_check_all;
	F->exact_bit = FExactZero;
	F->mask = ~0UL;
	F->n_exact=0;
	F->exact = 0;
	F->has_min=false;
	F->min=0;
	F->has_max=false;
	F->max=~0UL;

	F->re = 0;
	F->re_stuff = 0;
	return F;
}

lstHead* fieldCompile(Field*F){
	Regex* R;

	if (0 == F->re){
		return 0;
	}

	if (F->re_stuff){
		regexDestroy(F->re_stuff);
	}
	R = F->re_stuff = regexCreate();
	regexCompile(F->re,R);
	return lstCount(R->errors) ? R->errors : 0;
}

Field* fieldCopy(const Field*F){
	Field* G;

	G=fieldCreate();
	G->base = F->base;
	G->nth = F->nth;
	G->check = F->check;
	G->exact_bit = F->exact_bit;
	G->n_exact=F->n_exact;
	G->mask=F->mask;
	if (G->n_exact){
		G->exact = xalloc(G->n_exact*sizeof(unsigned long));
		memcpy(G->exact, F->exact, G->n_exact);
	}
	G->has_min = F->has_min;
	G->min = F->min;
	G->has_max = F->has_max;
	G->max = F->max;

	G->re = xcopy(F->re);
	G->re_stuff = 0;

	return G;
}

void fieldDestroy(Field*F){
	cxfree(F->exact);
	cxfree(F->re);
	if (F->re_stuff){
		regexDestroy(F->re_stuff);
	}
	memset(F,0x1,sizeof(Field));
}

lstCmp fieldCmp(const Field* a, const Field* b){
	int t;

	t = b->nth - a->nth;
	if (t>0){
		return lstLess;
	} else if (t<0){
		return lstGreater;
	}
	return lstEqual;
}

const lstFns lstFnsField = {
	(lstFnCopy)fieldCopy,
	(lstFnDelete)fieldDestroy,
	(lstFnCompare)fieldCmp,
	"%{field}"
};
