/*	provide a buffer between posix regular expressions and the
	rest of the world.
*/
#include <base.h>
#include <lst.h>
#include <xalloc.h>
#include <sys/types.h>
#include <slurp.h>
#include <field.h>

#include <pregex.h>
#include <memory.h>

/*	alternately use RE_SYNTAX_EGREP → close but not so picky
	N.B. we are not defining RE_INTERFALS or RE_NO_BK_BRACES
*/
static const int base_flags = RE_BACKSLASH_ESCAPE_IN_LISTS | 
	RE_CHAR_CLASSES | \
	RE_CONTEXT_INDEP_ANCHORS | \
	RE_CONTEXT_INDEP_OPS | \
	RE_CONTEXT_INVALID_OPS | \
	RE_NO_BK_PARENS | \
	RE_NO_BK_VBAR | \
	RE_NO_EMPTY_RANGES | \
	RE_UNMATCHED_RIGHT_PAREN_ORD;

void regexInit(void){
	re_set_syntax(base_flags);
}

Regex* regexCreate(void){
	Regex* R;

	R=xalloc(sizeof(Regex));
	R->iflag = false;
	R->errors = lstCreate(&lstFnsConstString);
	memset(&R->pbuf, 0, sizeof R->pbuf);
	memset(&R->regs, 0, sizeof R->regs);
	R->nregs = 30;
	R->starts = xalloc(31 * sizeof(R->starts));
	R->ends = xalloc(31 * sizeof(R->ends));

	return R;
}

void regexDestroy(Regex*R){
#if 0
	if (R->pbuf.allocated){
		xfree(R->pbuf.buffer);
	}
	cxfree(R->pbuf.fastmap);
#endif
	regfree(&R->pbuf);
	lstDestroy(R->errors);
	xfree(R->starts);
	xfree(R->ends);
	xfree(R);
}

boolean regexCompile(const char* ex, Regex* R){
	const char* estring;

	estring = re_compile_pattern(ex,strlen(ex), &R->pbuf);
	if (estring){
		lstAddEntry(R->errors, (char*)estring);
		return false;
	}
	R->pbuf.fastmap = xalloc(256);	/* read the source, luke */
	re_compile_fastmap(&R->pbuf);
	re_set_registers(&R->pbuf,&R->regs, R->nregs, R->starts, R->ends);

	return true;
}

boolean regexMatchRange(Regex* RE, Range*R){
	ul length;

	length = R->end - R->start;
	if (0 <= re_search(&RE->pbuf, R->start, length, 0, length, &RE->regs)){
		return true;
	}
	return false;
}

boolean regexMatch(Regex* RE, const char* pattern){
	ul length;

	length = pattern ? strlen(pattern) : 0;
	if (0 <= re_search(&RE->pbuf, pattern, length, 0, length, &RE->regs)){
		return true;
	}
	return false;
}
