/*
**
** RCS Info:	$Date: 2005/08/08 01:36:12 $
**		$Id: qstr.c,v 1.1 2005/08/08 01:36:12 quanstro Exp quanstro $
**		$Source: /home/quanstro/src/elocate/base/RCS/qstr.c,v $
**
** Description:
**
**	quick strings. inspired by a linux kernel ds of the same name.
**
** Limitations:
**
**	1. ymmv with qstrCmp().
**	2. qstr_conv() should move to stdprint library?
**
** History:
**
**	07-aug-2005 quanstro created.
*/
#include <baseInternal.h>
#include <xalloc.h>
#include <lst.h>
#include <print.h>
#include <qstr.h>
#include <hash.h>

#include <memory.h>

qstr* qstrCreate(void){
	qstr* q;

	q=xalloc(sizeof(qstr));
	memset(q,0,sizeof(qstr));
	return q;
}

void qstrInit(qstr* q, const char* name, int len){
	if (-1 == len){
		len = strlen(name);
	}
	q->hash = full_name_hash((uchar*)name, len);
	q->name = (char*)name;
	q->len = len;
}

qstr* qstrInst(const char* name, int len){
	qstr* q;

	q = qstrCreate();
	qstrInit(q, name, len);
	return q;
}

qstr* qstrCopy(const qstr* a){
	return qstrInst(xcopy(a->name), a->len);
}

lstCmp qstrCmp(const qstr* a, const qstr* b){ 
	if (a->hash > b->hash){
		return lstGreater;
	}
	return (b->hash - a->hash || b->len - a->len || memcmp(a->name, b->name, a->len))&(-1|1);
}

boolean qstreq(const qstr* a, const qstr* b){
	if (a->hash == b->hash && a->len == b->len && !memcmp(a->name, b->name, a->len)){
		return true;
	}
	return false;
}

void qstrDestroy(qstr* q){
	cxfree(q->name);
	xfree(q);
}

const lstFns lstFnsQStr = {
	(lstFnCopy)qstrCopy,
	(lstFnDelete)qstrDestroy,
	(lstFnCompare)qstrCmp,
	"%{qstr}"
};
