#define _GNU_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <assert.h>
#include <stdlib.h>
#include <errno.h>
#include <memory.h>

#include <baseInternal.h>

/* map copy-on-write buffer */
#define BYTES_LEFT (buf_alloc-(bufp-buf))
static boolean slurpByHand(int fd, bigmem* M){
	int nbytes;
	size_t size;
	char* buf;
	size_t buf_alloc;
	char* bufp;

	for (size=0, buf=0, buf_alloc=16*1024;;){
		buf_alloc *= 2;
		buf=xrealloc(buf, buf_alloc);
		
		for(bufp = buf+size;BYTES_LEFT>8 * 1024;){
		rstart:	nbytes = read(fd, bufp, BYTES_LEFT);
			switch(nbytes){
			case 0:
				goto finished;
			case -1:
				switch(errno){
				case EINTR:
					goto rstart;
				default:
					goto lossage;
				}
			default:
				break;
			}
			size+=nbytes;
			bufp+=nbytes;
		}
	}
lossage:
	xfree(buf);
	return false;

finished:
	M->base = buf;
	M->size = M->base_size = size;
	M->malloced = true;
	return true;
}

#define PAGE_MASK ((page_size-1) ^ ~0UL)
boolean slurpSliceFd(bigmem* M, int fd, size_t offset, size_t length){
	int rv;
	struct stat st;
	static int page_size = 0;

	rv = fstat(fd, &st);
	if (-1 == rv){
		return false;
	}

	if ((size_t)-1 == length){
		M->size = st.st_size;
		M->base_size = M->size;
	} else {
		/* magic calculations -- must pageboundarify*/
		if (0 == page_size){
			page_size = getpagesize();	/* no error return */
		}
		M->size = length;
		M->base_size = length + offset - (offset & PAGE_MASK);
		M->base_size += page_size-1;
		M->base_size &= ~(page_size-1);
		M->base_offset = offset;
	}

	M->base = mmap(0, M->base_size, PROT_READ, MAP_PRIVATE, fd, M->base_offset & PAGE_MASK);
	if ((char*)-1UL == M->base){
		if (!slurpByHand(fd,M)){
			return false;
		}
	}

	M->m = M->base + M->base_offset;
	M->fd = fd;

	return true;
}

boolean slurpSlice(bigmem* M, const char* file, size_t offset, size_t len){
	int fd;

	memset(M, 0, sizeof(bigmem));
	M->fd = -1;

	fd = open(file, O_RDONLY);
	if (-1 == fd){
		return false;
	}
	M->fd = fd;
	return slurpSliceFd(M, fd, offset, len);
}

boolean slurpFd(bigmem* M, int fd){
	memset(M, 0, sizeof(bigmem));
	M->fd = fd;
	return slurpSliceFd(M, fd,-1,-1);
}

boolean slurpExact(bigmem* M, const char* file){
	return slurpSlice(M, file,-1, -1);
}

void unslurp(bigmem*M){
	int rv;

	rv = munmap(M->base, M->base_size);
	if (-1 == rv){
		assert("munmap()");
	}
	close(M->fd);
	if (M->malloced){
		xfree(M->base);
	}
	memset(M, 0, sizeof(bigmem));
	M->fd = -1;
}
