#include <baseInternal.h>

#include <ctype.h>
#include <limits.h>

/* ain't character sets grand! dvalue is offset by one to take advantage of the compiler
   setting unset elements to 0.
*/
static const unsigned long ctab[256] = {
	['0'] = 0+1, ['1'] = 2,  ['2'] = 3,  ['3'] = 4,  ['4'] = 5,
	['5'] = 6,   ['6'] = 7,  ['7'] = 8,  ['8'] = 9,  ['9'] = 10,

	['a'] = 11,  ['b'] = 12, ['c'] = 13, ['d'] = 14, ['e'] = 15, ['f'] = 16, ['g'] = 17, 
	['h'] = 18,  ['i'] = 19, ['j'] = 20, ['k'] = 21, ['l'] = 22, ['m'] = 23, ['n'] = 24, 
	['o'] = 25,  ['p'] = 26, ['q'] = 27, ['r'] = 28, ['s'] = 29, ['t'] = 30, ['u'] = 31, 
	['v'] = 32,  ['w'] = 33, ['x'] = 34, ['y'] = 35, ['z'] = 36,

	['A'] = 11,  ['B'] = 12, ['C'] = 13, ['D'] = 14, ['E'] = 15, ['F'] = 16, ['G'] = 17, 
	['H'] = 18,  ['I'] = 19, ['J'] = 20, ['K'] = 21, ['L'] = 22, ['M'] = 23, ['N'] = 24, 
	['O'] = 25,  ['P'] = 26, ['Q'] = 27, ['R'] = 28, ['S'] = 29, ['T'] = 30, ['U'] = 31, 
	['V'] = 32,  ['W'] = 33, ['X'] = 34, ['Y'] = 35, ['Z'] = 36
};

boolean strtouln_r(ul* v, const char* p0, char const ** rend, ul len, ul base){
	unsigned long result;
	const char* end;
	unsigned long dvalue;
	unsigned long max_shift;
	unsigned long max_next;

	end=p0+len;

	while (p0<end && isspace(*p0)){
		p0++;
	}

	/* check for 0octal or 0xhex */
	if (0 == base){
		if (p0<end && '0' == *p0){
			p0++;
	
			if ((end-p0)>1 && ('x' == *p0 || 'X' == *p0)){
				base=16;
				p0++;
			} else {
				base=8;
			}
		} else {
			base = 10;
		}
	}

	/* pick off leading zeros */
	while(p0<end && '0' == *p0){
		p0++;
	}

	result = 0;
	max_shift = ULONG_MAX / base;
	max_next  = ULONG_MAX % base;

	for(; p0<end; p0++){
		dvalue = ctab[*p0];
		if (0 == dvalue || dvalue-->base){
			break;
		}
		/* detect overflow? */
		if (result>max_shift || (result==max_shift && dvalue>max_next)){
			if (rend) {
				*rend = p0;
			}
			*v = ~0;
			return false;
		}
		result = result*base+dvalue;
	}

	*v = result;
	if (rend){
		*rend = p0;
		return true;
	}
	return end==p0;
}

boolean strtoln_r(long* v, const char* p0, char const **rend, unsigned long ilen, long base){
	long sign;
	long result;
	const char* end;
	long dvalue;
	long max_shift;
	long max_next;

	end = p0 + ilen;

	while (p0<end && isspace(*p0)){
		p0++;
	}

	/* check for 0octal or 0xhex */
	if (0 == base){
		if (p0<end && '0' == *p0){
			p0++;
	
			if ((end-p0)>1 && ('x' == *p0 || 'X' == *p0)){
				base=16;
				p0++;
			} else {
				base=8;
			}
		} else {
			base = 10;
		}
	}


	/* pick off sign so we can verify length of number is good */
	sign=1;
	while(p0<end && (*p0 == '-' || *p0 == '+')){
		if (*p0 == '-'){
			sign = -sign;
		}
		p0++;
	}

	/* pick off leading zeros */
	while(p0<end && '0' == *p0){
		p0++;
	}

	result=0;
	max_shift = LONG_MAX / base;
	max_next  = LONG_MAX % base;

	for(; p0<end; p0++){
		dvalue = ctab[*p0];
		if (0 == dvalue || dvalue-->base){
			break;
		}
		/* detect overflow? */
		if (result>max_shift || (result==max_shift && dvalue>max_next)){
			if (rend) {
				*rend = p0;
			}
			*v = ~0;
			return false;
		}
		result = result*base+dvalue;
	}

	*v = result;
	if (rend){
		*rend = p0;
		return true;
	}
	return end==p0;
}

vlong llmax = LLONG_MAX;

boolean strtolln(vlong* v, const char* p0, unsigned long len, vlong base){
	vlong sign;
	vlong result;
	const char* digit;
	vlong dvalue;
	vlong max_shift;
	long max_next;
	vlong mfactor;

	while (len>0 && isspace(*p0)){
		len--;
		p0++;
	}

	/* check for 0octal or 0xhex */
	if (0 == base){
		if (len>0 && '0' == *p0){
			int c;

			len--;
			p0++;
			c = *p0;

			if (len>1 && ('x' == c || 'X' == c)){
				base=16;
				len--;
				p0++;
			} else if (len>1 && ('p' == c || 'P' == c)){
				base=2;
				len--;
				p0++;
			} else {
				base=8;
			}
		} else {
			base = 10;
		}
	}

	/* pick off sign so we can verify length of number is good */
	sign=1;
	while(len>0 && (*p0 == '-' || *p0 == '+')){
		if (*p0 == '-'){
			sign = -sign;
		}
		len--;
		p0++;
	}

	/* pick off leading zeros */
	while(len>0 && '0' == *p0){
		len--;
		p0++;
	}

	/* nothing left to convert, be nice and return 0 */
	if (0>=len){
		*v = 0;
		return true;
	}

	digit=p0+len-1;
	result=0;
	max_shift = LLONG_MAX / base;
	max_next  = LLONG_MAX % base;
	mfactor = 1;

	for(;digit>=p0;digit--, mfactor*=base){
		dvalue = ctab[(int)*digit];
		if (0 == dvalue || dvalue-->base){
			return false;
		}
		/* detect overflow? */
		if (result>max_shift || (result==max_shift && dvalue>max_next)){
			return false;
		}
		dvalue *= mfactor;
		result += dvalue;
	}
	*v = sign * result;
	return true;
}
