/*
**
**  RCS Info:		$Date: 2005/05/29 17:02:16 $ $Revision: 1.1 $
**
**			$Id: hash.h,v 1.1 2005/05/29 17:02:16 quanstro Exp $
**
**  Filename:		$Source: /home/quanstro/src/elocate/include/RCS/hash.h,v $
**
**  Originator:		erik quanstrom
**			202 328 3688
**			quanstro@quanstro.net
**
**  Description:
**
**	hash from linux/dentry.h from the linux 2.2 kernel.
**	this hasn't been used for bucketing; only for optimizing
**	strcmp() into qstr1.hash == qstr2.hash && qstr1.len == qstr2.len && strcmp();
**	obviously assuming that a count(qstr1 != qstr2) » count(qstr1 == qstr2)
**
**  Limitations:
**
**	1.  the linux 2.6 kernel is now using an R5 hash from reiserfs.
**	     i found a modified version of this to be better. ymmv ☺
**
**  Modification History:
**
** 	02-mar-2000 quanstro Created.
**	07-mar-2000 sward    Added ifndef.
**	30-jul-2005 quanstro new hash function.
**
******************************************************************************/

#ifdef AIX
#define __inline__
#endif

/* Name hashing routines. Initial hash value */
#define init_name_hash()		0

/* 2005-07-30: bakeoff results: for integers 1..10000

	1. best: modified reiser
		prevhash =  (prevhash + (c << 4) + (c >> 4)) * 11;
		return prevhash^c;

	2. second: plan9 c compiler algorithm
		return 3*prevhash+c;

	3. third: original kernel hash function
		prevhash = (prevhash << 4) | (prevhash >> (8*sizeof(unsigned long)-4));
		return prevhash ^ c;

	4. dead last: reiser R5 function. max depth 146!
		return (prevhash + (c << 4) + (c >> 4)) * 11;
*/
static __inline__ unsigned long partial_name_hash(unsigned long c, unsigned long prevhash)
{
	prevhash =  (prevhash + (c << 4) + (c >> 4)) * 11;
	return prevhash ^ c;
}

/* Finally: cut down the number of bits to a int value (and try to avoid losing bits) */
static __inline__ unsigned long end_name_hash(unsigned long hash)
{
	if ((long)hash<0)
		hash = ~hash;
	return (unsigned int)hash;
}

/* Compute the hash for a name string. */
static __inline__ unsigned int full_name_hash(const unsigned char * name, unsigned int len)
{
	unsigned long hash = init_name_hash();
	while (len--)
		hash = partial_name_hash(*name++, hash);
	return end_name_hash(hash);
}

