/*  
** RCS Info:	$Date: 2005/07/23 19:08:25 $
**		$Id: lst.h,v 1.4 2005/07/23 19:08:25 quanstro Exp quanstro $
**		$Source: /home/quanstro/src/elocate/include/RCS/lst.h,v $
*/
typedef struct lstEntry lstEntry;
typedef struct lstHead lstHead;

typedef enum {
	lstLess = -1,
	lstEqual = 0,
	lstGreater = 1
} lstCmp;

typedef void* (*lstFnCopy)(const void*);
typedef void  (*lstFnDelete)(void*);
typedef lstCmp(*lstFnCompare)(const void*, const void*);

typedef struct {
	void* (*copy)(const void*);
	void  (*delete)(void*);
	lstCmp (*compare)(const void*, const void*);
	const char* print;
} lstFns;

extern const lstFns lstFnsString;
extern const lstFns lstFnsConstString;
extern const lstFns lstFnsDummy;
extern const lstFns lstFnsDummyAssert;

/* fer strings */
lstCmp xcmp(const char*, const char*);

struct lstEntry {
	lstEntry* next;
	lstEntry* prev;
	void* data;
};

struct lstHead {
	long n;
	const lstFns* f;
	lstEntry* head;
	lstEntry* tail;
};

lstHead* lstCreate(const lstFns*);
void lstDestroy(lstHead*);
lstHead* lstCopy(const lstHead*);

static inline long lstCount(const lstHead* l){
	return l->n;
}

static inline boolean lstEmpty(const lstHead* l){
	if (0 == lstCount(l)){
		return true;
	}
	return false;
}

static inline boolean lstZero(const lstHead* l){
	if (0 == l || lstEmpty(l)){
		return true;
	}
	return false;
}

static inline lstHead* lstCCopy(const lstHead* l){
	if (l){
		return lstCopy(l);
	}
	return 0;
}

void lstAddEntry(lstHead *, void*);
void lstPushEntry(lstHead *, void*);
void lstInsertEntry(lstHead*, lstEntry*, void*);

void lstPushList(lstHead*, lstHead*);
void lstAddList(lstHead*, lstHead*);

void lstDeleteEntry(lstHead*, lstEntry**);
void lstDeleteLast(lstHead*);
void lstDeleteFirst(lstHead*);

void* lstRemoveEntry(lstHead*, lstEntry**);
void* lstRemoveLast(lstHead*);
void* lstRemoveFirst(lstHead*);

lstEntry* lstGetFirst(const lstHead *);
lstEntry* lstGetLast(const lstHead *);
lstEntry* lstGetNext(const lstEntry *);
lstEntry* lstGetPrev(const lstEntry *);

void* lstGetFirstData(const lstHead*, lstEntry**);
void* lstGetNextData(const lstHead*, lstEntry**);
void* lstGetLastData(const lstHead*, lstEntry**);
void* lstGetPrevData(const lstHead*, lstEntry**);
void* lstGetCurrData(const lstHead*, lstEntry**);

void** lst2tab(const lstHead*);
void lstSort(lstHead*);
void** lstSortTab(lstHead* l);		// list is destroyed

lstHead* lstStrstr2List(const char*, unsigned long, const char* );		/* separator is 1 string */
lstHead* lstStr2List(const char*, const char* );		/* seperators are list of char; not utf-aware */
lstHead* lstStrn2List(const char*, long, const char* );

lstEntry* lstScanForFnE(const lstHead*, const void*, lstFnCompare, lstEntry*);
lstEntry* lstScanForFn(const lstHead*, const void*, lstFnCompare);
lstEntry* lstScanForE(const lstHead*, const void*, lstEntry*);
lstEntry* lstScanFor(const lstHead*, const void*);
boolean lstInList(const lstHead*, const void*);

