/*
**
**  Copyright:		Copyright © public domain
**
**  RCS Info:		$Date: 2005/07/23 19:08:25 $ $Revision: 1.4 $
**
**			$Id: print.h,v 1.4 2005/07/23 19:08:25 quanstro Exp quanstro $
**
**  Filename:		$Source: /home/quanstro/src/elocate/include/RCS/print.h,v $
**
**  Originator:		Erik Quanstrom
**			202 328 3688
**			quanstro@quanstro.net
**
**  Organization:   	Erik Quanstrom
**			1455 Chapin ST NW
**			Washington DC 20009
**			Ph: 202 328 3688	FAX: zippidy
**
**  Description:
**
**	This is the libprint public domain library, a replacement for the
**	assorted standard I/O printing functions
**
**  Limitations:
**
**	1.  None
**
**  Modification History:
**
** 	14-Nov-1995 quanstro Created.
**	23-feb-1998 quanstro updated to v1.1
**	23-jul-2005 quanstro buku changes.
**
******************************************************************************/
/* from rc 1.4 - code by Byron and Paul */
#ifndef __nicer_print_h
#define __nicer_print_h

#include <stdarg.h>

#ifndef SIZE_T
typedef unsigned long SIZE_T;
#endif

const char* printVersion(void);

typedef enum {
	/* return values of conversion functions */
	FMT_flag = 0,			/* uhl#+-* ./[0-9] */
	FMT_verb = 1,			/* sc dioxXb eEfFgG rn CS */
	FMT_error = 2,			/* error during formatting */
} fmt_action;

enum {
	/* sizes of various things */
	FMT_MAXCONV = 256,	  	/* number of conversion characters */
	FMT_MAXLONGCONV = 256,		/* size of long conversion table */
	FMT_PRINT_ALLOCSIZE = 64,	/* grow by this many bytes when necessary */
	FMT_SPRINT_BUFSIZ = 1024	/* bytes */
};

typedef char* (*Alloc)(char *, SIZE_T);

typedef struct Format Format;
typedef fmt_action (Fconv)(Format *, int);
typedef fmt_action (Flongconv)(Format *, const char*);
typedef Fconv* Fmtconv;
typedef Flongconv* Fmtlongconv;

/* Format->flags values */
typedef enum {
	FMT_long	= (1<<0),	/* %l    */
	FMT_vlong	= (1<<1),	/* %ll   */
	FMT_short	= (1<<2),	/* %h    */
	FMT_byte	= (1<<3),	/* %hh   */
	FMT_unsigned	= (1<<4),	/* %u    */
	FMT_zeropad	= (1<<5),	/* %0    */
	FMT_leftside	= (1<<6),	/* %-    */
	FMT_altform	= (1<<7),	/* %#    */
	FMT_space	= (1<<8),	/* % 	 */
	FMT_sign	= (1<<9),	/* %+	 */
	FMT_comma	= (1<<10),	/* %,	 */
	FMT_f1set	= (1<<11),	/* %<n>  aka "width"*/
	FMT_f2set	= (1<<12)	/* %.<n> aka "precision" */
} fmt_flags;

struct Format {
	/* for the formatting routines */
	va_list args;
	fmt_flags flags;
	SIZE_T f1;
	SIZE_T f2;
	Fmtconv* fmttab;
	/* Fmtlongconv* longfmttab; */

	/* for the buffer maintainence routines */
	char *buf;
	char *bufbegin;
	char *bufend;
	int flushed;
	int (*grow)(Format *, SIZE_T);
	int error;
	int reqlen;
	union {
		int fd; 
		Alloc p; 
	} u;

	/* for the sake of reentrancy */
	void* client_data;
	int indent;
};

int fmtputc(Format *f, /*const*/ char c);
int fmtputr(Format *f, /*const Rune*/ unsigned short);		/* fixme? don't want to depend on rune.h */
int fmtappend(Format *format, const char *s, SIZE_T len);
int fmtcat(Format *format, const char *s);
Fmtconv fmtinstall(int c, Fmtconv f);
Fmtlongconv long_fmtinstall(const char*, Fmtlongconv);
int fmtengine(Format *format, const char *fmt);
int fmtprint(Format *format, const char *fmt,...);
int fmt_fprint_flush (Format *format, SIZE_T more);
int vfprint(int fd, const char *fmt, va_list ap);
int fprint(int fd, const char *fmt,...);
int print(const char *fmt,...);
int eprint(const char *fmt,...);
char *fmt_memprint (Format *format, const char *fmt, SIZE_T* len);
int fmt_memprint_grow (Format *format, SIZE_T more);
char *vsmprint (Alloc a, SIZE_T* len, const char* fmt, va_list ap);
char *smprint (Alloc alloc, SIZE_T* len, const char* fmt, ...);
int fmt_snprint_grow (Format *format, SIZE_T more);
char* palloc(char*, SIZE_T);
char *mprint(const char *fmt,...);
int vsnprint(char *buf, int buflen, const char *fmt, va_list ap);
int snprint(char *buf, int buflen, const char *fmt,...);
int sprint(char *buf, const char *fmt,...);
void fmtpad(Format *, SIZE_T, char);

/* 
	buffered output.

	format will flush when you reach bufsize. the maximum number of 
	calls to write(2) will be
		buffer_size \ min(buffer_size, sys_max_write) + 1

	do this:

		buf = xalloc(bigsize);

		fmt_fprint_init(fd,&format, buf, bigsize);
				print like a dog.
		fmt_print_close(&format);

		xfree(buf);
*/
int fmt_fprint_init(int fd, Format* format, char* buf, SIZE_T buf_size);
int fmt_fprint_close(Format* format);

#endif /* __nicer_print_h */
