/*
**
**  Copyright:		(c) 1992 AT&T - All rights reserved.
**
**  RCS Info:		$Date: 2005/07/23 19:08:25 $ $Revision: 1.4 $
**
**			$Id: rune.h,v 1.4 2005/07/23 19:08:25 quanstro Exp quanstro $
**
**  Filename:		$Source: /home/quanstro/src/elocate/include/RCS/rune.h,v $
**
**  Originator:		erik quanstrom
**			202 328 3688
**			quanstro@quanstro.net
**
**  Organization:	erik quanstrom
**			1455 chapin st nw
**			washington dc 20009
**			Ph: 202 328 3688	FAX: zippidy
**
**  Description:
**
**	utf-8 support from plan9; character classification from me.
**
**  Limitations:
**
**	1.  four- to six-byte sequences are not supported.
**	2.  Rune type is only 16 bits--fixing #1 will be difficult.
**
**  Modification History:
**
** 	07-jun-1998 quanstro Created.
**	02-sep-2005 quanstro added rune classification.
**
******************************************************************************/

typedef	unsigned short		Rune;

enum
{
	UTFmax		= 3,		/* maximum bytes per rune */
	Runesync	= 0x80,		/* cannot represent part of a utf sequence (<) */
	Runeself	= 0x80,		/* rune and utf sequences are the same (<) */
	Runeerror	= 0xFFFD	/* decoding error in utf */
};

/*
 * new rune routines
 */
extern	int	runetochar(char*, const Rune*);
extern	int	chartorune(Rune*, const char*);
extern	int	runelen(long);
extern	int	fullrune(const char*, const int);

/*
 * rune routines from converted str routines
 */
extern	long	utflen(const char*);		/* was countrune */
extern	char*	utfrune(char*, const long);
extern	char*	utfrrune(char*, const long);
extern	char*	utfutf(const char*, const char*);

/*
 * rune classification
 */
Rune tolowerrune(Rune);
Rune toupperrune(Rune);
Rune totitlerune(Rune);

int islowerrune(Rune);
int isupperrune(Rune);
int isalpharune(Rune);
int istitlerune(Rune);
int isspacerune(Rune);
int isdigitrune(Rune);

int digitrunevalue(Rune);
