#include <base.h>
#include <xalloc.h>
#include <lst.h>
	
#include <memory.h>

lstHead* lstStrstr2List(const char* start, unsigned long len, const char* string){
	lstHead* l;
	const char* p;
	size_t slen;
	size_t skip;

	l = lstCreate(0);
	if (0 == start){
		return l;
	}

	slen = strlen(string);
	for(;; start += skip, len -= skip){
		p = memchr(start, *string, len);
		if (!p){
			break;
		}
		skip = 1 + p - start;

		if (skip+slen > len+1){
			break;
		}
		if (0==memcmp(p, string, slen)){
			lstAddEntry(l, xcopyn(start, p-start));
		}
	}
	if (*start){
		lstAddEntry(l, xcopy(start));
	}
	return l;
}

lstHead* lstStr2List(const char* s, const char* usep){
	static const char* sep=" \t\n";
	const char* anchor;
	long len;
	lstHead* l;

	if (0 == usep){
		usep = sep;
	}

	l = lstCreate(0);
	if (0 == s){
		return l;
	}
	
	for (anchor=s+strspn(s,usep); (len=strcspn(anchor, usep)); ){
		lstAddEntry(l, xcopyn(anchor, len));
		anchor+=len;
		anchor+=strspn(anchor, usep);
	}

	return l;
}

static long strncspn(const char* s, long len, const char* usep){
	char map[0xff];
	const char* p;
	const char* end;
	int c;

	memset(map,0,0xff);
	while ((c=(uchar)*usep++))
		map[c]=1;
	for(p = s, end = s+len; map[(uchar)*s++]; ){
		if (s == end)
			return 0;
	}
	return p-s-1;
}

static long strnspn(const char*s, long len, const char* usep){
	char map[0xff];
	const char* p;
	const char* end;
	int c;

	memset(map,0,0xff);
	while ((c=(uchar)*usep++))
		map[c]=1;
	for(p = s, end = s+len; 0 == map[(uchar)*s++]; ){
		if (s == end)
			return 0;
	}
	return p-s-1;
}

lstHead* lstStrn2List(const char* s, long L, const char* usep){
	static const char* sep=" \t\n";
	const char* anchor;
	long len;
	long skip;
	lstHead* l;

	if (0 == usep){
		usep = sep;
	}

	l = lstCreate(0);
	if (0 == s){
		return l;
	}
	
	skip = strnspn(s, L, usep);
	anchor = s+skip;
	L -= skip;

	while ((len=strncspn(anchor, L, usep))){
		lstAddEntry(l, xcopyn(anchor, len));
		anchor+=len;
		L-=len;
		skip = strnspn(anchor, L, usep);
		anchor+=skip;
		L-=skip;
	}

	return l;
}

#if DAVID
static const char* fast_strstr(const char* start, size_t len, const char* string){
        const char* p;
        size_t slen;
        size_t skip;

        slen = strlen(string);
        for(;; start += skip, len -= skip){
                p = (char*)memchr(start, *string, len);
                if (!p){
                        return 0;
                }
                skip = 1 + p - start;

                if (skip+slen > len+1){
                        return 0;
                }
                if (0==memcmp(p, string, slen)){
                        return p;
                }
        }
        return 0;
}


lstHead* lstStrstr2List(const char* orig, const char* sep) {
	lstHeader l;
	const char* ptr;
	const char* nxt;
	char* tmp;
	int seplen, ptrlen, origlen;

	l = lstCreate(0);
	if (!orig || !*orig) return l;

	seplen = strlen(sep);
	origlen = strlen(orig);

	for (ptr = orig, ptrlen = origlen, nxt = fast_strstr(ptr, ptrlen, sep); ptr && nxt; ptr = nxt + seplen, ptrlen = origlen - (ptr - orig), nxt = fast_strstr(ptr, ptrlen, sep)) {
		lstAddEntry(l, copyn2(ptr, nxt - ptr));
	}
	if (ptr) {
		lstAddEntry(l, copy(ptr));
	}

	return l;
}
#endif
