#include <lstInternal.h>
#include <vec.h>
#include <memory.h>
#define internal static

static inline long newSize(unsigned long old_size){
	if (0 == old_size){
		old_size = 16;
	} else if (old_size < 4096/sizeof(void**)){
		old_size *= 2;
	} else {
		old_size += 4096;
	}
	return old_size;
}
		
static void vecCheckSize(vecHead*v){
	unsigned long oldp;

	if (v->n+1 < v->n_alloc){
		return;
	}
	oldp=v->n_alloc;
	v->n_alloc = newSize(v->n_alloc);
	v->vec = realloc(v->vec, v->n_alloc * sizeof(vecEntry));
	memset(v->vec + oldp, 0, sizeof(vecEntry)*(v->n_alloc-oldp));
}

vecHead* vecCreate(const lstFns* fns){
	vecHead* v;
	
	v=xalloc(sizeof(vecHead));
	v->n = 0;
	v->n_alloc = 0;
	v->vec=0;
	if (fns){
		v->f = fns;
	} else {
		v->f = &lstFnsString;
	}
	vecCheckSize(v);

	return v;
}

void vecDestroy(vecHead* v){
	vecEntry* e;

	for(e=vecGetFirst(v); e; e=vecGetNext(e)){
		v->f->delete(e->data);		
	}
	cxfree(v->vec);
	xfree(v);
}

vecHead* vecCopy(const vecHead* old){
	vecHead* v;
	const vecEntry* e;

	v = vecCreate(old->f);
	for(e=vecGetFirst(old); e; e = vecGetNext(e)){
		vecAddEntry(v, v->f->copy(e->data));
	}

	return v;
}

void vecAddEntry(vecHead* v, void*data){
	vecEntry* e;

	vecCheckSize(v);
	v->n++;
	e = v->vec+v->n;
	e->data = data;
}

void vecRemoveLast(vecHead*v){
	switch(vecCount(v)){
	case 0:
		break;
	default:
		v->f->delete((v->vec+v->n)->data);
		v->n--;
	}
}

vecEntry* vecGetFirst(const vecHead* v){
	return v->n ? v->vec+1 : 0;
}

vecEntry* vecGetLast(const vecHead* v){
	return v->n ? v->vec + v->n : 0;
}

vecEntry* vecGetNext(const vecEntry *e){
	const vecEntry* e2;
	e2=e+1;
	if (e2->data){
		return (vecEntry*)e2;
	}
	return 0;
}

vecEntry* vecGetPrev(const vecEntry *e){ /* cf. vecCheckSize() for trick that makes this legit */
	const vecEntry* e2;
	e2=e-1;
	if (e2->data){
		return (vecEntry*)e2;
	}
	return 0;
}

void* vecGetFirstData(const vecHead*v, vecEntry**e){
	*e = v->vec+1;
	if (*e){
		return (*e)->data;
	}
	return 0;
}

void* vecGetNextData(const vecHead*v, vecEntry**e){
	USED(v);
	*e=vecGetNext(*e);
	return *e ? (*e)->data : 0;
}

void* vecGetLastData(const vecHead*v, vecEntry**e){
	*e = v->vec+v->n;
	return (*e)->data;
}

void* vecGetPrevData(const vecHead*l, vecEntry**e){
	USED(l);
	*e = (*e)-1;
	if (*e){
		return (*e)->data;
	}
	return 0;
}


/* evil function */
#define CA (int(*)(const void *, const void *))
void vecSort(vecHead* v){
	quicksort_ptr((void**)(v->vec+1), vecCount(v), sizeof(vecEntry), v->f->compare);
}

vecEntry* vecScanForFn(const vecHead* v, const void* item, lstFnCompare compare){
	vecEntry* e;

	for(e=vecGetFirst(v); e; e=vecGetNext(e)){
		if (lstEqual == compare(e->data, item)){
			return e;
		}
	}
	return 0;
}

vecEntry* vecScanFor(const vecHead* v, const void* item){
	return vecScanForFn(v,item, v->f->compare);
}

boolean vecInVec(const vecHead* v, const void* item){
	if (vecScanFor(v,item)){
		return true;
	}
	return false;
}
