#include <mailhdr.h>
#include <rune.h>
#include <xalloc.h>
#include <latin_tab.h>
#include <latin2_tab.h>
#include <windows_tab.h>
#include <gb_tab.h>

#include <string.h>

static boolean gb2utf8(boolean* did, int base, const Rune* _tab, const char* bin, int inlen, char** bout, int* outlen){
	const char* p;
	const char* end;
	char* out;
	char* pout;
	int c;
	int oldc;
	int clen;
	USED(base);

	pout = out = xalloc(UTFmax*inlen/2);
	end = bin+inlen;
	oldc = -1;

	for(p=bin; p<end; ){
		c=(uchar)*p++;
		if (-1 != oldc){
			if (oldc>=0x81 && c>=0x40) {
				clen = runetochar(pout, &_tab[192*(oldc-0x81)+(c-0x40)]);
				pout+=clen;
			} else {
				pout += sprint(pout, "%C", Runeerror);
			}
			oldc=-1;
		} else if (c<0x80){
			*pout++=c;		// ascii
		} else {
			oldc=c;
		}
	}
	*bout = out;
	*outlen = pout-out;
	*did = true;

	return true;
}

static boolean latin2utf8(boolean* did, int base, const Rune*tab, const char*bin, int inlen, char** bout, int* outlen){
	const char* p;
	const char* end;
	char* out;
	char* pout;
	int c;
	int clen;

	pout = out = xalloc(UTFmax*inlen);
	
	end = bin+inlen;
	for(p=bin; p<end; ){
		c=(uchar)*p++;
		if (c<base){
			*pout++=c;
		} else {
			clen = runetochar(pout, &tab[c-base]);
			pout+=clen;
		}
	}
	*bout = out;
	*outlen = pout-out;
	*did = true;

	return true;
}

static boolean ascii2utf8(boolean* did, int base, const Rune*tab, const char*bin, int inlen, char** bout, int* outlen){
	USED(tab);
	USED(did);
	USED(base);

	*bout = (char*)bin;
	*outlen = inlen;

	return true;
}

typedef struct {
	const char* name;
	const Rune* tab;
	int base;
	boolean (*toutf8)(boolean* did, int base, const Rune*tab, const char*bin, int inlen, char** bout, int* outlen);
} Xlate;

// should bsearch this or something; this was supposed to be a small table. ☺
static Xlate Xtab[] = {
	{"US-ASCII",	0,		-1,	ascii2utf8	},
	{"UTF-8",		0,		-1,	ascii2utf8	},
	{"unknown-8bit",	0,		-1,	ascii2utf8	},	// chris' mailer.
	{"ISO-8859-1",	latin1_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-2",	latin2_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-3",	latin3_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-4",	latin4_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-5",	latin5_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-6",	latin6_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-7",	latin7_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-8",	latin8_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-9",	latin9_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-10",	latina_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-15",	latinf_tab,	0xa0,	latin2utf8	},
	{"windows-1250",	windows_1250,	0x80,	latin2utf8	},
	{"windows-1251",	windows_1251,	0x80,	latin2utf8	},
	{"windows-1252",	windows_1252,	0x80,	latin2utf8	},
	{"windows-1253",	windows_1253,	0x80,	latin2utf8	},
	{"windows-1254",	windows_1254,	0x80,	latin2utf8	},
	{"windows-1255",	windows_1255,	0x80,	latin2utf8	},
	{"windows-1256",	windows_1256,	0x80,	latin2utf8	},
	{"windows-1257",	windows_1257,	0x80,	latin2utf8	},
	{"windows-1258",	windows_1258,	0x80,	latin2utf8	},
	{"gb2312",	(Rune*)gb_tab,	0,	gb2utf8		},
};

#define nxlate	(sizeof(Xtab)/sizeof(Xtab[0]))

boolean toutf8(boolean* did, const char* charset, const char* bin, int inlen, char** bout, int* outlen){
	ul i;
	
	*did = false;
	if (!inlen){
		return true;
	}

	for(i=0; i<nxlate; i++){
		if (0 == strcasecmp(Xtab[i].name, charset)){
			return Xtab[i].toutf8(did, Xtab[i].base, Xtab[i].tab, bin, inlen, bout, outlen);
		}
	}
	return false;
}
