#define _GNU_SOURCE
#include <time.h>
#include <stdlib.h>
#include <memory.h>

#include <mailhdr.h>
#include <xalloc.h>
#include <print.h>

static int m_tab[]	= {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
static int l_tab[] 	= {31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

static int leapyear(int y){
	y += 1900;
	if ((0 == y%100 && y%400) || y%4)
		return false;
	return true;
}

static int mdays(int m, int leap){
	int *tab;
	int j;
	int d;

	tab=m_tab;
	if (leap) {
		tab=l_tab;
	}

	for(d=0, j=0; j<m; j++){
		d += tab[j];
	}
	return d;
}

static int daysm(int d, int leap, int* m){
	int *tab;
	int j;

	tab=m_tab;
	if (leap) {
		tab=l_tab;
	}

	for(j=0; d>tab[j]; j++){
		d -= tab[j];
	}
	*m = j;
	return d;
}

static int ydays(int y){
	int d;

	for (d=0; 70 <= --y; ){
		if (leapyear(y)){
			d+= 1;
		}
		d+=365;
	}
	return d;
}
		
	// we just don't trust glibc;
static void adjust_tm(struct tm* tm, int diffsec){
	int leap;
	int sec;
	int y;

	leap = leapyear(tm->tm_year);

	sec = tm->tm_sec + tm->tm_min*60 + tm->tm_hour*3600;
	sec += ( ydays(tm->tm_year) + mdays(tm->tm_mon, leap) + tm->tm_mday - 1 ) * 86400;
	sec += diffsec;

	tm->tm_sec	= sec % 60;	sec /=60;
	tm->tm_min	= sec % 60;	sec /=60;
	tm->tm_hour	= sec % 24;	sec /= 24;

	for(y=70; sec >= 365; y++){
		if (leapyear(y)){
			if (365==y){
				break;
			}
			sec--;
		}
		sec -= 365;
	}

	tm->tm_year = y;
	tm->tm_yday = sec;
	tm->tm_mday = 1 + daysm(sec, leapyear(y), &tm->tm_mon);
	tm->tm_wday = (tm->tm_wday + diffsec/86400) % 7;
}

static char* do_date(const char* date, const char* fmt, ul* conv_sec){
	const char* dflt_fmt = "%Y.%m.%d %H:%M";
	struct tm tm;
	const char* s;
	char buf[20];
	time_t sec;
	int delta;

	memset(&tm,0,sizeof(tm));
	if (!fmt){
		fmt = dflt_fmt;
	}

	s = strptime(date, "%A, %d %B %Y %H:%M:%S", &tm);
	if (s && (delta=strtol(s, 0, 10))){
		sec = mktime(&tm);
		localtime_r(&sec, &tm);
		sec -= -tm.tm_gmtoff +delta*36 + tm.tm_isdst*3600;
		localtime_r(&sec, &tm);
		strftime(buf, sizeof(buf), fmt, &tm);
	} else {
		strftime(buf, sizeof(buf), fmt, &tm);	
		sec = mktime(&tm);
		sec -= tm.tm_isdst*3600;
	}

	*conv_sec=sec;
	return xcopy(buf);
}

static const char* test_tab[] = {
	"Sat, 17 Sep 2005 00:14:57 PDT",
	"Sat, 17 Sep 2005 01:29:11 +0200",
	"Sun, 31 Oct 2004 13:48:55 -0500",
	"Wed, 23 Mar 2005 12:58:03 +0100",
	"Sat, 14 May 2005 01:57:34 +0000",
	"Wed, 25 May 2005 09:03:53 -0600 (MDT)",
	"Sun, 29 May 2005 22:00:16 -0600 (MDT)",
	"Fri, 3 Jun 2005 08:45:15 -0600 (MDT)",
	"Fri, 3 Jun 2005 12:54:10 -0600 (MDT)",
	"Mon, 6 Jun 2005 06:08:00 -0600",
	"Sat, 18 Jun 2005 18:52:53 -0600 (MDT)",
	"Mon, 25 Jul 2005 18:01:32 +0200",
	"Sat, 13 Aug 2005 22:31:17 -0400",
	"Mon, 15 Aug 2005 12:23:22 +0200",
	"Mon, 15 Aug 2005 14:37:46 +0200",
	"Tue, 23 Aug 2005 19:10:47 -0700",
	"Sat, 27 Aug 2005 23:55:47 -0400",
	"Fri, 2 Sep 2005 15:47:11 +1000",
	"Fri, 9 Sep 2005 13:14:18 -0500",
	"Mon, 12 Sep 2005 09:20:12 -0500",
	"Tue, 13 Sep 2005 17:26:07 -0700",
	"Fri, 16 Sep 2005 18:20:08 +0100",
	"Fri, 16 Sep 2005 13:18:31 -0400",
	"Fri, 16 Sep 2005 18:48:58 -0600",
	"Fri, 16 Sep 2005 21:42:25 -0500 (CDT)",
	"Fri, 16 Sep 2005 22:00:08 -0500 (CDT)",
	0
};

void main(void){
	const char** s;
	char* result;
	ul sec;

	for(s=test_tab; *s; s++){
		result=do_date(*s, 0, &sec);
		print("%s\n", result);
		xfree(result);
	}
}
