#include <mailhdr.h>
#include <string.h>
#include <xalloc.h>
#include <rune.h>

#include <assert.h>
#include <ctype.h>

typedef enum {
	bad_okay=0,
	bad_quote=1,		// must quote
	bad_slash=1<<1,		// must \-escape.
	bad_2047=1<<2,		// must rfc-2047-encode.
	bad_dot=1<<3		// . 2822 wierdness
} bad;

typedef enum {
	quote_no = false,
	quote_yes = true,
	quote_2047 = 2
} qtype;

static const char* qtab[2+1] = {
	0,
	"\"",
	0
};

static const bad atom_tab[Runeself+1] = {
	['\n']	= 	bad_2047,
	['\r']	= 	bad_2047,
	['\t']	= 	bad_2047,
	[' '] 	= 	bad_quote,
	['"']	=	bad_slash,
	[',']	=	bad_quote,
	['(']	=	bad_quote,
	[')']	=	bad_quote,
	['.']	=	bad_dot,		// only if the next character is a space.
	[':']	=	bad_quote,
	[';']	=	bad_quote,
	['<']	=	bad_quote,
	['>']	=	bad_quote,
	['@']	=	bad_quote,
	['[']	=	bad_quote,
	['\\']	=	bad_slash,
	[']']	=	bad_quote,
};

static const bad phrase_tab[Runeself+1] = {
	['\n']	= 	bad_2047,
	['\r']	= 	bad_2047,
	['\t']	= 	bad_2047,
//	[' '] 	= 	bad_quote,
	['"']	=	bad_slash,
	[',']	=	bad_quote,
	['(']	=	bad_quote,
	[')']	=	bad_quote,
	['.']	=	bad_dot,		// special only iff first or last character in dot-atom.
	[':']	=	bad_quote,
	[';']	=	bad_quote,
	['<']	=	bad_quote,
	['>']	=	bad_quote,
	['@']	=	bad_quote,
	['[']	=	bad_quote,
	['\\']	=	bad_slash,
	[']']	=	bad_quote,
};

#if 0
// rules are different for mimestrings; here's the table

static const bad mime_tab[Runeself+1] = {
	['\n']	=	bad_2047,
	['"']	=	bad_slash,
	[',']	=	bad_quote,
	['/']	=	bad_quote,
	['=']	=	bad_quote,
	['(']	=	bad_quote,
	[')']	=	bad_quote,
	['.']	=	bad_quote,
	[':']	=	bad_quote,
	[';']	=	bad_quote,
	['<']	=	bad_quote,
	['>']	=	bad_quote,
	['?']	=	bad_quote,
	['@']	=	bad_quote,
	['[']	=	bad_quote,
	['\\']	=	bad_slash,
	[']']	=	bad_quote,
};
#endif

// bend the rules for viewage.
static qtype quotable_view(const bad* tab, const char*s, SIZE_T len, SIZE_T f2, SIZE_T* fmt_chars){
	SIZE_T i;
	SIZE_T c;
	qtype q;
	SIZE_T b;
	bad r;
	Rune rune;
	unsigned char uc;

	q=quote_no;
	for(i=0,b=0,c=0;c<len && b<f2;c++){
		uc = (unsigned char)s[i];
		if (Runeself <= uc){
			i+=chartorune(&rune, s+i);
			r=bad_okay;
		} else {
			r=tab[uc];
			i+=1;
		}
		switch(r){
		case bad_slash:
			if(b+1>=f2){
				goto done;
			}
			b+=2;
			break;
		case bad_dot:
			if (!(0 == i || len == i) && !isspace(s[i])) {
				b++;
				break;
			}
		case bad_quote:
			if (q!=quote_yes){
				if (b+2 >= f2){
					goto done;
				}
				q=quote_yes;
				b+=2;
			}
		default:
			b++;
		}
	}
done:
	*fmt_chars=b;
	return q;
}

// strict rules for sending ze email
static qtype quotable_send(const bad* tab, const char*s, SIZE_T len, SIZE_T f2, SIZE_T* fmt_chars){
	SIZE_T i;
	SIZE_T c;
	qtype q;
	SIZE_T b;
	bad r;
	unsigned char uc;

	q=quote_no;
	for(i=0,b=0,c=0;c<len && b<f2;c++){
		uc = (unsigned char)s[i];
		if (uc >= 0x20 && uc<=0x7f){
			r=tab[uc];
		}else{
			r=bad_2047;
		}
		i+=1;

		switch(r){
		case bad_2047:
			// screw it: encode-o-rama, and you can keep your stinking f2;
			xfree(smprint(palloc, &b, "=?utf-8?B?" "%.*{base64}" "?=", len, s));
			q=quote_2047;
			goto done;
		case bad_slash:
			if(b+1>=f2){
				goto done;
			}
			b+=2;
			break;
		case bad_dot:
			if (!(0 == i || len == i) && !isspace(s[i])) {
				b++;
				break;
			}
		case bad_quote:
			if (q!=quote_yes){
				if (b+2 >= f2){
					goto done;
				}
				q=quote_yes;
				b+=2;
			}
		default:
			b++;
		}
	}
done:
	*fmt_chars=b;
	return q;
}

// this function is a little wrong. only one term < 70-some characters should
// be stuffed in this way. 
static void stuff_send(Format* f, SIZE_T chars, const char* s, qtype q, const bad* tab){
	SIZE_T b;
	SIZE_T n;
	bad r;
	unsigned char uc;
	const char* quote;
	int l;

	if (quote_2047 == q){
		fmtprint(f, "=?utf-8?B?");
		fmtprint(f, "%.*{base64}", chars, s);
		fmtprint(f, "?=");
		return;
	}
		
	b=0;
	quote = qtab[(int)q];
	if (quote){
		l = strlen(quote);
		fmtappend(f,quote,l);
		b+=2*l;
	}

	for(;b<chars;s+=n){
		uc = (unsigned char)*s;
		if (Runeself <= uc){
			goto do2047;
		} else {
			r=tab[uc];
			n=1;
		}
		switch(r){
		case bad_2047:
		do2047:
			assert(0 == "cant get here");
		case bad_slash:
			b++;
			if (b>=chars){
				goto done;		/* can't leave unstuffed garbage */
			}
			fmtappend(f,"\\", 1);
		case bad_dot:
		default:
			fmtappend(f,s,n);
			b++;
		}
	}
done:
	if(quote){
		fmtappend(f,quote,l);
	}
}

static void stuff_view(Format* f, SIZE_T chars, const char* s, qtype q, const bad* tab){
	SIZE_T b;
	SIZE_T n;
	bad r;
	Rune rune;
	unsigned char uc;
	const char* quote;
	int l;

	b=0;
	quote = qtab[(int)q];
	if (quote){
		l = strlen(quote);
		fmtappend(f,quote,l);
		b+=2*l;
	}

	for(;b<chars;s+=n){
		uc = (unsigned char)*s;
		if (Runeself <= uc){
			n=chartorune(&rune, s);
			r=bad_okay;
		} else {
			r=tab[uc];
			n=1;
		}
		switch(r){
		case bad_slash:
			b++;
			if (b>=chars){
				goto done;		/* can't leave unstuffed garbage */
			}
			fmtappend(f,"\\", 1);
		case bad_dot:
		default:
			fmtappend(f,s,n);
			b++;
		}
	}
done:
	if(quote){
		fmtappend(f,quote,l);
	}
}

static fmt_action p(qtype (*view)(const bad*, const char*, SIZE_T, SIZE_T, SIZE_T*), void (*stuff)(Format*, SIZE_T, const char*, qtype, const bad*), const bad* tab, Format *f, const char* fmt){
	const char* s;
	SIZE_T extra;
	SIZE_T maxbytes;
	SIZE_T maxconv;
	SIZE_T minwidth;
	qtype quote;
	USED(fmt);

	s = va_arg(f->args, const char*);
	if (!s) s = "(null)";

	if (f->flags & FMT_altform) {
		maxbytes = (f->flags & FMT_f2set) ? f->f2 : (SIZE_T)utflen(s);
	} else {
		maxbytes = utflen(s);
		if ((f->flags & FMT_f2set) && (f->f2 < maxbytes)){
			maxbytes = f->f2;
		}
	}

	quote = view(tab, s, maxbytes,(f->flags & FMT_f2set) ? f->f2 : ~0UL, &maxconv);
	minwidth = (f->flags & FMT_f1set) ? (SIZE_T)f->f1 : maxbytes;

	if (minwidth>maxconv){
		extra = minwidth-maxconv;
	} else {
		extra=0;
	}

	if (f->flags & FMT_leftside) {
		stuff(f,maxconv, s, quote, tab);
		fmtpad(f, extra, ' ');
	} else {
		fmtpad(f, extra, ' ');
		stuff(f,maxconv, s, quote, tab);
	}

	return FMT_verb;
}

fmt_action long_atom_conv(Format *f, const char* fmt){
	return p(quotable_view, stuff_view, atom_tab, f, fmt);
}

fmt_action long_fullatom_conv(Format *f, const char* fmt){
	return p(quotable_send, stuff_send, atom_tab, f, fmt);
}

fmt_action long_phrase_conv(Format *f, const char* fmt){
	if (f->flags & FMT_altform)
		return long_fullphrase_conv(f,fmt);
	return p(quotable_view, stuff_view, phrase_tab, f, fmt);
}

fmt_action long_fullphrase_conv(Format *f, const char* fmt){
	return p(quotable_send, stuff_send, phrase_tab, f, fmt);
}
