#include <mailhdr.h>
#include <rune.h>
#include <xalloc.h>
#include <latin_tab.h>
#include <latin2_tab.h>
#include <windows_tab.h>
#include <gb_tab.h>
#include <jis_tab.h>
#include <big5_tab.h>

#include <string.h>
#include <assert.h>

static boolean gb2utf8(boolean* did, int base, const Rune* _tab, const char* bin, int inlen, char** bout, int* outlen){
	const char* p;
	const char* end;
	char* out;
	char* pout;
	uchar c;
	int oldc;
	int clen;
	USED(base);

	pout = out = xalloc(UTFmax*inlen/2);
	end = bin+inlen;
	oldc = -1;

	for(p=bin; p<end; ){
		c=(uchar)*p++;
		if (-1 != oldc){
			if (oldc>=0x81 && c>=0x40) {
				clen = runetochar(pout, &_tab[192*(oldc-0x81)+(c-0x40)]);
				pout+=clen;
			} else {
				pout += sprint(pout, "%C", Runeerror);
			}
			oldc=-1;
		} else if (c<0x80){
			*pout++=c;		// ascii
		} else {
			oldc=c;
		}
	}
	*bout = out;
	*outlen = pout-out;
	*did = true;

	return true;
}

// hkscs range: 88-8b, 9d-a0, c6-c8, f9-fe
static boolean big52utf8(boolean* did, int base, const Rune* _tab, const char* bin, int inlen, char** bout, int* outlen){
	const char* p;
	const char* end;
	char* out;
	char* pout;
	uchar c;
	int c1;
	int clen;
	USED(base);

	pout = out = xalloc(UTFmax*inlen/2);
	end = bin+inlen;

	for(p=bin; p<end; ){
		c=(uchar)*p++;
		if (c<0x80){
			*pout++ = c;
		} else {
			c1=(uchar)*p++;
			if (c<0xa1 || c>0xf9 || c==0xc7 || c==0xc8){
				pout += sprint(pout, "%C", Runeerror);
			} else {
				if (c>0xc6)
					c-=2;
				c-= 0xa1;
				clen = runetochar(pout, &_tab[224*c+c1]);
				pout+=clen;
			}
		}
	}
	*bout = out;
	*outlen = pout-out;
	*did = true;

	return true;
}

static boolean latin2utf8(boolean* did, int base, const Rune*tab, const char*bin, int inlen, char** bout, int* outlen){
	const char* p;
	const char* end;
	char* out;
	char* pout;
	uchar c;
	int clen;

	pout = out = xalloc(UTFmax*inlen);
	
	end = bin+inlen;
	for(p=bin; p<end; ){
		c=(uchar)*p++;
		if (c<base){
			*pout++=c;
		} else {
			clen = runetochar(pout, &tab[c-base]);
			pout+=clen;
		}
	}
	*bout = out;
	*outlen = pout-out;
	*did = true;

	return true;
}

static boolean ascii2utf8(boolean* did, int base, const Rune*tab, const char*bin, int inlen, char** bout, int* outlen){
	USED(tab);
	USED(did);
	USED(base);

	*bout = (char*)bin;
	*outlen = inlen;

	return true;
}

static const Rune rune_feff = 0xfeff;

static boolean mustswap(Rune feff){
	static const uchar rb[2] = "\xfe\xff";

//	if (feff != (Rune)rb)
	if (memcmp(&feff, rb, 2))
		return true;
	return false;
}


static void byteswap(Rune feff, char* _buf, SIZE_T len){
	uchar* buf;
	uchar s;
	uchar* end;

	assert(0 == len%2);
	buf = (uchar*)_buf;
	if (!mustswap(feff))
		return;

	for(end=buf+len; buf<end; buf+=2){
		s=buf[0];
		buf[0] = buf[1];
		buf[1] = s;
	}
}

static const uchar map64[] = {
	['A'] = 1,	['B'] = 1,	['C'] = 1,	['D'] = 1,	['E'] = 1,	['F'] = 1,	['G'] = 1,	['H'] = 1,
	['I'] = 1,	['J'] = 1,	['K'] = 1,	['L'] = 1,	['M'] = 1,	['N'] = 1,	['O'] = 1,	['P'] = 1,
	['Q'] = 1,	['R'] = 1,	['S'] = 1,	['T'] = 1,	['U'] = 1,	['V'] = 1,	['W'] = 1,	['X'] = 1,
	['Y'] = 1,	['Z'] = 1,	['a'] = 1,	['b'] = 1,	['c'] = 1,	['d'] = 1,	['e'] = 1,	['f'] = 1,
	['g'] = 1,	['h'] = 1,	['i'] = 1,	['j'] = 1,	['k'] = 1,	['l'] = 1,	['m'] = 1,	['n'] = 1,	
	['o'] = 1,	['p'] = 1,	['q'] = 1,	['r'] = 1,	['s'] = 1,	['t'] = 1,	['u'] = 1,	['v'] = 1,	
	['w'] = 1,	['x'] = 1,	['y'] = 1,	['z'] = 1,	['0'] = 1,	['1'] = 1,	['2'] = 1,	['3'] = 1,	
	['4'] = 1,	['5'] = 1,	['6'] = 1,	['7'] = 1,	['8'] = 1,	['9'] = 1,	['+'] = 1,	['/'] = 1,
};
static const char* encode64 = 
"ABCDEFGHIJKLMNOPQRSTUVWXYZ" "abcdefghijklmnopqrstuvwxyz" "0123456789" "+/";

static const uchar map64alt[] = {
	['A'] = 1,	['B'] = 1,	['C'] = 1,	['D'] = 1,	['E'] = 1,	['F'] = 1,	['G'] = 1,	['H'] = 1,
	['I'] = 1,	['J'] = 1,	['K'] = 1,	['L'] = 1,	['M'] = 1,	['N'] = 1,	['O'] = 1,	['P'] = 1,
	['Q'] = 1,	['R'] = 1,	['S'] = 1,	['T'] = 1,	['U'] = 1,	['V'] = 1,	['W'] = 1,	['X'] = 1,
	['Y'] = 1,	['Z'] = 1,	['a'] = 1,	['b'] = 1,	['c'] = 1,	['d'] = 1,	['e'] = 1,	['f'] = 1,
	['g'] = 1,	['h'] = 1,	['i'] = 1,	['j'] = 1,	['k'] = 1,	['l'] = 1,	['m'] = 1,	['n'] = 1,	
	['o'] = 1,	['p'] = 1,	['q'] = 1,	['r'] = 1,	['s'] = 1,	['t'] = 1,	['u'] = 1,	['v'] = 1,	
	['w'] = 1,	['x'] = 1,	['y'] = 1,	['z'] = 1,	['0'] = 1,	['1'] = 1,	['2'] = 1,	['3'] = 1,	
	['4'] = 1,	['5'] = 1,	['6'] = 1,	['7'] = 1,	['8'] = 1,	['9'] = 1,	['+'] = 1,	[','] = 1,
};
static const char* encode64alt = 
"ABCDEFGHIJKLMNOPQRSTUVWXYZ" "abcdefghijklmnopqrstuvwxyz" "0123456789" "+,";

static const char*sixtyfourend(const uchar* map, const char* s, int l){
	const char* end;
	end=s+l;
	while (map[(int)*s] && s<end){
		s++;
	}
	return s;
}

static boolean _utf72utf8(boolean* did, const char*bin, int inlen, char** bout, int* outlen, const uchar* map, const char* t, const char* E){
	const char* p;
	const char* end;
	char* out;
	char* pout;
	int r;
	const char* sep[2];
	char* raw;
	SIZE_T six4;

	// premature optimization is the root of all evil.
	sep[0] = memchr(bin, t[0], inlen);
	if (!sep[0]){
		return ascii2utf8(did, 0, 0, bin, inlen, bout, outlen);
	}

	pout = out = xalloc(inlen);	// can't get longer.
	end = bin+inlen;

	for(p=bin; p<end; p=sep[1]+1){
		sep[0] = memchr(p, t[0], end-p);

		r = sep[0] ? sep[0]-p : end - p;
		memcpy(pout, p, r);
		pout+=r;
		if (!sep[0]){
			break;
		}

		sep[1] = sixtyfourend(map, sep[0], end-sep[0]);
		if (!sep[1]){
			sep[1] = end;
		}
		
		// special case
		r = sep[1] - sep[0] - 1;
		if (0 == r){
			*pout++ = t[0];
		} else {
			raw = smprint(palloc, &six4, "%#.*/{unbase64}", r, E, sep[0]+1);
			//raw = mprint("%#.*{unbase64}", r, sep[0]+1);
			byteswap(rune_feff, raw, six4);
			pout += snprint(pout, inlen-(pout-out), "%#.*S", six4/2, raw);
			xfree(raw);
		}
	}

	*bout = out;
	*outlen = pout-out;
	*did = true;

	return true;
}

// useless and depricated.
static boolean utf72utf8(boolean* did, int base, const Rune*tab, const char*bin, int inlen, char** bout, int* outlen){
	static const char t[] = { '+', '-'};
	USED(base);
	USED(tab);
	return _utf72utf8(did, bin, inlen, bout, outlen, map64, t, encode64);
}

// not really used for anything except for IMAP.
static boolean utf7alt2utf8(boolean* did, int base, const Rune*tab, const char*bin, int inlen, char** bout, int* outlen){
	static const char t[] = { '&', '-'};
	USED(base);
	USED(tab);
	return _utf72utf8(did, bin, inlen, bout, outlen, map64alt, t, encode64alt);
}

static boolean utf162utf8(boolean* did, int base, const Rune*tab, const char*bin, int inlen, char** bout, int* outlen){
	char* rbuf;
	Rune r;
	USED(base);
	USED(tab);

	if (inlen<2){
		*bout = 0;
		*outlen = 0;
		return true;
	}

	r = *(Rune*)bin;
	if (r != 0xfeff && r != 0xfffe){
		return false;
	}

	if (mustswap(*bin)){
		inlen -= 2;
		rbuf = xcopyn(bin+2, inlen);
		byteswap(*bin, rbuf, inlen);
		*bout = smprint(palloc, (SIZE_T*)outlen, "%#.*S", inlen, rbuf);
		xfree(rbuf);
	} else {
		*bout = smprint(palloc, (SIZE_T*)outlen, "%#.*S", inlen, bin);
	}
	*did = true;
	return true;
}

static boolean utf16LE2utf8(boolean* did, int base, const Rune*tab, const char*bin, int inlen, char** bout, int* outlen){
	char* rbuf;
	USED(base);
	USED(tab);

	if (mustswap(0xfffe)){
		inlen -= 2;
		rbuf = xcopyn(bin+2, inlen);
		byteswap(*bin, rbuf, inlen);
		*bout = smprint(palloc, (SIZE_T*)outlen, "%#.*S", inlen, rbuf);
		xfree(rbuf);
	} else {
		*bout = smprint(palloc, (SIZE_T*)outlen, "%#.*S", inlen, bin);
	}
	*did = true;
	return true;
}

static boolean utf16BE2utf8(boolean* did, int base, const Rune*tab, const char*bin, int inlen, char** bout, int* outlen){
	char* rbuf;
	USED(base);
	USED(tab);

	if (mustswap(0xfeff)){
		inlen -= 2;
		rbuf = xcopyn(bin+2, inlen);
		byteswap(*bin, rbuf, inlen);
		*bout = smprint(palloc, (SIZE_T*)outlen, "%#.*S", inlen, rbuf);
		xfree(rbuf);
	} else {
		*bout = smprint(palloc, (SIZE_T*)outlen, "%#.*S", inlen, bin);
	}
	*did = true;
	return true;
}

typedef struct {
	const char* name;
	const Rune* tab;
	int base;
	boolean (*toutf8)(boolean* did, int base, const Rune*tab, const char*bin, int inlen, char** bout, int* outlen);
} Xlate;

/* shift jis encoding: for j≡first byte; k≡second byte

when j is from 33 to 96 (red/orange), s = (j+1)/2 + 112
when j is from 97 to 126 (blue/turquoise), s = (j+1)/2 + 176
when j is odd (red/blue), t = k + 31, plus one more if k > 95
when j is even (orange/turquoise), t = k + 126

*/

static boolean sjis2utf8(boolean* did, int base, const Rune* tab, const char* buf, int inlen, char** bout, int* outlen){
	char* out;
	char* outp;
	uchar* p;
	uchar* end;
	uchar c;
	int s;
	int odd;
	USED(base);

	if (inlen%2){
		return false;
	}

	out = outp = xalloc(UTFmax*(inlen/2+1));
	end = (uchar*)(buf+inlen);
	for(p = (uchar*)buf; p<end; p++) {
		c=*p++;
		odd = c%2;
		s = (c+1)/2;
		if (s>=33 && s<=126){
			if (s<=96){
				s+=112;
			} else {
				s+=176;
			}
		} else {
			goto loose;
		}

		c=*p++;
		s = (s<<8)+c;

		if (odd){
			if (c>95)
				s++;
			s+=31;
		} else {
			s+=126;
		}

		outp += runetochar(outp, &tab[s]);
	}

	*did = true;
	*bout = out;
	*outlen = outp-out;
	return true;

loose:
	xfree(out);
	return false;
}

// could be called ku-ten to utf-8;
// the shifts (subtractions) are due to table compression; 
static boolean jis2utf8(boolean* did, int base, const Rune* tab, const char* buf, int inlen, char** bout, int* outlen){
	char* out;
	char* outp;
	uchar* p;
	uchar* end;
	uchar c;
	int s;
	USED(base);

	if (inlen%2){
		return false;
	}

	out = outp = xalloc(UTFmax*(inlen/2+1));
	end = (uchar*)(buf+inlen);
	for(p = (uchar*)buf; p<end; p++) {
		c=*p++;
		if (c<0x21 || c>0x7f)
			goto loose;
		c-=0x21;
		s=c*(0x80-0x20);

		c=*p++;
		if (c<0x21 || c>0x7f)
			goto loose;
		s+=c-0x21;
		outp += runetochar(outp, &tab[s]);
	}

	*did = true;
	*bout = out;
	*outlen = outp-out;
	return true;

loose:
	xfree(out);
	return false;
}

typedef struct {
	int iso;
	const char* desc;
	const char* name;
	const Rune* tab;
	int base;
	boolean (*toutf8)(boolean* did, int base, const Rune*tab, const char*bin, int inlen, char** bout, int* outlen);
} Xlate2;

/* full table: 
	x is the buffer; 
	y is the charset type: e.g. single-byte 94-character.

	g0	g1	g2	g3
1/94	(	)	*	+	
1/96	 	-	.	/
n/94	$(	$)	$*	$+	g0 exception $[@AB]
n/96		$-	$.	$/

1-byte 94-character
@	ISO 646 IRV: 1983 
A	BS 4730 (UK) 
B	ANSI X3.4-1968 (ASCII) 
C	NATS Primary Set	Finland and Sweden 
I	JIS X 0201 Katakana 
J	JIS X 0201 Roman (Latin) 

1-byte 96-character A
A	ISO 8859-1 Latin-1 
B	ISO 8859-2 Latin-2 
C	ISO 8859-3 Latin-3 
D	ISO 8859-4 Latin-4 
F	ISO 8859-7 Latin/Greek 
G	ISO 8859-6 Latin/Arabic 
H	ISO 8859-8 Latin/Hebrew 
L	ISO 8859-5 Latin/Cyrillic 

multibyte 94-character 
@	JISX 0208-1978 Japanese 
A	GB 2312-80 Chinese 
B	JISX 0208-1983 Japanese 
C	KSC 5601 Korean 
D	JISX 0212-1990 Japanese 
E	CCITT Extended GB (ISO-IR-165) 
F	CNS 11643-1992 Set 1 (Taiwan) 
H	CNS 11643-1992 Set 2 (Taiwan) 
I	CNS 11643-1992 Set 3 (Taiwan) 
J	CNS 11643-1992 Set 4 (Taiwan) 
K	CNS 11643-1992 Set 5 (Taiwan) 
L	CNS 11643-1992 Set 6 (Taiwan) 
M	CNS 11643-1992 Set 7 (Taiwan)

*/

static const Xlate2 j_charset[] = {
	{ 42,	"jis x 0208-1978",		"$@",	jisx_208,		-1,	jis2utf8		},
	{ 149,	"ksg5601-1987",		"$(C",	0,		-1,	jis2utf8		},
	{ 159,	"jis x 0212-1990",		"$(D",	jisx_212,		-1,	jis2utf8		},
	{ 58,	"gb2312-1980",		"$A",	(Rune*)gb_tab,	0,	gb2utf8		},
	{ 87,	"jis x 0208-1983",		"$B",	jisx_208,		-1,	jis2utf8		},
	{ 6,	"ascii",			"(B",	0,		0xff,	ascii2utf8	},
	{ 14,	"jis x 0201-roman",	"(J",	jis_roman_tab,	0x50,	latin2utf8	},
};

static const Xlate2* Xascii = &j_charset[5];

#define nj_charset (sizeof j_charset / sizeof j_charset[0])
#define ESC 033

const Xlate2* encoding(const char* s, int l){
	unsigned int i;
	const char* p;
	const char* t;
	const char* end;

	end = s+l;

	for(i=0; i<nj_charset; i++){
		p = j_charset[i].name;
		t = s;
		while (t<end && *p == *t){
			p++;
			t++;
		}
		if (t<=end && ! *p){
			return &j_charset[i];
		}
	}
	return 0;
}

static const char* esc_end(const char* s, int l, const Xlate2* X){
	const char* end;
	char c;

	// deep ... cheating we don't need to go line-by-line
	// if we are converting ascii, a common case.
	if (X == Xascii){
		return memchr(s, ESC, l);
	}

	end = s+l;

	while (s<end && '\n' != (c = *s) && ESC != c){
		s++;
	}
	if (s==end)
		s--;
	return s;
}

// @ u+0040 is the index
const Rune* g2_tab[0x80] = {
	['A'] = latin1_tab,
	['B'] = latin2_tab,
	['C'] = latin3_tab,
	['D'] = latin4_tab,
	['F'] = latin7_tab,
	['G'] = latin6_tab,
	['H'] = latin8_tab,
	['L'] = latin5_tab,
};

static boolean encode2utf8(boolean* did, int base, const Rune* tab, const char* s, int inlen, char** bout, int* outlen){
	const char* p;
	int skip;
	const Rune* g2;
	int c;
	char* out;
	char* pout;
	char* part;
	boolean pdid;
	int partlen;
	const Xlate2* X;
	static const int g2_base = 0xa0;		// BUG()
	/*boolean*/ int reset;

	p = memchr(s, ESC, inlen);
	if (!p){
		return ascii2utf8(did,base,tab,s,inlen,bout,outlen);
	}

	pout = out = xalloc(UTFmax*inlen/2);		// codes are 2 bytes
	g2 = 0;
	reset = true;
	for(; inlen && p; s=p, p = esc_end(s, inlen, X)){
		if (reset){
			X = Xascii;
			reset = false;
		}
		skip=p-s;
		if (skip){
			pdid = false;
			if (!X->toutf8(&pdid, X->base, X->tab, s, skip, &part, &partlen)){
				goto loose;
			}
			memcpy(pout, part, partlen);
			pout+=partlen;
			if (pdid){
				xfree(part);
			}
			inlen -= skip;
		}
		if (!inlen){
			break;
		}
		reset = '\n' == *p++;
		inlen--;
		if (reset){
			*pout++ = '\n';
			continue;
		}
		if (!inlen){
			break;
		}
		if (inlen<2){
			goto loose;
		}

		c = p[0];
		if ('.' == c) {
			if (inlen<2){
				goto loose;
			}
			c=*++p;
			if (c>=0x7f){
				goto loose;
			}
			g2 = g2_tab[c];
			if (!g2){
				goto loose;
			}
			p++; inlen -= 2;
			continue;
		} else if ('N' == c){
			if (inlen<2){
				goto loose;
			}
			p++; inlen-=2;
			c=0xff & (*p++ + 0x80);
			if (!g2){
				goto loose;
			}
			if (c<g2_base){
				*pout++=(uchar)c;
			} else {
				pout += runetochar(pout, &g2[c-g2_base]);
			}
			continue;
		}

		if (inlen<3){
			goto loose;
		}

		X = encoding(p, inlen);
		if (!X) {
			goto loose;
		}

		partlen = strlen(X->name);
		p += partlen;
		inlen -= partlen;
	}

	memcpy(pout, s, inlen);
	pout += inlen;

	*bout = out;
	*outlen = pout - out;
	*did = true;
	return true;

loose:
	xfree(out);
	return false;
}

// BUG() non-64bit compatable.
static uchar cbucky = 0x80;
static ul lbucky = 0x80808080UL;

const uchar*
buckybit(const char* _buf, int n)
{
	const uchar* buf;
	ul *lbuf;
	ul* end;
	ul* p;
	int r;
	int m;

	m = sizeof(ul);
	if (0 == n)
		return 0;
	buf = (const uchar*)_buf;

	r=n%m;
	if (0 == r)
		r=m;
	n-=r;
	lbuf = (ul*)buf;
	end = lbuf + n/m;

	for(p = lbuf; p<end; p++)
		if (*p & lbucky)
			break;

	if (p<end)
		r=m;
	buf=(uchar*)p;
	for(buf=(uchar*)p; r--; buf++)
		if (*buf & cbucky)
			return buf;

	return 0;
}

// guess-o-matic
boolean unknowntoutf8(boolean* did, int base, const Rune* tab, const char* s, int inlen, char** bout, int* outlen){
	const char* p;
	int l;
	Rune r;
	USED(base);
	USED(tab);

	p = (const char*)buckybit(s, inlen);
	if (0 == p){
		return toutf8(did, "us-ascii", s, inlen, bout, outlen);
	}

	l = inlen - (p-s);
	if (0 == l){
		return toutf8(did, "iso-8859-1", s, inlen, bout, outlen);
	}

	if (fullrune(p,l) && chartorune(&r, p) && Runeerror != r){
		return toutf8(did, "utf-8", s, inlen, bout, outlen);
	}
	return toutf8(did, "iso-8859-1", s, inlen, bout, outlen);
}

/*	remaining tcs stuff to support:

	utf1	UTF-1 (ISO 10646 Annex A)
	av	Alternativnyj Variant
	ov	Osnovnoj Variant
	sf1	ISO-646: Finnish/Swedish SF-1 variant
	sf2	ISO-646: Finnish/Swedish SF-2 variant (recommended)
	ujis	EUC-JX: JIS 0208
*	ms-kanji	Microsoft, or Shift-JIS
*	euc-k	Korean EUC: ASCII+KS C 5601 1987
	tis	Thai+ASCII (TIS 620-1986)
	viet1	Vietnamese VSCII-1 (1993)
	viet2	Vietnamese VSCII-2 (1993)
	macrom	Macintosh Standard Roman character set
	next	NEXTSTEP character set
	atari	ATARI-ST character set
	ebcdic	EBCDIC
*/

static Xlate Xtab[] = {
	{"US-ASCII",	0,		-1,	ascii2utf8	},
	{"UTF-8",		0,		-1,	ascii2utf8	},
	{"UTF-7",		0,		-1,	utf72utf8	},
	{"UTF-7-alt",	0,		-1,	utf7alt2utf8	},
	{"UTF-16",	0,		-1,	utf162utf8	},
	{"UTF-16LE",	0,		-1,	utf16LE2utf8	},
	{"UTF-16BE",	0,		-1,	utf16BE2utf8	},
	{"ISO-2022-JP",	0,		0,	encode2utf8	},
	{"ISO-8859-1",	latin1_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-2",	latin2_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-3",	latin3_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-4",	latin4_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-5",	latin5_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-6",	latin6_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-7",	latin7_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-8",	latin8_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-9",	latin9_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-10",	latina_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-11",	latinb_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-13",	latind_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-14",	latine_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-15",	latinf_tab,	0xa0,	latin2utf8	},
	{"ISO-8859-16",	lating_tab,	0xa0,	latin2utf8	},
	{"big5",		big5_tab,	-1,	big52utf8	},
	{"dec-mcs",	dec_mcs,	0xa0,	latin2utf8	},
	{"gb2312",	(Rune*)gb_tab,	0,	gb2utf8		},
//	{"hkscs",		hkscs_tab,	-1,	hkscs2utf8	},
	{"JIS-X-0201",	jis_roman_tab,	0x50,	latin2utf8	},
	{"JIS-X-0208",	jisx_208,		0,	jis2utf8		},
	{"JIS-X-0212",	jisx_212,		0,	jis2utf8		},
	{"KOI8-R",	koi8_r,		0x80,	latin2utf8	},
	{"KOI8-U",	koi8_u,		0x80,	latin2utf8	},
	{"SJIS-X-0208",	jisx_208,		0,	sjis2utf8		},	// name?
	{"SJIS-X-0212",	jisx_212,		0,	sjis2utf8		},	// name?
	{"viscii",		viscii,		0,	latin2utf8	},
	{"ibm424",	ibm424,		0,	latin2utf8	},
	{"ibm437",	ibm437,		0,	latin2utf8	},	// ibm pc
	{"ibm850",	ibm850,		0x80,	latin2utf8	},
//	{"ibm852",	ibm852,		?,	latin2utf8	},
//	{"ibm856",	ibm856,		?,	latin2utf8	},
//	{"ibm862",	ibm862,		?,	latin2utf8	},
	{"ibm866",	ibm866,		0x80,	latin2utf8	},
	{"windows-1250",	windows_1250,	0x80,	latin2utf8	},
	{"windows-1251",	windows_1251,	0x80,	latin2utf8	},
	{"windows-1252",	windows_1252,	0x80,	latin2utf8	},
	{"windows-1253",	windows_1253,	0x80,	latin2utf8	},
	{"windows-1254",	windows_1254,	0x80,	latin2utf8	},
	{"windows-1255",	windows_1255,	0x80,	latin2utf8	},
	{"windows-1256",	windows_1256,	0x80,	latin2utf8	},
	{"windows-1257",	windows_1257,	0x80,	latin2utf8	},
	{"windows-1258",	windows_1258,	0x80,	latin2utf8	},
	{"ucode",		ucode,		0xa0,	latin2utf8	},	// russian ucode
	{"unknown-8bit",	0,		-1,	unknowntoutf8	},	// chris' mailer.
	{"x-mac-cyrillic",	mac_cyrillic,	0x80,	latin2utf8	},
};

#define nxlate	(sizeof(Xtab)/sizeof(Xtab[0]))

boolean toutf8(boolean* did, const char* charset, const char* bin, int inlen, char** bout, int* outlen){
	ul i;
	
	*did = false;
	if (!inlen){
		return true;
	}

	if (0 == charset){
		return unknowntoutf8(did, 0, 0, bin, inlen, bout, outlen);
	}

	for(i=0; i<nxlate; i++){
		if (0 == strcasecmp(Xtab[i].name, charset)){
			return Xtab[i].toutf8(did, Xtab[i].base, Xtab[i].tab, bin, inlen, bout, outlen);
		}
	}
	return false;
}
