// now for dates...  with pedantic adhearance to 2822.
#include <mailhdr.h>

#include <time.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

static int m_tab[]	= {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
static int l_tab[] 	= {31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

static boolean leapyear(int y){
	y += 1900;
	if ((0 == y%100 && y%400) || y%4)
		return false;
	return true;
}

static int mdays(int m, boolean leap){
	int *tab;
	int j;
	int d;

	tab=m_tab;
	if (leap) {
		tab=l_tab;
	}

	for(d=0, j=0; j<m; j++){
		d += tab[j];
	}
	return d;
}

static int ydays(int y){
	int d;

	for (d=0; 70 <= --y; ){
		if (leapyear(y)){
			d+= 1;
		}
		d+=365;
	}
	return d;
}
static int _mktime(struct tm* t){
	boolean leap;
	int s;
	int d;
	leap = leapyear(t->tm_year);

	s=t->tm_sec;
	s+=t->tm_min*60;
	s+=t->tm_hour*3600;
	d=t->tm_mday-1;
	d+=mdays(t->tm_mon, leap);
	d+=ydays(t->tm_year);
	return s + d*86400;
}

static int daysm(int d, int leap, int* m){
	int *tab;
	int j;

	tab=m_tab;
	if (leap) {
		tab=l_tab;
	}

	for(j=0; d>=tab[j]; j++){
		d -= tab[j];
	}
	*m = j;
	return d;
}

void _mktm(struct tm* t, long s){
	int y;

	memset(t,0,sizeof(struct tm));
	
	t->tm_sec = s%60;		s /= 60;
	t->tm_min = s%60;	s /= 60;
	t->tm_hour = s%24;	s /= 24;

	t->tm_wday = (s + 4) % 7;	// epoch is 4= Thu, Jan 01 1970 00:00:00

	for(y=70; s >= 365; y++){
		if (leapyear(y)){
			if (365==y){
				break;
			}
			s--;
		}
		s -= 365;
	}
	t->tm_year = y;
	t->tm_yday = s;
	t->tm_mday = 1 + daysm(s, leapyear(y), &t->tm_mon);
}

static const char* engdow[] = {
	"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", 0
};
static const char* engmon[] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec", 0
};

static int tindex(const char** tab, const char* s, int n){
	int i;
	
	for(i=0; tab[i]; i++){
		if (!strncmp(tab[i], s, n))
			return i;
	}
	return -1;
}

static int scan2d(const char* s){
	int d;
	int c;
	
	c=*s++;
	if (!isdigit(c))
		return -1;
	d=c-'0';

	c=*s++;
	if (!isdigit(c))
		return -1;
	d = d*10 + c-'0';

	if (isdigit(*s))
		return -1;

	return d;
}

static const char* scan12d(const char* s, int* r){
	const char* p;
	int d;
	int c;
	
	p=s;
	c=*s++;
	if (!isdigit(c))
		return p;
	d=c-'0';

	c=*s;
	if (!isdigit(c))
		goto end;
	d = d*10 + c-'0';
	s++;

end:	if (isdigit(*s))
		return p;

	*r=d;
	return s;
}

// scan 2822 timezone/obsolete timezone.
int scan_tz(const char*s){
	const char* p;
	char z;
	int l;
	int h;
	
again:	while(isspace(*s))
		s++;
	p=s;
	if ('-' == *p || '+' == *p)
		p++;
	for(; isalnum(*p); p++)
		;
	l=p-s;

	// old pegasus mail does this.
	if (l==3 && ('-' == *p || '+' == *p) && !strncmp(s,"GMT",3)) {
		s=p;
		goto again;
	}

	// fortunately for us, +0000 is the default.
	// so we only have to parse stuff that's not zero.
	switch(l){
	case 1:
	case 2:	// UT is the same as zero; go ahead and try to turn into a number.
		return 100*strtol(s, 0,10);
	case 3:
		z=*s++;
		if (z == 'G')
			break;

		switch(z){
		default:  goto out;
		case 'E':	h=-4; break;
		case 'C':	h=-5; break;
		case 'M':	h=-6; break;
		case 'P':	h=-7; break;
		}
	
		z=*s++;
		switch(z){
		case 'D':	break;
		case 'S':   h--; break;
		default:	goto out;
		}

		if ('T' != *s)
			goto out;
		return h * 100;
	case 5:
		return strtol(s, 0, 10);
	default:	
		break;
	}
out:	return 0;
}

#define CFWS(s) s = scancomment(0,0,s);

// scan as much of an rfc 2822 date as possible.
// return point at which scanning failed.
// sec == 0 indicates failure. so no "Wed, 31 Dec 1969 24:59:59 -0500 (EST)", please.

const char* scandate(const char* s, void* tp, long* sec, long* offset){
	struct tm t;
	const char* p;
	int l;
	int i;

	memset(&t, 0, sizeof(struct tm));
	if (tp) memset(tp, 0, sizeof(struct tm));
	*sec = 0;
	*offset = 0;

	CFWS(s);
	l = strlen(s);
	if (l>4 && ',' == s[3]){
		i = tindex(engdow, s, 3);
		if (-1 == i){
			return s;
		}
		t.tm_wday = i;
		l-=4;
		s+=4;
	}
	CFWS(s);

	s = scan12d(s,&t.tm_mday);

	CFWS(s);

	i = tindex(engmon, s, 3);
	if (-1 == i){
		return s;
	}
	t.tm_mon = i;
	s+=3;

	CFWS(s);

	p = s;
	i = strtol(s,(char**)&s,10);
	if (s-p<2)
		return s;
	if (i<100){
		// i+=1900;
	} else {
		i-=1900;
	}
	if (i<0){
		i=0;
	}

	t.tm_year = i;

	CFWS(s);

	s = scan12d(s, &t.tm_hour);		// i have examples!
	if (':' != *s)
		return s;
	s++;
	t.tm_min = scan2d(s);
	if (-1 == t.tm_min)
		return s;
	s+=2;
	if (':' == *s){
		t.tm_sec = scan2d(++s);
		if (-1 == t.tm_sec)
			return s-1;
		s+=2;
	}

	CFWS(s);

	*offset = scan_tz(s);
	*sec = _mktime(&t);
	if (*offset) {
		*sec -= 3600*(*offset/100) + 60*(*offset%100);
		_mktm(&t, *sec);
	}

	s = scancomment(0,0,s);		// get that traditional (EST) comment.
	if (tp) memcpy(tp, &t, sizeof(struct tm));
	return s;
}

// print
// don't use strftime(), which uses the current locale.

void mailtime(Format* f, long sec, int offset){
	struct tm t;

	sec += (offset/100)*60*60 + (offset%100)*60;
	_mktm(&t,sec);
	fmtprint(f, "%s, %02d %s ", engdow[t.tm_wday], t.tm_mday, engmon[t.tm_mon]);
	fmtprint(f, "%d %02d:%02d:%02d %+04d", t.tm_year+1900, t.tm_hour, t.tm_min, t.tm_sec, offset);
}

// there is no gettz() function; this will have to do for linux-like systems.
// the goal is to have this be our only connection to glibc timezones.
int localgmtoffset(void){
	static int offset = ~0;

	if (~0 == offset){
		extern long timezone;
		tzset();
		offset = - timezone;
	}
	return offset;
}

char* date2local(const char* date, const char* fmt, ul* conv_sec){
	const char* dflt_fmt = "%Y.%m.%d %H:%M";
	struct tm t;
	long sec;
	long delta;

	if (!fmt){
		fmt = dflt_fmt;
	}

	scandate(date,0,&sec,&delta);		// epoch seconds
	sec += localgmtoffset();		// adjust time to local timezone.
	_mktm(&t,sec);

	*conv_sec=sec;
	return mprint("%/{strftime}", fmt, &t);
}


// our strftime();
static void tmengine(Format * f, const char *fmt, struct tm* t);

static fmt_action a_conv(Format* f, int c, struct tm* tm){
	USED(c);
	fmtprint(f, "%s", engdow[tm->tm_wday % 7]);
	return FMT_verb;
}

static fmt_action b_conv(Format* f, int c, struct tm* tm){
	USED(c);
	fmtprint(f, "%s", engmon[tm->tm_wday % 12]);
	return FMT_verb;
}

static fmt_action d_conv(Format* f, int c, struct tm* tm){
	USED(c);
	fmtprint(f, "%02d", tm->tm_mday);
	return FMT_verb;
}

static fmt_action H_conv(Format* f, int c, struct tm* tm){
	USED(c);
	fmtprint(f, "%02d", tm->tm_hour);
	return FMT_verb;
}

static fmt_action I_conv(Format* f, int c, struct tm* tm){
	USED(c);
	fmtprint(f, "%02d", tm->tm_hour%12);
	return FMT_verb;
}

static fmt_action j_conv(Format* f, int c, struct tm* tm){
	USED(c);
	fmtprint(f, "%02d", tm->tm_yday);
	return FMT_verb;
}

static fmt_action m_conv(Format* f, int c, struct tm* tm){
	USED(c);
	fmtprint(f, "%02d", 1+tm->tm_mon);
	return FMT_verb;
}

static fmt_action M_conv(Format* f, int c, struct tm* tm){
	USED(c);
	fmtprint(f, "%02d", tm->tm_min);
	return FMT_verb;
}

static fmt_action p_conv(Format* f, int c, struct tm* tm){
	static const char* utab[] = {"AM", "PM"};
	static const char* ltab[] = {"AM", "PM"};
	static const char** tab;
	int i;

	tab=ltab;
	i=0;
	if (isupper(c))
		tab = utab;
	if (tm->tm_hour>=12)
		i=1;

	fmtappend(f, tab[i], 2);
	return FMT_verb;
}

static fmt_action S_conv(Format* f, int c, struct tm* tm){
	USED(c);
	fmtprint(f, "%02d", tm->tm_sec);
	return FMT_verb;
}

static fmt_action T_conv(Format* f, int c, struct tm* tm){
	USED(c);
	tmengine(f, "%H:%M:%S", tm);
	return FMT_verb;
}

static fmt_action Y_conv(Format* f, int c, struct tm* tm){
	USED(c);
	fmtprint(f, "%d", tm->tm_year+1900);
	return FMT_verb;
}

static fmt_action z_conv(Format* f, int c, struct tm* tm){
	int o;
	USED(c);

	o = tm->tm_gmtoff / 60;	// seconds
	if (o<0){
		o=-o;
		fmtappend(f,"-",1);
	} else {
		fmtappend(f,"+",1);
	}

	fmtprint(f, "%02d%02d", o/60, o);
	return FMT_verb;
}

static fmt_action literal_conv(Format* f, int c, struct tm* tm){
	USED(tm);
	fmtprint(f, "%c", c);
	return FMT_verb;
}

static fmt_action fmt_badconv(Format *format, int c, struct tm*tm) {
	USED(tm);
	eprint("%{strftime}: bad conversion char '%c'\n", c);
	format->error = 1;
	return FMT_verb;
}

typedef fmt_action (*Ftm)(Format *, int, struct tm*);

static Ftm Ftab[0x100];

static Ftm* insure_inittab(Ftm* tab) {
	int i;

	if (tab[0]) {
		return tab;
	}

	for (i = 0; i <= 0xff; i++) {
		tab[i] = fmt_badconv;
	}

	tab['a'] = a_conv;
//	tab['A'] = A_conv;	
	tab['b'] = b_conv;
//	tab['B'] = B_conv;	
	tab['d'] = d_conv;
	tab['H'] = H_conv;
	tab['I'] = I_conv;
	tab['j'] = j_conv;
	tab['m'] = m_conv;
	tab['M'] = M_conv;
	tab['p'] = p_conv;
	tab['P'] = p_conv;
	tab['S'] = S_conv;
	tab['T'] = T_conv;
	tab['Y'] = Y_conv;
	tab['z'] = z_conv;
	tab['%'] = literal_conv;

	return tab;
}

static void
tmengine(Format * f, const char *fmt, struct tm* t) {
	const unsigned char *s;
	Ftm* tab;
	int c;

	s = (const unsigned char*)fmt;
	tab = insure_inittab(Ftab);

	for (;;) {
		switch ((c = *s++)) {
		case '%':
			//f->flags = f->f1 = f->f2 = 0;
			//f->client_data = 0;
			do c = *s++; while (FMT_flag == (c = tab[c](f, c, t)));
			break;
		case '\0':
			return /*f->buf - f->bufbegin + f->flushed*/;
		default:
			fmtputc(f, c);
			break;
		}
	}
}

fmt_action long_strftime_conv(Format* f, const char* _fmt){
	const char* dfmt = "%a, %d %b %Y %H:%M:%S %z";
	const char* fmt = 0;
	struct tm* t;

	t = va_arg(f->args, struct tm*);
	if (!t){
		return FMT_verb;
	}
	if (_fmt){
		fmt=_fmt;
	}
	if (!fmt){
		fmt = f->client_data;
	}
	if (!fmt){
		fmt = dfmt;
	}

	tmengine(f, fmt, t);
	return FMT_verb;
}
