#include <mailhdr.h>
#include <print.h>
#include <xalloc.h>

#include <unistd.h>
#include <memory.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>

static int execpath(lstHead* path, int c, const char** argv) {
	char fullpath[512];
	lstEntry* e;
	const char* p;
	char* v[128];

	// close(1);
	// dup2(2, 1);
	memcpy(v, argv, c*sizeof(char*));
//	v[0] = fullpath;
	v[c]=0;

	for(p = lstGetFirstData(path, &e); e; p = lstGetNextData(path,&e)){
		snprint(fullpath, 512, "%s/%s", p, argv[0]);
		eprint("exec: c=%d: %{array %s}\n", c, v);
       		execve(fullpath, v, environ);
	}
	eprint("exec: fails %r\n");
 	return errno;
}

static int execute(int c, const char** v){
	lstHead* l;
	const char* s;
	int rv;

	eprint("execute(%d, {%{array %s}})\n", c, v);
	s = getenv("PATH");
	if (!s){
		s = "/bin:/usr/bin";
	}

	l = lstStrstr2List(s, strlen(s), ":");
	eprint("\tpath (%d) %{list}\n", lstCount(l), l);
	rv = execpath(l, c, v);
	
	lstDestroy(l);
	return rv;
}

enum { DELTA = 8 * 1024 };

static void pipeclose(int* fd){
	close(fd[0]);
	close(fd[1]);
}

// parent->pfd[1] -> child->pfd[0] (cmd) child->cfd[1] -> parent->cfd[0]

boolean pipecmd(int c, const char** v, const char* in, int in_len, char** out, ul* out_len){
	int pfd[2];
	int cfd[2];
	int wfd[2];
	int pid;
	boolean status;
	int efd;

	if (-1 == pipe(pfd) || -1 == pipe(cfd)){
		status = false;
		goto out;
	}

	switch((pid=fork())){
	case -1:
		status = false;
		break;
	case 0:
		close(0);
		close(1);
		close(2);
		if (-1 == dup2(pfd[0], 0)){
			exit(7);
		}
		if (-1 == dup2(cfd[1], 1)){
			exit(7);
		}
		close(pfd[1]);
		close(cfd[0]);

		efd = open("/tmp/childout", O_WRONLY|O_CREAT|O_APPEND, 0600);
		//efd = open("/dev/null", O_WRONLY|O_APPEND);
		if (2 != efd){
			dup2(efd,2);
		}

		execute(c, v);
		exit(7);
	default:
		wfd[0] = cfd[0];
		wfd[1] = pfd[1];
		close(pfd[0]);
		close(cfd[1]);
		status = runpipe(pid, wfd, in, in_len, out, out_len);
	}

out:	pipeclose(pfd);
	pipeclose(cfd);

	return status;
}
