#include <mailhdr.h>
#include <xalloc.h>

#include <memory.h>
#include <string.h>

Head* headCreateList(const char* name, lstHead* l){
	Head* H;

	H = xalloc(sizeof(Head));
	memset(H,0,sizeof(Head));
	H->name = xcopy(name);
	H->values = l;

	return H;
}

Head* headCreate(const char* name, const char* value){
	lstHead* l;

	l = lstCreate(&lstFnsString);
	if (value && *value){
		lstAddEntry(l, xcopy(value));
	}
	return headCreateList(name, l);
}

Head* headCopy(const Head* H0){
	Head* H;

	H = xalloc(sizeof(Head));
	memset(H,0,sizeof(Head));
	H->name = xcopy(H0->name);
	H->values = lstCopy(H0->values);

	return H;
}
	
void headDestroy(Head* H){
	xfree(H->name);
	lstDestroy(H->values);
	xfree(H);
}

lstCmp headCmp(const Head* a, const Head* b){
	int i;
	
	i = strcasecmp(a->name, b->name);
	if (i<0){
		return lstLess;
	}
	if (i>0){
		return lstGreater;
	}
	return lstEqual;
}

static const char* repeat_tab[] = {
	"Received",
	0
};

static boolean repeater(const char* name){
	const char** v;

	for(v=repeat_tab; *v; v++){
		if (0 == strcasecmp(name, *v)){
			return true;
		}
	}
	return false;
}

static const struct {
	const char* name;
	const char* sep;
} sep_tab[] = {
	{"References", ""},
	{0,	","},
};

const char* getSep(const char* name){
	int i;

	for(i=0; sep_tab[i].name; i++){
		if (0 == strcasecmp(name, sep_tab[i].name)){
			break;
		}
	}
	return sep_tab[i].sep;
}

void head_conv_int(Format* f, const Head* h, const char* lstsep, const char* elefmt){
	lstEntry* e;
	const char* s;
	char a[30];
	char b[30];

	if (repeater(h->name)){
		snprint(a, 30, "%%s: %s\n", elefmt);
		for(s=lstGetFirstData(h->values, &e); e; s=lstGetNextData(h->values, &e)){
			fmtprint(f, a, h->name, s);
		}
	} else {
		fmtprint(f, "%s: ", h->name);
		snprint(a, 30, "%s%s", getSep(h->name), lstsep);
		snprint(b, 30, "%%/{list %s}\n", elefmt);
		fmtprint(f, b, a, h->values);
	}	
}

fmt_action long_head_conv(Format*f, const char* fmt){
	Head* h;

	h = va_arg(f->args, Head*);
	if (!h){
		fmtprint(f, "(null)");
		return FMT_verb;
	}
	if (f->flags & FMT_altform){
		fmtprint(f, "%{list}", h->values);
		return FMT_verb;
	}
	if (!fmt){
		fmt = "\n\t";
	}

	head_conv_int(f, h, fmt, h->values->f->print);
	return FMT_verb;
}

// BUG() doesn't work.
fmt_action long_headquote_conv(Format*f, const char* fmt){
	Head* h;
	char a[30];

	h = va_arg(f->args, Head*);
	if (!h){
		fmtprint(f, "(null)");
		return FMT_verb;
	}
	if (f->flags & FMT_altform){
		fmtprint(f, "%{list}", h->values);
		return FMT_verb;
	}
	if (!fmt){
		fmt = "\n\t";
	}

	snprint(a,30,"%%#%s", 1+h->values->f->print);
	head_conv_int(f, h, fmt, a);
	return FMT_verb;
}

const lstFns lstFnsHead = {
	(lstFnCopy)headCopy,
	(lstFnDelete)headDestroy,
	(lstFnCompare)headCmp,
	"%{head}"
};

boolean isAddressHdr(const char* n){
	static const char* tab[] = {"Bcc", "Cc", "Delivered-To", "Errors-To", "From", "List-Post", "Mail-Followup-To", "Reply-To", "Return-Path", "Sender", "To", 0};
	const char** p;
	int i;

	for(p=tab; *p; p++){
		i = strcasecmp(*p, n);
		if (0 == i)
			return true;
		if (i > 0)
			break;
	}
	return false;
}

boolean headAddress(const Head* h){
	return isAddressHdr(h->name);
}
		
lstEntry* look_heade(lstHead*l, const char* name, lstEntry* e){
	Head h;

	memset(&h, 0, sizeof(h));
	h.name = (char*)name;

	return lstScanForE(l, &h, e);
}

Head* look_head(lstHead* l, const char* name){
	const lstEntry* e;
	
	e = look_heade(l, name, lstGetFirst(l));
	if (e){
		return e->data;
	}
	return 0;
}

const char* look_head1(lstHead* l, const char* name){
	Head* h;
	lstEntry* e;

	h = look_head(l, name);
	if (h){
		return lstGetFirstData(h->values, &e);
	}
	return 0;
}

void replace_head(lstHead* l, const char* name, lstHead* values){
	Head* h;

	h = look_head(l, name);
	if (!h){
		h = headCreateList(name, 0);
		lstAddEntry(l, h);
	} else {
		lstDestroy(h->values);
	}
	h->values = lstCopy(values);
}

void replace_head1(lstHead* l, const char* name, const char* value){
	lstHead* m;

	m = lstCreate(&lstFnsString);
	lstAddEntry(m, xcopy(value));
	return replace_head(l, name, m);
}

#if 0
lstHead* headOrderdList(const char** order, lstHead* l){
	lstHead* o;
	const Head* h;
	
	o = lstCreate(&lstFnsHead);

	for(; *order; order++){
		h = look_head(l, *order);
		if (h){
			lstAddEntry(o, headCopy(h));
		}
	}

	return o;
}
#endif
