#include <mailhdr.h>
#include <xalloc.h>
#include <rune.h>

#include <ctype.h>
#include <string.h>
#include <memory.h>

// must not accept \n as a space.
//
static inline int spacep(int c){
	if ('\t' == c || ' ' == c){
		return 1;
	}
	return 0;
}

static char*rfc_2047_body(int encoding, SIZE_T l, const char*s, SIZE_T* newl){
	switch(tolower(encoding)){
	default:
		return 0;
	case 'b':
		return smprint(palloc, newl, "%#.*{unbase64}", l, s);
	case 'q':
		return smprint(palloc, newl, "%#.*{unquoted_printable}", l, s);
	}
	return 0;
}

static char* rfc_2047_decode(const char* s, int* N){
	int n;
	char* t;
	char* t0;
	SIZE_T l;
	const char* p;
	const char* end;
	char* utf8;
	SIZE_T utf8_len;
	boolean utf8_alloc;
	char encoding;
	char* decode;
	SIZE_T decode_len;
	boolean decode_alloc;
	char charset[25];

	n = *N;
	t = 0;

	// unroll first check; premature optimization.
	p = s;
	end = strstrn(s, "=?", n);
	if (!end){
		return 0;
	}
	t0 = t = xalloc(UTFmax*n);		// assume all 3-byte decodes.

	for (; end; end = strstrn(p, "=?", n)){
		if (n < (int)sizeof("?=1?b?=?")) {
			goto loose;
		}

		l = end-p;		// copy non-encode stuff.
		memcpy(t, p, l);
		t += l;
		p += l+2;
		n -= l+2;

		end=memchr(p, '?', n);	// find character encoding
		if (!end) break;
		l = end-p;
		if (l>24) goto loose;
		memcpy(charset, p, l);
		charset[l] = 0;
		p += l+1;
		n -= l+1;

		if (n<3) goto loose;		// find transfer encoding
		encoding = *p;
		p+=2;
		n-=2;

		end = strstrn(p, "?=", n);
		if (!end) goto loose;

		l=end-p;
		decode = rfc_2047_body(encoding, l, p, &decode_len);
		if (decode){
			decode_alloc=true;
		} else {
			goto loose;
		}

		if (!toutf8(&utf8_alloc, charset, decode, decode_len, &utf8, (int*)&utf8_len)){
			if (decode_alloc){
				xfree(decode);
			}
			goto loose;
		}

		memcpy(t, utf8, utf8_len);
		t += utf8_len;
		if (decode_alloc){
			xfree(decode);
		}
		if (utf8_alloc){
			xfree(utf8);
		}

		n -= end-p+2;
		p = end +2;
	}
	memcpy(t, p, n);
	t+=n;

	*t = 0;
	*N = t-t0;
	return t0;
loose:
	cxfree(t0);
	return 0;
}

static char* getLine(const char** S, int* N){
	const char* end;
	int endl;
	lstHead* parts;
	char* part;
	const char* s;
	int n;

	n = *N;
	s = *S;

	parts = lstCreate(&lstFnsString);
	
	do {
		end = memchr(s, '\n', n);
		if (end){
			endl = end-s;
		} else {
			endl = n;
			end = s+n;
		}

		part = mprint("%.*{trim}", endl, s);
		if (part){
			lstAddEntry(parts, part);
		}

		if (!end){
			break;
		}

		n-= endl+1;
		s+=endl+1;

	} while (n>0 && spacep(*s));

	part = mprint("%/{list}", " ", parts);
	lstDestroy(parts);

	*S = s;
	*N = n;
	return part;
}

lstHead* crack_header(const char* s, int n){
	lstHead* l;
	char* line;
	char* split;
	char* demangle;
	const char* t;
	int split_n;

	l = lstCreate(&lstFnsHead);

	for(;n>0 && *s == '\n';){
		s++; n--;
	}

	for(;n>0 && *s != '\n';){
		line = getLine(&s, &n);
		if (!line){
			break;
		}
		split = strchr(line, ':');
		if (!split){
			xfree(line);
			continue;
		}
		*split = 0;
		for(split++; spacep(*split); split++) {}
		split_n = strlen(split);
		demangle = rfc_2047_decode(split, &split_n);
		t = demangle ? demangle : split;
		lstAddEntry(l, headCreate(line, t));
		xfree(line);
		cxfree(demangle);
	}

	return l;
}
