#include <base.h>
#include <lst.h>
#include <print.h>

// atom_conv.c
fmt_action long_atom_conv(Format *f, const char* fmt);
fmt_action long_fullatom_conv(Format *f, const char* fmt);
fmt_action long_phrase_conv(Format* f, const char* fmt);
fmt_action long_fullphrase_conv(Format* f, const char* fmt);


// nv.c
typedef struct {
	char* name;
	char* value;
} nv;
nv* nvCreate(void);
nv* nvInst(char* name, char* value);
nv* nvCopy(const nv*);
void nvDestroy(nv*);
const char* nvInTab(const char* name, const nv** v);
fmt_action long_nv_conv(Format*f, const char* fmt);
extern const lstFns lstFnsNv;

// scan.c
const char* scan2(const char*p, const int match, const char*end);
const char* scanstr(char* t, int l, const char*s);
const char* scanmimestr(char* t, int l, const char*s);
const char* scanstrws(char* t, int l, const char*s);
const char* scanmimestrws(char* t, int l, const char*s);
const char* scancomment(char* t, int l, const char*s);
const char* scansqbracket(char* t, int l, const char*s);
const char* scanphrase(char*t, int l, char* c, int cl, const char*s);

// head.c
typedef struct {
	char* name;
	lstHead* values;
} Head;
Head* headCreateList(const char* name, lstHead* l);
Head* headCreate(const char* name, const char* value);
Head* headCopy(const Head*);
void headDestroy(Head*);
lstCmp headCmp(const Head* a, const Head* b);
const char* getSep(const char* name);
fmt_action long_head_conv(Format*f, const char* fmt);
fmt_action long_headquote_conv(Format*f, const char* fmt);
extern const lstFns lstFnsHead;
boolean isAddressHdr(const char*);

lstEntry* look_heade(lstHead*l, const char* name, lstEntry* e);
Head* look_head(lstHead* l, const char* name);
const char* look_head1(lstHead* l, const char* name);
void replace_head(lstHead* l, const char* name, lstHead* values);
void replace_head1(lstHead*, const char*, const char*);

// header.c
lstHead* crack_header(const char* s, int n);

// mimestring.c
typedef enum {
	Dzero,
	Dinline,
	Dattach,
	Dform_data,
	Dsignal,
	Dalert,
	Dicon,
	Drender,
	Dsession,
	Dsigned_receipt_protocol,
	Dsigned_receiped_micalg,
	Daib,
	Dearly_session,
	Dinfinite
} mDispo;

typedef struct {
	char* type;
	char* charset;
	char* name;
	char* boundary;
	lstHead* extra;
	mDispo dispo;
	lstHead* dextra;
} mimeString;

mimeString* mimeStringCreate(void);
mimeString* mimeStringCopy(const mimeString* m1);
void mimeStringDestroy(mimeString*m);
fmt_action long_mimeString_conv(Format* f, const char* fmt);
fmt_action long_mimeStringDispo_conv(Format* f, const char* fmt);
mimeString* str2mimeString(const char* s);
boolean str2mimeStringDispo(mimeString* m, const char* s);
extern const lstFns lstFnsMimeString;
char* mimeStringBoundary(char* buf, int lbuf, const char* purpose);

// mimedoc.c
typedef struct {
	mimeString* m;

	lstHead* head_in;
	lstHead* head_out;

	boolean body_allocated;
	char* body_in;
	int n_in;

	// ↓ all processed results
	lstHead* subdocs;

	boolean out_allocated;
	char* body_out;
	int n_out;

	lstHead* lines;
} mimeDoc;

mimeDoc* mimeDocCreate(void);
mimeDoc* mimeDocCopy(const mimeDoc* A);
void mimeDocDestroy(mimeDoc* D);
void mimeDocSetBody(mimeDoc* D, char* s, int n, boolean allocated);
fmt_action long_mimeDoc_conv(Format* f, const char* fmt);
extern const lstFns lstFnsMimeDoc;

// none; 
typedef struct {
	const char* from;
	const char* to;
	boolean re;			// regular expression
	const char* data;
	lstHead* (*xform)(lstHead*, const Head*, const char* data);
} Xform;
// mailhdr.c; needs to be moved to xform.c
lstHead* xform(lstHead* ls, const Xform* xform);

//mimeparse.c ; ref is mimesection references -- just pass in the whole file.
lstHead* lineifyBody(const char*s, int n);
void decodeDocBody(const Xform* X, mimeDoc* D);
mimeDoc* parseBody(const char* s, int n);
mimeDoc* parseBodyRef(const char* ref, const char* s, int n);
mimeDoc* decodeBodyRef(const Xform*, const char* ref, const char* s, int n);
boolean unmime(const Xform* X, mimeDoc* D);
fmt_action long_un64_conv(Format* f, const char* fmt);
fmt_action long_unquoted_printable_conv(Format* f, const char*fmt);

// maddress.c 

typedef struct {
	char* gecos;
	char* user;
	lstHead* domain;
	lstHead* members;	// group.
	char* comment;
} mAddress;

mAddress* mAddressCreate(const char* gecos, const char* user, const char* domain, const char* comment);
mAddress* mAddressCopy(const mAddress* A);
void mAddressDestroy(mAddress* A);
lstCmp mAddressCmp(const mAddress* A, const mAddress* B);
fmt_action long_mAddress_conv(Format* f, const char* fmt);
fmt_action long_mAddressBasic_conv(Format* f, const char* fmt);
extern const lstFns lstFnsmAddress;

lstHead* str2mAddresses(const char* s, int n, lstHead* l);
mAddress* str2mAddress(const char* s, int n);
mAddress* _str2mAddress(const char** s, int* n);

// msrc.c
typedef enum {
	mSrcZero,
	mSrcReceived,
	mSrcReplyTo,
	mSrcSender,
	mSrcFrom,
	mSrcInfinite
} mSrcType;

typedef struct {
	mSrcType type;
	nv* from;		// sending host	value = address
	nv* by;			// recpient host
	char* via;		// physical path
	lstHead* with;		// link protocol
	char* msgid;
	mAddress* _for;
	long datetime;
	long offset;		// ↓ for printing purposes.
} mSrc;

mSrc* mSrcCreate(void);
mSrc* mSrcInst(mSrcType type, nv* from, nv* by, char* via, lstHead* with, char* msgid, mAddress* _for, long t, long o);
mSrc* mSrcCopy(const mSrc* R);
void mSrcDestroy(mSrc* R);
fmt_action long_mSrc_conv(Format* f, const char* fmt);
extern const lstFns lstFnsMSrc;
mSrc* receivedstr2mSrc(const char* s);

// date.c
void mailtime(Format* f, long sec, int offset);
const char* scandate(const char* s, void */*struct tm* */, long* sec, long* offset);
int scan_tz(const char*);
int localgmtoffset(void);
char* date2local(const char* date, const char* fmt, ul* conv_sec);
fmt_action long_strftime_conv(Format*, const char*);

// dirents.c
lstHead* getDirents(const char* name);

// pipecmd.c
boolean pipecmd(int c, const char** v, const char* in, int, char** out, ul*);

// latin_tab.c
boolean toutf8(boolean* did, const char* charset, const char* bin, int, char** bout, int*);

// run*.c execcmd.c
boolean runpipe(int, int* fd, const char* in, int, char** out, ul*);

// mbox.c 
mimeDoc* mbox2mimeDoc(const char* ref, const char* s, int n);

// ref.c
boolean mimeDocRef(mimeDoc** D, const char* file);

// util.c
const char* strstrn(const char* s, const char* find, int n);
void lowerString(char*);
void mailprintInstall(void);
int safe_write(int, const char*, int);
int safe_write_all(int, const char*, int);
int safe_read(int, char*, int);
