#include <mailhdr.h>

#include <ctype.h>
#include <string.h>

// the reason for the extra work here is because "modern" unix tools
// don't quote the '^From ' lines in the middle of mime email. so...
// we've got to make sure to honor mime encoding regardless of what
// they stick in there.

static void skipMboxJunk(const char** S, int* N){
	const char* p;
	const char* s;
	int n;

	s=*S; n=*N;
	
	p = memchr(s,'\n', n);
	if (!p) return;
	n -= 1+p-s;
	s = p+1;

	while (n && '>' == *s){
		p = memchr(s,'\n', n);
		if (!p) break;
		n -= 1+p-s;
		s = p+1;
	}
	*S=s;
	*N=n;
}

static boolean mboxFrom(const char**S, int* N){
	const char* s;
	int n;

	s=*S; n=*N;
	while (isspace(*s) && n>8){
		s++;
		n--;
	}

	if (n>8 && !strncmp(s, "From ", 5)){
		skipMboxJunk(&s,&n);
		*S=s;
		*N=n;
		return true;
	}
	return false;
}

boolean isMbox(const char*s, int n){
	return mboxFrom(&s,&n);
}

static const char* mboxEnd(const char* mstring, const char* s, int n){
	char boundary[100];
	mimeString* m;
	const char* nl;
	const char* end;

	m = 0;
	if (mstring){
		m = str2mimeString(mstring);
	}

	if (m && m->boundary){
		snprint(boundary, 100, "\n--%s--\n", m->boundary);
		end = strstrn(s, boundary, n);
		if (end){
			n -= end-s;
			s = end;
		}
		mimeStringDestroy(m);
	}

	nl = strstrn(s, "\n\nFrom ", n);
	if (nl){
		return nl+1;
	}
	end=s+n-1;
	if (end[0] == '\n' && end[-1] == '\n'){
		end--;
	}
	return end;
}

static const char fake_header[] = 
"Content-Type: multipart/digest; boundary=mailhdr-internal\n"
"X-Mailhdr: created from mbox\n";

mimeDoc* mbox2mimeDoc(const char* ref, const char* s, int n){
	lstHead* l;
	lstHead* head;
	const char* stype;
	const char* end;
	mimeDoc* D;
	mimeDoc* msg;
	int d;

	l = lstCreate(&lstFnsMimeDoc);

	while (mboxFrom(&s, &n)){
		head = crack_header(s,n);
		stype = look_head1(head, "Content-Type");
		end = mboxEnd(stype, s, n);
		d=end-s;
		D = parseBodyRef(ref, s, d);
		if (D){
			msg = mimeDocCreate();
			msg->m = str2mimeString("message/rfc822");
			str2mimeStringDispo(msg->m, "inline");
			msg->head_in = lstCreate(&lstFnsHead);
			lstAddEntry(msg->head_in, headCreate("Content-Disposition", "inline"));
			msg->subdocs = lstCreate(&lstFnsMimeDoc);
			lstAddEntry(msg->subdocs, D);
			lstAddEntry(l, msg);
			//lstAddEntry(l,D);
		}
		lstDestroy(head);
		n -=d;
		s = end;
	}

	D = mimeDocCreate();
	D->head_in = crack_header(fake_header, sizeof(fake_header));
	//D->head_out = lstCreate(&lstFnsHead);	// bzzzt ... wrongo.
	D->m = str2mimeString("multipart/digest; boundary=mailhdr-internal");
	D->subdocs = l;

	return D;
}

