#include <mailhdr.h>
#include <xalloc.h>

#include <string.h>
#include <memory.h>
#include <ctype.h>

static mimeDoc* decodeBody(const Xform* X, const char* s, int n);

lstHead* lineifyBody(const char*s, int n){
	lstHead* l;
	const char* p;
	int delta;

	l=lstCreate(&lstFnsString);
	
	for(;n; s+=delta, n-=delta){
		p=memchr(s, '\n', n);
		if (!p){
			break;
		}
		delta=p-s;
		lstAddEntry(l, xcopyn(s, delta));
		delta++;
	}
	if (n>0) {
		lstAddEntry(l, xcopyn(s, n));
	}
	return l;
}

mimeDoc* parseMultipart(mimeDoc* D0){
	const char* part[2];
	int i;
	const char* anchor;
	const char* nl;
	int len;
	int blen;
	int dlen;
	mimeDoc* D;
	int n;

	if (!D0->m->boundary){
		return 0;
	}

	D0->subdocs = lstCreate(&lstFnsMimeDoc);

	blen = strlen(D0->m->boundary);
	anchor=D0->body_out;
	i=0; part[0] = part[1] = 0;

	n = D0->n_out;
	for(; n; anchor=nl+1, n-=len+1){
		nl = memchr(anchor, '\n', n);
		if (0 == nl){
			break;		// should we do fancy checks here?
		}
		len = nl-anchor;

		// they add an extra -- at the beginning and an extra -- 
		// for the last boundary. we don't really care.

		switch(len-blen){
		case 2:
		case 4:
			if ('-' != anchor[0] || '-' != anchor[1]){
				continue;
			}
			if (!strncmp(anchor+2, D0->m->boundary, blen)){
				break;
			}
		default:
			continue;
		}

		switch(i){
		case 1:
			part[i++] = anchor-1;
			i = 0;

			dlen = part[1] - part[0] + 1;

			D = parseBody(part[0], dlen);
			if (D){
				lstAddEntry(D0->subdocs, D);
			}
			// fallthrough.
		case 0:
			part[i++] = nl+1;
			break;
		}
	}

	// ignore trailing junk and unmatched boundaries pfft.
	return D0;
}

static struct {
	const char* desc;
	char* (*decode)(const char*, int);
	const char* fmt;
} decode_tab[] = {
//	{"7bit",			0,	"%#.*s"			},
//	{"8bit",			0,	"%#.*s"			},
//	{"binary",		0,	"%#.*s"			},
	{"base64",		0,	"%#.*{un64}"		},
	{"quoted-printable",	0,	"%.*{unquoted_printable}"	},
	{0,			0,	"%#.*s"			},
};
#define last_decode (sizeof(decode_tab)/sizeof(decode_tab[0]))

fmt_action long_un64_conv(Format* f, const char* fmt){
	char* s;
	char* b;
	char* p;
	char* q;
	char* end;
	SIZE_T maxbytes;
	USED(fmt);

	s = va_arg(f->args, char*);
	if (f->flags & FMT_f2set) {
		maxbytes = f->f2;
	} else {
		maxbytes = strlen(s);
	}

	if (0 == maxbytes) {
		return FMT_verb;
	}

	b = xalloc(maxbytes);
	end = s+maxbytes;
	for(p=s, q=b; p<end; p++){
		if (isspace(*p)){
			continue;
		}
		*q++=*p;
	}
	
	fmtprint(f, "%#.*{unbase64}", q-b, b);
	xfree(b);

	return FMT_verb;
}

// use altform for header quoted-printable;
fmt_action long_unquoted_printable_conv(Format* f, const char*fmt){
	char* s;
	SIZE_T maxbytes;
	SIZE_T remain;
	SIZE_T i;
	int c;
	ul l;
	boolean altform;
	USED(fmt);

	s = va_arg(f->args, char*);
	altform = f->flags & FMT_altform ? true : false;
	if (f->flags & FMT_f2set) {
		maxbytes = f->f2;
	} else {
		maxbytes = strlen(s);
	}

	if (0 == maxbytes) {
		return FMT_verb;
	}

	for(i=0; i<maxbytes; ){
		c=s[i];
		switch(c){
		case '=':
			c=s[i+1];
			remain = maxbytes-i;
			if (1<=remain && '\n' == c){
				i+=2;
				break;
			}
			if (3<=remain && isxdigit(c) && isxdigit(s[i+2])){
				strtouln(&l, s+i+1, 2, 16);
				fmtputc(f, (char)l);
				i+=3;
				break;
			}
		case '_':
			if (altform){
				fmtappend(f, " ", 1);
				i++;
				break;
			}
		default:
			fmtappend(f, s+i++, 1);
			break;
		}
	}
	
	return FMT_verb;
}

static char* _decode_body(const char* encoding, SIZE_T* len, const char*body, int n){
	unsigned i;
	char buf[100];

	snprint(buf, 100, "%{trim}", encoding);
	lowerString(buf);

	for(i=0; i<last_decode-1; i++){
		if (0 == strcmp(buf, decode_tab[i].desc)){
			break;
		}
	}

	return smprint(palloc, len, decode_tab[i].fmt, n, body);
}

static boolean multipart_all(const Xform* X, mimeDoc* d){
	mimeDoc* D;
	lstHead* l;
	lstEntry* e;

	l = d->subdocs;
	if (!l){
		return true;
	}
	for(D = lstGetFirstData(l, &e); e; D = lstGetNextData(l, &e)){
		decodeDocBody(X, D);
	}
	return true;
}

static boolean multipart_mix(const Xform* X, mimeDoc* PD){
	return multipart_all(X,PD);
}

static boolean multipart_pick(const Xform* X, mimeDoc* D0, boolean (*check)(const char**, const char*), const char** check_tab){
	mimeDoc* D;
	lstEntry* e;
	Head* h;

	if (!multipart_all(X,D0)){
		return false;
	}

	for(D = lstGetFirstData(D0->subdocs, &e); e; D = lstGetNextData(D0->subdocs, &e)){
		if (check(check_tab, D->m->type)){
			goto found;
		}
	}
	return false;

	/* seek & destroy: doc, mimeString, headers */
found:	mimeDocSetBody(D0, D->body_out, D->n_out, D->out_allocated);
	D->out_allocated = false;

	mimeStringDestroy(D0->m);
	D0->m = mimeStringCopy(D->m);

	for(h=lstGetFirstData(D->head_out, &e); e; h=lstGetNextData(D->head_out, &e)){
		replace_head(D0->head_out, h->name, h->values);
	}
	for(h=lstGetFirstData(D->head_in, &e); e; h=lstGetNextData(D->head_in, &e)){
		replace_head(D0->head_in, h->name, h->values);
	}

	lstDestroy(D0->subdocs);
	D0->subdocs = 0;

	return true;	
}

static boolean in_tab(const char** tab, const char* s){
	if (!s){
		return false;
	}
	for(; *tab; tab++){
		if (0 == strcasecmp(*tab, s)){
			return true;
		}
	}
	return false;
}

static boolean not_in_tab(const char** tab, const char*s){
	if (in_tab(tab,s)){
		return false;
	}
	return true;
}

boolean multipart_alt(const Xform* X, mimeDoc* D){
	static const char* okay[] = {"text/plain", 0};

	return multipart_pick(X,D,in_tab,okay);
}

static boolean multipart_apple(const Xform* X, mimeDoc* D){
	static const char* bad[] = {"application/applefile", 0};

	return multipart_pick(X,D,not_in_tab,bad);
}

static boolean mime_file(const Xform* X, mimeDoc* D){
	Head* h;
	USED(X);

	lstAddEntry(D->head_out, h = headCreate("Content-Type", 0));
	lstAddEntry(h->values, mprint("%#{mimeString}", D->m));
	return true;
}

static boolean mime_rfc822(const Xform* X, mimeDoc* D){
	D->subdocs = lstCreate(&lstFnsMimeDoc);
	lstAddEntry(D->subdocs, decodeBody(X, D->body_out, D->n_out));

	return true;
}

static boolean mime_html(const Xform* X, mimeDoc* D){
	static const char* v[] = {"htmlfmt", 0};
	boolean stat;
	char* s;
	SIZE_T len;
	USED(X);

	stat = pipecmd(1, v, D->body_out, D->n_out, &s, &len);
	if (stat){
		mimeDocSetBody(D, s, len, true);
	}
	return stat;
}

#define P "t=/tmp/msf\"$$\" ; cat > $t ; wvText $t $t\".out\" ; cat $t\".out\" ; rm -f $t $t\".out\""
#define arguments ((sizeof(v)/ sizeof(char**))-1)

boolean mime_msword(const Xform* X, mimeDoc* D){
	static const char* v[] = {"bash", "-c", P, 0};
//	static const char* v[] = {"msword2txt", 0};
	boolean stat;
	char* s;
	SIZE_T len;
	USED(X);

	stat = pipecmd(arguments, v, D->body_out, D->n_out, &s, &len);
	if (stat){
		mimeDocSetBody(D, s, len, true);
	}
	return stat;
}
#undef P
#undef arguments

static boolean mime_nop(const Xform* X, mimeDoc* D){
	USED(D);
	USED(X);

	return true;
}

typedef struct {
	const char* type;
	boolean (*f)(const Xform*, mimeDoc*);
} Mtab;

static const Mtab mtype_tab[] = {
	{	"text/plain",		mime_nop	},
	{	"text/html",		mime_html	},
	{	"application/octet-stream",	mime_file	},
	{	"application/msword",	mime_msword	},
	{	"message/delivery-status",	mime_nop	},
	{	"message/rfc822",		mime_rfc822	},
	{	"multipart/alternative",	multipart_alt	},
	{	"multipart/appledouble",	multipart_apple	},	// fuckers.
	{	"multipart/digest",	multipart_mix	},
	{	"multipart/mixed",	multipart_mix	},
	{	"multipart/parallel",	multipart_mix	},
	{	"multipart/related",	multipart_mix	},
	{	"multipart/report",	multipart_mix	},
	{	0,			mime_nop	},
};

static const Mtab mtype_fallback[] = {
	{	"text",			mime_nop	},
	{	"multipart",		multipart_mix	},
	{	"application",		mime_file	},
	{	"message",		mime_rfc822	},
	{	"image",			mime_file	},
	{	"audio",			mime_file	},
	{	"video",			mime_file	},
	{	0,			mime_nop	},	// shouldn't be.
};

static boolean setMime(mimeDoc* D){
	const char* s;

	if (D->m){
		return true;
	}

	s = look_head1(D->head_in, "Content-Type");
	if (s){
		D->m = str2mimeString(s);
	}
	if (!D->m){
		return false;
	}
	
	s = look_head1(D->head_in, "Content-Disposition");
	if (s){
		str2mimeStringDispo(D->m, s);
	}
	
#if 0
	s = look_head1(D->head_in, "Content-Description");
	if (s){
		mimeStringDesc(m, s);
	}
#endif

	if (D->m){
		return true;
	}
	return false;
}

boolean unmime(const Xform* X, mimeDoc* D){
	const Mtab* T;
	const char* type;
	const char* slash;
	char firstpart[65];

	if (!setMime(D)){
		return true;
	}

	type = D->m->type;
	if (!type || !(slash=strchr(type, '/'))){
		return mime_nop(X, D);
	}

	for (T=mtype_tab; T->type; T++){
		if (0 == strcasecmp(type, T->type)){
			goto done;
		}
	}

	memcpy(firstpart, type, slash-type);
	firstpart[slash-type]=0;

	for (T=mtype_fallback; T->type; T++){
		if (0 == strcasecmp(firstpart, T->type)){
			break;
		}
	}
	
done:	return T->f(X, D);
}

boolean mimeparse(mimeDoc* D){
	mimeDoc* x;

	if (!setMime(D)){
		return true;
	}

	if (strncmp(D->m->type, "multipart/", 10)){
		return true;
	}
	x = parseMultipart(D);
	return x ? true : false;
}

static boolean isText(const mimeString* m){
	if (!m || !m->type || !strncmp(m->type, "text/", 5))
		return true;
	return false;
}

boolean uncharset(mimeDoc* D){
	boolean new_body;
	boolean rv;
	char* s;
	int n;

	rv = toutf8(&new_body, D->m->charset, D->body_out, D->n_out, &s, &n);
	if (new_body){
		mimeDocSetBody(D, s, n, true);
	}
	// this whole section is questionable.
	if (rv && isText(D->m)){
		if (D->m->charset){
			xfree(D->m->charset);
			D->m->charset = xcopy("utf-8");
		} else {
			D->m->charset = xcopy("unknown-8bit");
		}
	}
	return rv;
}

static const char* fail_charset = "charset conversion failed.\ncannot convert %s\n";
static const char* fail_unmime = "document conversion failed.\ncannot convert %s\n";

static boolean findBody(char** S, int* N){
	char* s;
	char* nl;
	int n;
	int delta;

	s=*S;
	n=*N;

	for(; n; ){
		nl = memchr(s, '\n', n);
		if (!nl){
			return false;
		}
		delta=nl-s+1;
		s += delta;
		n-=delta;
		if (n && *s == '\n'){
			s++; n--;
			break;
		}
	}

	if (n){
		*S=s;
		*N=n;
		return true;
	}
	return false;
}

mimeDoc* parseBody(const char* s, int n){
	mimeDoc* D;
	char* m;
	SIZE_T nblen;
	char* nbody;
	const char* type;

	D = mimeDocCreate();
	D->body_allocated = false;
	D->body_in = (char*)s;
	D->n_in = n;

	D->head_in = crack_header(s, n);
	if (!D->head_in){
		mimeDocDestroy(D);
		return 0;
	}

	if (!findBody((char**)&s, &n)){
		return D;
	}

	type = look_head1(D->head_in, "Content-Transfer-Encoding");
	if (!type){
		mimeDocSetBody(D, (char*)s, n, false);
	} else {
		// we can't reset Content-Transfer-Encoding; no place to put it.
		nbody = _decode_body(type, &nblen, s, n);
		mimeDocSetBody(D, nbody, nblen, true);
	}

	if (!mimeparse(D)){
		m = mprint(fail_unmime, D->m->type);
		mimeDocSetBody(D, m, strlen(m), true);
	}

	return D;
}

mimeDoc* parseBodyRef(const char* ref, const char* s, int n){
	mimeDoc* D;

	D = parseBody(s, n);
	if (D) mimeDocRef(&D, ref);
	return D;
}

static const char* sfnet = "-------------------------------------------------------";

static boolean allWS(const char*s){
	for (;*s; s++){
		if (!isspace(*s)){
			return false;
		}
	}
	return true;
}

static void removeAds(lstHead* body){
	lstEntry* e;
	const char* s;
	int i;
	
	for(;;){
		while ((s = lstGetLastData(body,&e))){
			if (!allWS(s)){
				break;
			}
			lstDeleteLast(body);
		};

		i=0;
		for(s=lstGetLastData(body,&e); i<15 && e; s=lstGetPrevData(body,&e)){
			i++;
			if (0 == strcmp(s, sfnet)){
				goto delete_ad;
			}
		}
		return;

delete_ad:	for(; e; lstGetNextData(body,&e)){
			lstDeleteEntry(body, &e);
		}
	}
}

void decodeDocBody(const Xform* X, mimeDoc* D){
	char* m;

	if (!D){
		return;
	}
	D->head_out = xform(D->head_in, X);
	if (!D->m){
		D->m = str2mimeString("text/plain");
	}

	if (lstZero(D->subdocs)){
		if (!uncharset(D)){
			m = mprint(fail_charset, D->m->charset);
			mimeDocSetBody(D, m, strlen(m), true);
		} else if (!unmime(X, D)) {
			m = mprint(fail_unmime, D->m->type);
			mimeDocSetBody(D, m, strlen(m), true);
		}

		if (!D->lines){
			D->lines = lineifyBody(D->body_out, D->n_out);
			removeAds(D->lines);
		}
	} else {
		 if (!unmime(X, D)) {
			m = mprint(fail_unmime, D->m->type);
			mimeDocSetBody(D, m, strlen(m), true);
		}
	}
}

static mimeDoc* decodeBody(const Xform* X, const char* s, int n){
	mimeDoc* D;

	D = parseBody(s, n);
	decodeDocBody(X,D);

	return D;
}

mimeDoc* decodeBodyRef(const Xform* X, const char* ref, const char* s, int n){
	mimeDoc* D;

	D = parseBody(s, n);
	if (D) mimeDocRef(&D, ref);
	if (D) decodeDocBody(X,D);

	return D;
}
