#include <mailhdr.h>
#include <xalloc.h>

#include <memory.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <unistd.h>

mimeString* mimeStringCreate(void){
	mimeString* m;

	m=xalloc(sizeof(mimeString));
	memset(m,0,sizeof(mimeString));
	return m;
}

char* mimeStringBoundary(char* buf, int l, const char* purpose){
	if (!purpose){
		purpose = "x";
	}

	if (l){
		snprint(buf, l, "mailhdr.%s.%09x.%06x", purpose, time(0), getpid());
		return 0;
	}
	return mprint("mailhdr.%s.%09x.%06x", purpose, time(0), getpid());
}

mimeString* mimeStringCopy(const mimeString* m1){
	mimeString* m2;

	m2 = mimeStringCreate();
	m2->type = cxcopy(m1->type);
	m2->charset = cxcopy(m1->charset);
	m2->name = cxcopy(m1->name);
	m2->boundary = cxcopy(m1->boundary);
	if (m1->extra) {
		m2->extra = lstCopy(m1->extra);
	}
	m2->dispo = m1->dispo;
	if (m1->dextra){
		m2->dextra = lstCopy(m1->dextra);
	}
	return m2;
}

void mimeStringDestroy(mimeString*m){
	cxfree(m->type);
	cxfree(m->charset);
	cxfree(m->name);
	cxfree(m->boundary);
	if (m->extra){
		lstDestroy(m->extra);
	}
	if (m->dextra){
		lstDestroy(m->dextra);
	}
	xfree(m);
}

fmt_action long_mimeString_conv(Format* f, const char* fmt){
	mimeString* m;
	boolean semi;
	lstEntry* e;
	const nv* s;
	static const char* dflt_fmt = ";%\n\t";

	m = va_arg(f->args, mimeString*);
	if (!m || (!m->type && !m->charset && !m->name && !m->boundary && !m->extra)){
		return FMT_verb;
	}
	if (!fmt){
		fmt = dflt_fmt;
	}

	semi=false;
	if (!(f->flags & FMT_altform)){
		fmtprint(f, "Content-Type: ");
	}

	if (m->type) {
		fmtprint(f, "%s", m->type);
		semi=true;
	}

	if (m->charset) {
		if (semi) fmtprint(f, "; "); semi=true;
		fmtprint(f, "charset=%s", m->charset);
	}

	if (m->name) {
		if (semi) fmtprint(f, fmt); semi=true;
		fmtprint(f, "name=%{atom}", m->name);
	}

	if (m->boundary) {
		if (semi) fmtprint(f, fmt); semi=true;
		fmtprint(f, "boundary=%{atom}", m->boundary);
	}

	if (m->extra){
		for(s = lstGetFirstData(m->extra,&e); e; s = lstGetNextData(m->extra,&e)){
			if (semi) fmtprint(f, fmt); semi=true;
			fmtprint(f, "%{nv}", s);
		}
	}

	return FMT_verb;
}

static const char* dispo_tab[Dinfinite+1] = {
	"", "inline", "attachment", "form-data", "signal", "alert", "icon", "render",
	"session", "signed-receipt-protocol", "signed-receipted-micalg", "aib",
	"early-session", "(BUG Dinfinite)", 
};

static const char* dispop_tab[Dinfinite+1] = {
	"inline", "inline", "attachment", "form-data", "signal", "alert", "icon", "render",
	"session", "signed-receipt-protocol", "signed-receipted-micalg", "aib",
	"early-session", "(BUG Dinfinite)", 
};

static mDispo str2mDispo(const char* name){
	mDispo i;

	for(i=Dzero; i<Dinfinite; i++)
		if (0 == strcmp(name, dispo_tab[i]))
			break;
	return i;
}

fmt_action long_mimeStringDispo_conv(Format* f, const char* fmt){
	mimeString* m;
	boolean semi;
	lstHead* l;
	lstEntry* e;
	const nv* s;
	static const char* dflt_fmt = ";%\n\t";

	m = va_arg(f->args, mimeString*);
	if (!m || (!m->dispo && !m->name && !m->dextra)){
		return FMT_verb;
	}
	if (!fmt){
		fmt = dflt_fmt;
	}

	semi=false;
	if (!(f->flags & FMT_altform)){
		fmtprint(f, "Content-Disposition: ");
	}

	fmtprint(f, "%s", dispop_tab[m->dispo]);
	semi=true;

	if (m->name) {
		if (semi) fmtprint(f, fmt); semi=true;
		fmtprint(f, "filename=%{atom}", m->name);
	}

	l = m->dextra;
	if (l){
		for(s = lstGetFirstData(l,&e); e; s = lstGetNextData(l,&e)){
			if (semi) fmtprint(f, fmt); semi=true;
			fmtprint(f, "%{nv}", s);
		}
	}

	return FMT_verb;
}

const lstFns lstFnsMimeString = {
	(lstFnCopy)mimeStringCopy,
	(lstFnDelete)mimeStringDestroy,
	0,
	"%{mimeString}"
};

// BUG() / isn't a special character via 2822 
static const char* scanType(const char* s, char** type){
	char t[100];
	char c[100];
	int l;
	int i;
	char* p;

	l = sizeof(t);
	p = t;

	do {
		s = scanstrws(p, l, s);
		s = scancomment(p, sizeof(c), s);
		i = strlen(p);
		p+=i;
		l-=i;
		if ('/' != *s){
			break;
		}
		s++;
		l--;
		*p++ = '/';	
	} while (l);
	
	*p = 0;
	*type = xcopy(t);
	lowerString(*type);
	return s;
}

mimeString* str2mimeString(const char* s){
	const char* p;
	mimeString* m;
	char label[100];
	char value[100];

	m = mimeStringCreate();
	p = scanType(s, &m->type);

	while (';' == *p){
		p++;
		p = scanmimestrws(label, sizeof label, p);
		p = scancomment(0,0,p);
		if (*p != '='){
			continue;
		}

		p++;

		p = scanmimestrws(value, sizeof value, p);
		p = scancomment(0,0,p);

		if (0==strcasecmp(label, "charset")){
			m->charset = xcopy(value);
		} else if (0 == strcasecmp(label, "filename") || 0 == strcasecmp(label, "name")){
			if (!m->name) m->name = xcopy(value);
		} else if (0 == strcasecmp(label, "boundary")){
			m->boundary = xcopy(value);
		} else {
			if (!m->extra){
				m->extra = lstCreate(&lstFnsNv);
			}
			lstAddEntry(m->extra, nvInst(xcopy(label), xcopy(value)));
		}
	}

	lowerString(m->charset);
	return m;
}

static const char* scanDispo(const char*s, mDispo* d){
	char t[100];

	s = scanstrws(t,sizeof t, s);
	s = scancomment(0, 0, s);
	lowerString(t);
	*d = str2mDispo(t);
	
	return s;
}

/* additional fields that could be handled if they had any meaning for us:

creation-date		date when content was created       		[RFC2183]
modification-date		date when content was last modified  	[RFC2183]
read-date                 	date when content was last read          	[RFC2183]
name                      	original field name in form              	[RFC2388]
voice                     	type or use of audio content              	[RFC2421]
handling                  	wheter or not processing is required    	[RFC3204]

*/

boolean str2mimeStringDispo(mimeString* m, const char* s){
	const char* p;
	char label[100];
	char value[100];

	p = scanDispo(s, &m->dispo);

	while (';' == *p){
		p++;
		p = scanmimestrws(label, sizeof label, p);
		p = scancomment(0,0,p);
		if (*p != '='){
			continue;
		}

		p++;

		p = scanmimestrws(value, sizeof value, p);
		p = scancomment(0,0,p);

		if (0 == strcasecmp(label, "filename") || 0 == strcasecmp(label, "name")){
			if (!m->name) m->name = xcopy(value);
		} else {
			if (!m->extra){
				m->dextra = lstCreate(&lstFnsNv);
			}
			lstAddEntry(m->dextra, nvInst(xcopy(label), xcopy(value)));
		}
	}

	lowerString(m->charset);
	return true;
}

