#include <mailhdr.h>
#include <xalloc.h>
#include <sys/types.h>
#include <slurp.h>
#include <pregex.h>

#include <memory.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>

/* 	valid references: (should we change the mime part sep to ditch the /re/?)
	.1.2			duh.
	.1
	/re/			match regular expression.	
	address , address		range matching.
	,			the whole kit and kaboodle

    	valid expressions:
	address	field			
		g/From/quanstro@quanstro.net/
	,	g/content-id/<xyz@cmu.edu>/
	.1	g/subject/runetype\.c/
	.1	g/body/runetype\.c/
		v:content-type:text/plain:

	
*/

typedef enum {
	rZero,
	rAddr,		
	rRE,
	rInfinte
} mRefType;

typedef struct {
	mRefType type;
	lstHead* addr;
	char sep;
	char* re;
	Regex* R;
} mRef;

mRef* mRefCreate(void){
	mRef* r;

	r=xalloc(sizeof(mRef));
	memset(r,0,sizeof(mRef));
	return r;
}

mRef* mRefInst(lstHead* addr, char sep, char* re){
	mRef* r;

	r = mRefCreate();
	r->addr = addr;
	r->re = re;
	if (addr){
		r->type = rAddr;
	} else if (re){
		r->sep = sep;
		r->type = rRE;
		r->R = regexCreate();
		regexCompile(r->re, r->R);
	}
	return r;
}

mRef* mRefCopy(const mRef* r){
	return mRefInst(r->addr ? lstCopy(r->addr) : 0, r->sep, cxcopy(r->re));
}

void mRefDestroy(mRef* r){
	switch(r->type){
	case rRE:
		xfree(r->re);
		if (r->R){
			regexDestroy(r->R);
		}
		break;
	case rAddr:
		lstDestroy(r->addr);
		break;
	default:
		break;
	}
	xfree(r);
}

fmt_action long_mRef_conv(Format* f, const char* fmt){
	mRef* r;
	USED(fmt);

	r = va_arg(f->args, mRef*);
	if (!r){
		return FMT_verb;
	}
	
	switch(r->type){
	case rZero:
		break;
	case rRE:
		fmtprint(f, "%c%s%c", r->sep, r->re, r->sep);
		break;
	case rAddr:
		fmtprint(f, "%{list}", r->addr);
		break;
	default:
		assert(0 == "bad type in long_mRef_conv");
		break;
	}

	return FMT_verb;
}

const char* findsep(const char*p, const int match){
	int c;
	for(p++; (c=*p); p++){
		if ('\\' == c){
			p++;
		} else if (match == c){
			break;
		}
	}
	return p;
}


mRef* str2mRef(const char**S){
	const char* p;
	const char* s;
	const char* q;
	char sep;
	lstHead* l;
	int c;

	s = *S;

	while (isspace(*s)){
		s++;
	}

	c = *s;
	if (0 == c){
		return 0;
	}

	if ('.' == c){
		p=s;
		while ('.' == *s++){
			while (isdigit(*++s)){
			}
		}

		*S = --s;
		c = s-p;
		if (c<=1){
			return 0;
		}

		l = lstCreate(&lstFnsString);
		lstAddEntry(l, xcopyn(p, c));
		return mRefInst(l,0,0);
	}

	// isalpha() ?
	if (islower(*s)){
		return 0;
	}

	sep=*s;
	p = findsep(s, sep);
	if (!p){
		return 0;
	}

	q = p+1;
	while (isspace(*q)){
		q++;
	}
	*S = q;

	return mRefInst(0, sep, xcopyn(s+1, p-s-1));
}

typedef struct {
	boolean neg;	// address
	mRef* a;
	mRef* b;
	char rtype;	// name
	char sep;
	nv* field;
	Regex* name;
	Regex* value;	// currently no negitive matching for this.
} mExp;

#define New(var, type) var = xalloc(sizeof(type)); memset(var, 0, sizeof(type))

mExp* mExpCreate(void){
	mExp* E;

	New(E, mExp);
	return E;
}

boolean mExpSetField(mExp* E, const char* name, int namel, const char* value, int valuel){
	char* n;
	char* v;
	boolean b;

	b = true;
	n=0;
	if (namel){
		n = xcopyn(name, namel);
		E->name = regexCreate();
		b &= regexCompile(n, E->name);
	}

	v = 0;
	if (valuel){
		v = xcopyn(value, valuel);
		E->value = regexCreate();
		b &= regexCompile(v, E->value);
	}
	E->field = nvInst(n, v);
	return b;
}

mExp* mExpCopy(const mExp* E){
	mExp* n;
	boolean b;
	const char* x;
	const char* y;

	n = mExpCreate();
	n->neg = E->neg;
	if (E->a){
		n->a = mRefCopy(E->a);
	}
	if (E->b){
		n->b = mRefCopy(E->b);
	}
	n->rtype = E->rtype;
	if (E->field){
		x = E->field->name;
		y = E->field->value;
		b = mExpSetField(n, x, x ? strlen(x) : 0, y, y ? strlen(y) : 0);
		assert(b);
	}
	return n;
}

void mExpDestroy(mExp* E){
	if (E->a){
		mRefDestroy(E->a);
	}
	if (E->b){
		mRefDestroy(E->b);
	}
	if (E->field){
		nvDestroy(E->field);
	}
	if (E->name){
		regexDestroy(E->name);
	}
	if (E->value){
		regexDestroy(E->value);
	}

	xfree(E);
}

fmt_action long_mExp_conv(Format* f, const char* fmt){
	mExp* e;
	boolean sp;
	USED(fmt);

	e = va_arg(f->args, mExp*);
	if (!e){
		return FMT_verb;
	}
	
	sp = false;
	if (e->neg){
		fmtappend(f,"!",1);
	}
	if (e->a && e->b){
		fmtprint(f, "%{mRef},%{mRef}", e->a, e->b);
		sp=true;
	} else if (e->a){
		fmtprint(f, "%{mRef}", e->a);
		sp=true;
	}

	if (sp){
		fmtappend(f, " ", 1);
		sp=false;
	}

	if (e->rtype){
		fmtputc(f,e->rtype);
	}
	
	if (e->field){
		fmtprint(f, "%c%s%c%s%c", e->sep, e->field->name, e->sep, e->field->value, e->sep);
	}

	return FMT_verb;
}

mExp* str2mExp(const char** S){
	mExp* E;
	const char* s;
	const char* s1;
	const char* s2;
	char c;
	boolean b;

	s = *S;
	if (!s){
		return 0;
	}
	
	E = mExpCreate();

	while (isspace(*s)){
		s++;
	}

	if ('!' == *s){
		s++;
		E->neg = true;
	} 

	E->a = str2mRef(&s);

	if (',' == *s){
		s++;
		E->b = str2mRef(&s);
	}

	c = *s;
	switch(c){
	case 'g':
	case 'v':
		E->rtype = c;
		break;
	case 0:
		*S = s;
		return E;
	default:
		*S = s;
		mExpDestroy(E);
		return 0;
	}

	s++;
	while (isspace(*s)){
		s++;
	}

	E->sep = *s++;
	if (!E->sep){
		return E;
	}

	s1 = findsep(s, E->sep);
	s2 = 0;
	if (s1){
		s2 = findsep(s1+1, E->sep);
	}

	if (s2){
		b = mExpSetField(E, s, s1-s, s1+1, s2-s1-1);
		s = s2 + 1;
	} else {
		b = mExpSetField(E, s, s1-s-1, 0, 0);
		s = s1 + 1;
	}

	*S=s;
	if (!b){
		mExpDestroy(E);
		return 0;
	}
	return E;
}

static int _mimeCmp(const char* a, const char* b){
	int i;
	int j;
	int d;
	while (*a == *b && *b == '.'){
		a++; b++;
		i = strtoul(a,(char**)&a,10);
		j = strtoul(b,(char**)&b,10);

		d=j-i;
		if (d){
			return d;
		}
	}
	return (*a == *b) ? 0 : -1;
}

static int mimeCmp(const mRef* a, const char* b){
	lstEntry*e;

	switch(a->type){
	case rAddr:
		return _mimeCmp(lstGetFirstData(a->addr,&e), b);
	case rRE:
	default:
		break;
	}
	return -1;
}

static boolean field(const mimeDoc* d, const mExp* E){
	boolean b;
	Head* h;
	lstEntry* e;
	const char* s;
	lstEntry* f;
	boolean neg;

	neg = 'v' == E->rtype;
	for(h=lstGetFirstData(d->head_in, &e); e; h=lstGetNextData(d->head_in, &e)){
		b = regexMatch(E->name, h->name);
		if (neg) {
			b=!b;
		}
		if (!b){
			continue;
		}

		for(s=lstGetFirstData(h->values,&f); f; s=lstGetNextData(h->values,&f)){
			b = regexMatch(E->value, s);
			//if (valueneg) {
			//	b=!b;
			//}
			if (b) {
eprint("matched value %s\n", s);
				return true;
			}
		}
	}
	return false;
}

static boolean check(mimeDoc* doc, const mExp* E, const char* address){
	int b;
	int i;

	b = 1;
	if (E->a){
		i = mimeCmp(E->a, address);
		if (E->b){
			b &= i>=0;
		} else {
			b &= i==0;
		}
	}
	
	if (E->b){
		i = mimeCmp(E->b, address);
		b &= i<=0;
	}

	if (E->neg){
		b = !b;
	}

	if (!b || !E->field){
		// eprint("check %s → %{boolean}\n", address, b);
		return b ? true : false;
	}

	return field(doc, E);
}

static boolean walk_field(mimeDoc** dp, const mExp* E, const char* address){
	mimeDoc* d;
	mimeDoc* doc;
	char a[128];
	int n;
	boolean status;
	lstHead* l;

	doc = *dp;
	if (check(doc,E, address)){
		return true;
	}

	if (lstZero(doc->subdocs)){
		return false;
	}

	n=0;
	l = lstCreate(&lstFnsMimeDoc);
	while((d=lstRemoveFirst(doc->subdocs))){
		snprint(a, 128, "%s.%02d", address, n++);
		status = walk_field(&d, E, a);
		if (!status){
			mimeDocDestroy(d);
		} else {
			lstAddEntry(l,d);
		}
	}
	lstDestroy(doc->subdocs);
	doc->subdocs = l;

	if (!lstCount(doc->subdocs)){
		return false;
	}
	if (1 == lstCount(doc->subdocs)){
		d = lstRemoveFirst(doc->subdocs);
		mimeDocDestroy(doc);
		*dp = d;
	}

	return true;
}

boolean applyRestrictions(mimeDoc** D, const char*field){
	mExp* E;
	boolean st;

	E = str2mExp(&field);
	if (!E){
		return false;
	}
	st=false;
	if (field){
		st=walk_field(D, E, "");
	}
	mExpDestroy(E);
	return st;
}

boolean mimeDocRef(mimeDoc** D, const char* file){
	const char* p;
	boolean r;

	p = strchr(file, '!');
	if (!p){
		return true;
	}

	r = applyRestrictions(D,p+1);
	if (!r) {
		mimeDocDestroy(*D);
		*D=0;
	}
	return r;
}
