/*
 *	use pthreads() to excess.
*/

#include <mailhdr.h>
#include <print.h>
#include <xalloc.h>

#include <unistd.h>
#include <memory.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pthread.h>

typedef struct {
	int pid;
	int *fd;
	const char* in;
	int in_len;
	char** out;
	SIZE_T* out_len;
} Args;

static const char* pthread_rv[] = { 0, "1" };

enum { DELTA = 8 * 1024 };

static void _pthread_exit(int unix_rv){
	int index=1;

	if (unix_rv){
		index=0;
	}
	pthread_exit((void*)pthread_rv[index]);
}

static void* writer(void* varg){
	const Args* A;
	int in_len;
	const char* in;
	int bytes;
	int rv;

	A = (Args*)varg;
	in = A->in;
	in_len = A->in_len;
	rv = 0;

	while (in_len) {
		bytes = safe_write(A->fd[1], in, in_len);
		switch(bytes){
		case -1:
			rv = -1;
		case 0:
			goto done;
		default:
			in_len-=bytes;
			in+=bytes;
		}
	};

done:	close(A->fd[1]);
	_pthread_exit(rv);
	return 0;
}

static void* reader(void* varg){
	Args* A;
	int bytes;
	int rv;
	char* s;
	int salloc;
	int si;

	A = (Args*)varg;
	rv = 0;

	salloc = A->in_len + DELTA;
	si = 0;
	s = xalloc(salloc);

	do {
		if (salloc-si < DELTA){
			s = xrealloc(s, salloc+=DELTA);
		}
	
		bytes = safe_read(A->fd[0], s + si, salloc-si);
		switch(bytes){
		case -1:
			eprint("%r\n");
			rv=-1;
		case 0:
			break;
		default:
			si+=bytes;
			break;
		}
	} while(bytes>0);
	
	if (-1 != rv){
		*A->out = s;
		*A->out_len = si;
	} else {
		xfree(s);
		s=0;
	}
	pthread_exit(s);
	return 0;
}

boolean runpipe(int pid, int* fd, const char* in, int in_len, char** out, SIZE_T* out_len){
	char* statw;
	char* statr;
	pthread_t thread[2];
	Args A;

	A.pid = pid;
	A.fd = fd;
	A.in = in;
	A.in_len = in_len;
	A.out = out;
	A.out_len = out_len;

	if (pthread_create(&thread[0], 0, writer, &A)){
		return false;
	}
	if (pthread_create(&thread[1], 0, reader, &A)){
		pthread_cancel(thread[0]);
		return false;
	}

	pthread_join(thread[0], (void**)&statw);
	pthread_join(thread[1], (void**)&statr);

	return statw && statr;
}
