#include <mailhdr.h>

#include <string.h>
#include <ctype.h>
#include <stdlib.h>

static const char* _scancommentz(const char* s, const char* m);

const char* scan2(const char*p, const int match, const char*end){
	int c;
	for(p++; p<end; p++){
		c=*p;
		if ('\\' == c){
			p++;
		} else if (match == c){
			break;
		}
	}
	return p;
}

enum {
	Lin	= -1,
	Lout	= -3,
	Lspecial	= 1,
	Ldot	= 2,			// automagic phrase pasting. e.g. .
	Lfws	= 3,			// folding white space
	Lesc	= 4,			// e.g. backslash.
	Lquote	= 5,
};

static const char mapu[0x100] = {		// unquoted
	[0]	= Lspecial,
	['\t']	= Lfws,
	['\n']	= Lfws,
	['\r']	= Lfws,
	[' '] 	= Lfws,
	['(']	= Lin,
	[')']	= Lout,
	['<']	= Lspecial,
	['>']	= Lspecial,
	['@']	= Lspecial,
	[',']	= Lspecial,
	[';']	= Lspecial,
	[':']	= Lspecial,
	['\\']	= Lesc,
	['"']	= Lquote,
	['[']	= Lspecial,
	[']']	= Lspecial,
 	['.']	= Ldot,			// goofy special case.
};

static const char mapm[0x100] = {		// mime
	[0]	= Lspecial,
	['\t']	= Lfws,
	['\n']	= Lfws,
	['\r']	= Lfws,
	[' '] 	= Lfws,
	['(']	= Lin,
	[')']	= Lout,
	['<']	= Lspecial,
	['>']	= Lspecial,
	['@']	= Lspecial,
	[',']	= Lspecial,
	[';']	= Lspecial,
	[':']	= Lspecial,
	['\\']	= Lesc,
	['"']	= Lquote,
//	['/']	= 1,			// uncomment?
	['[']	= Lspecial,
	[']']	= Lspecial,
 	['.']	= Ldot,			// goofy special case.
	['?']	= Lspecial,
	['=']	= Lspecial,
};

static const char mapq[0x100] = {
	[0]	= Lspecial,
	['\\']	= Lesc,
	['\n']	= Lfws,
	['\r']	= Lfws,
	['"']	= Lquote,
};

static const char mapc[0x100] = {
	[0]	= Lspecial,
	['\\']	= Lesc,
	['(']	= Lin,
	[')']	= Lout,
};

static const char mapsq[0x100] = {
	[0]	= Lspecial,
	['\\']	= Lesc,
	['[']	= Lin,
	[']']	= Lout,
};

// '.' is very special. as per rfc 2822. 3.2.4. "dot-atom".
// this function should be called _scandotatom()
// l != strlen(s); l = length(t);

static const char*
_scanstr(const char* map, char* t, int l, const char *s) {
	const char* m;
	const char* reject;
	const char* rs;
	int c;

	if ('.' == *s) {
		goto done;
	}
	reject=0;
	rs=s;
	m = map;
	while (*s){
		while(l-- && 0 == (c=m[*(uchar*)s]))
			*t++ = *s++;
		if (0 == l)
			break;
		if (Lesc == c) {
			s++; l--;
			if (!l) {
				break;
			}
			if (*s){
				*t++ = *s++;
			}
			continue;
		} else if (Lin == c){
			s = _scancommentz(s+1, mapc);
			continue;
		}
		if (m == map){
			if (Lquote == c){
				m = mapq;
				s++;
				continue;
			} else if (Ldot == c){
				reject = t; rs = s;
				*t++ = *s++;
				s = scancomment(0,0,s);
				continue;
			}
		} else {
			if (Lquote == c) {
				m = map;
				s++; 		// eat closing quote.
				//continue; 	// not allowed by rfc.
			} else if (Lfws == c){
				while (isspace(*s))
					s++;
				if (l){
					*t++ = ' ';
					l--;
				}
				continue;
			}
		}
		break;
	}

	if (m != map){
		// no closing quote found
	} else {
		if (t == reject){
			s = rs;
			t--;
		}
	}
done:	*t = 0;
	return s;
}

/* royal asspain alert: we just have to scan until we've scanned one string too many.
    then reject that string and return the rest.

    probablly should be handled as a list? 
*/

static const char*
_scanphrase(const char* m, char* buf, int l, char* comment, int cl, const char* s){
	const char* p;
	char* bend;
	int lp;
	uchar c;

	bend = buf;
	for(p=s; l>0;s=p){
		p = scanstrws(buf, l, s);
		p = scancomment(comment, cl, p);

		c=*(uchar*)p;
		if (p == s || '@' == c || 0 == c){	/* reject */
			*bend = 0;
			return s;
		}

		switch(m[c]){
		case Ldot:
			*buf++ = *p++;
			l--;
			continue;
		case Lfws:
		case Lquote:
		case 0:
			break;
		default:
			return p;
		}

		lp = strnlen(buf, l);
		buf += lp;
		bend = buf;
		*buf++ = ' ';
		*buf = 0;
		l -= lp+1;
	}
	return p;
}

const char*
scanstr(char* t, int l, const char *s) {
	return _scanstr(mapu, t, l, s);
}

const char*
scanmimestr(char* t, int l, const char *s) {
	return _scanstr(mapm, t, l, s);
}

const char*
scanphrase(char* t, int l, char* comment, int cl, const char* s){
	return _scanphrase(mapu, t, l, comment, cl, s);
}

const char*
_scancomment(char* t, int l, const char* s, const char* m){
	int c;
	int level;

	level=1;
	for(;;){
		while(l-- && (c=m[*(uchar*)s]) == 0)
			*t++ = *s++;
		if (0 == l)
			break;
		
		if (c<0){
			level += c+2;
			s++;
			if (0 != level) {
				continue;
			}
			break;
		}

		if (Lesc == c) {
			if (*++s)
				*t++ = *s++;
			continue;
		}
		break;
	}

	*t = 0;
	return s;
}

const char*
_scancommentz(const char* s, const char* m){
	int c;
	int level;

	level=1;
	for(;;){
		while((c=m[*(uchar*)s]) == 0)
			s++;
		
		if (c<0){
			level += c+2;
			s++;
			if (0 != level) {
				continue;
			}
			break;
		}

		if (Lesc == c) {
			if (*s++)
				s++;
			continue;
		}
		break;
	}

	return s;
}

const char*
scancomment(char* t, int l, const char* s){
	while (isspace(*s)){
		s++;
	}

	while ('(' == *s){
		if (l){
			s = _scancomment(t, l, 1+s, mapc);
		} else {
			s = _scancommentz(1+s, mapc);
		}
		while (isspace(*s)){
			s++;
		}
	}
	//*t = 0;
	return s;
}

const char*
scansqbracket(char* t, int l, const char* s){
	*t = 0;
	while (isspace(*s)){
		s++;
	}

	if ('[' != *s){
		//*t = 0;
		return s;
	}

	s = _scancomment(t, l, 1+s, mapsq);
	while (isspace(*s)){
		s++;
	}

	return s;
}

const char*
scanstrws(char* t, int l, const char* s){
	s = scancomment(0, 0, s);
	s = scanstr(t, l, s);
	
	while (isspace(*s)){
		s++;
	}

	return s;
}

const char*
scanmimestrws(char* t, int l, const char* s){
	s = scancomment(0, 0, s);
	s = scanmimestr(t, l, s);
	
	while (isspace(*s)){
		s++;
	}

	return s;
}

