/*
**
** RCS Info:	$Date: 2005/08/04 01:35:07 $
**		$Id: dconv.c,v 1.4 2005/08/04 01:35:07 quanstro Exp quanstro $
**		$Source: /home/quanstro/src/elocate/print/RCS/dconv.c,v $
**
** Description:
**
**	convert integers.
**
** Limitations:
**
**	1. none
**
** History:
**
**	02-jul-2005 quanstro created.
**	22-jul-2005 quanstro added %X; added %, support; rewrite utoa nonrecursively.
*/

#include <printInternal.h>
#include <rune.h>

static int lutoa(unsigned long long u, char* p, unsigned int radix, const char* digit, int comma){
	int n;

	n=0;
	do {
		if (comma && n%4 == 3){
			n++;
			*p-- = ',';
		}
		*p-- = digit[u%radix];
		u /= radix;
		n++;
	} while (u);

	return n;
}

static int utoa(unsigned long u, char* p, unsigned int radix, const char* digit, int comma){
	int n;

	n=0;
	do {
		if (comma && n%4 == 3){
			n++;
			*p-- = ',';
		}
		*p-- = digit[u%radix];
		u /= radix;
		n++;
	} while (u);

	return n;
}

static void intconv(Format *format, unsigned int radix, int upper, const char *altform) {
	static const char * const table[] = {
		"0123456789abcdefghijklmnopqrstuvwxyz",
		"0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ",
	};
	char padchar;
	SIZE_T len, pre, zeroes, padding, width;
	long n, flags, sign;
	unsigned long u;
	long long ln;
	unsigned long long lu;
	char number[64+64/3], prefix[20];
	char* n_end = number + sizeof number;

	if (radix > 36){
		return;
	}

	flags = format->flags;
	if (flags & FMT_unsigned) {
		if (flags & FMT_vlong) {
			ln = (unsigned long long)va_arg(format->args, long long);
		} else if (flags & FMT_long) {
			n = (unsigned long)va_arg(format->args, long);
		} else if (flags & FMT_short) {
			n = (unsigned long)(unsigned short)va_arg(format->args, int);
		} else if (flags & FMT_byte) {
			n = (unsigned long)(unsigned char)va_arg(format->args, int);
		} else {
			n = (unsigned long)(unsigned int)va_arg(format->args, int);
		}
		sign=1;
	} else {
		if (flags & FMT_vlong) {
			ln = (long long)va_arg(format->args, long long);
		} else if (flags & FMT_long) {
			n = (long)va_arg(format->args, long);
		} else if (flags & FMT_short) {
			n = (long)(short)va_arg(format->args, int);
		} else if (flags & FMT_byte) {
			n = (char)va_arg(format->args, int);
		} else {
			n = (long)va_arg(format->args, int);
		}
		sign = flags & FMT_vlong ? ln>=0 : n>=0;
	}

	pre = 0;
	if (sign){
		u = n;
		lu = ln;
		if (flags & FMT_sign){
			prefix[pre++] = '+';
		} else if (flags & FMT_space){
			prefix[pre++] = ' ';
		}
	} else {
		prefix[pre++] = '-';
		u = -n;
		lu = -ln;
	}

	if (flags & FMT_altform) {
		while (*altform) {
			prefix[pre++] = *altform++;
		}
	}

	if (flags & FMT_vlong) {
		len = lutoa(lu, n_end, radix, table[upper], flags & FMT_comma);
	} else {
		len = utoa(u, n_end, radix, table[upper], flags & FMT_comma);
	}
	
	if ((flags & FMT_f2set) && format->f2 > len) {
		zeroes = format->f2 - len;
	} else {
		zeroes = 0;
	}

	width = pre + zeroes + len;
	if ((flags & FMT_f1set) && format->f1 > width) {
		padding = format->f1 - width;
	} else {
		padding = 0;
	}

	padchar = ' ';
	if (padding > 0 && flags & FMT_zeropad) {
		padchar = '0';
		if ((flags & FMT_leftside) == 0) {
			zeroes += padding;
			padding = 0;
		}
	}

	if (0 == (flags & FMT_leftside)) {
		fmtpad(format, padding, padchar);
	}
	fmtappend(format, prefix, pre);
	fmtpad(format, zeroes, '0');
	fmtappend(format, n_end +1 - len, len);
	if (flags & FMT_leftside) {
		fmtpad(format, padding, padchar);
	}
}

#define CONV(d, base, upper, altstring) \
	fmt_action d(Format*fmt, int c){\
		USED(c);\
		intconv(fmt, base, upper, altstring);\
		return FMT_verb;\
	}

CONV(bconv,  2, 0, "0b")
CONV(oconv,  8, 0, "0")
CONV(dconv, 10, 0, "")
CONV(xconv, 16, 0, "0x")
CONV(Xconv, 16, 1, "0X")
