/*
**
**  Classification:	UNCLASSIFIED
**
**  Copyright:		(c) Copyright 1995 IBM Corporation
**			This program is protected by the copyright law as an
**			unpublished work.
**
**  RCS Info:		$Date: 2005/07/19 03:45:44 $ $Revision: 1.4 $
**
**			$Id: runeconv.c,v 1.4 2005/07/19 03:45:44 quanstro Exp quanstro $
**
**  Filename:		$Source: /home/quanstro/src/elocate/print/RCS/runeconv.c,v $
**
**  Originator:		Erik Quanstrom
**			(703)205-6195
**			quanstro@dev.infomkt.ibm.com
**
**  Organization:	IBM Corporation / infoMarket
**			3190 Fairview Park Drive
**			Falls Church, VA  22042
**			Ph: (703)205-5600	FAX: (703)205-5704
**
**  Description:
**
**	This is the libprint public domain library, a replacement for the
**	assorted standard I/O printing functions
**
**  Limitations:
**
**	1.  None
**
**  Modification History:
**
**	23-feb-1998 quanstro Created.
**	01-jul-2005 quanstro updated to match Plan 9 development.
**
******************************************************************************/
/* Contributed by Erik Quanstrom <quanstro@sartre.minerva.bah.com> */

#include "printInternal.h"
#include <rune.h>

static SIZE_T runestrlen(const Rune *r) {
	const Rune *s = r;
	while (*r) r++;
	return r-s;
}

static Rune stupid[] = { 923, 0};

fmt_action Sconv (Format *format, int c) {
	USED(c);
	SIZE_T maxrunes, minwidth, extra;
	Rune *R;
	char s[UTFmax]; 
	int i;

	R = va_arg(format->args, Rune*);
/*	if (!R) R = L"Λ";  some compilers choke on this */
	if (!R) R = stupid;

	if (format->flags & FMT_altform) {
		/* maxrunes may overrun R,  but this way we don't have 
	   	   to assume that R is null terminated when f2 is set. */
		maxrunes = (format->flags & FMT_f2set) ? format->f2 : runestrlen(R);
	} else {
		/* do f2 like printf */
		maxrunes = runestrlen(R);
		if ((format->flags & FMT_f2set) && (format->f2 < maxrunes))
			maxrunes = format->f2;
	}

	minwidth = (format->flags & FMT_f1set) ? format->f1 : maxrunes;
	extra = (minwidth > maxrunes) ? minwidth - maxrunes : 0;

	if (format->flags & FMT_leftside) {
		while (maxrunes--) {
			i = runetochar(s, R++);
			fmtappend(format, s, i);
		}
		fmtpad(format, extra, ' ');
	} else {
		fmtpad(format, extra, ' ');
		while (maxrunes--) {
			i = runetochar(s, R++);
			fmtappend(format, s, i);
		}
	}
	
	return FMT_verb;
}

fmt_action Cconv(Format *format, int c) {
	USED(c);

	fmtputr(format, (Rune)va_arg(format->args, int));
	return FMT_verb;
}
