/*
**
** RCS Info:	$Date: 2005/07/19 03:45:44 $
**		$Id: sconv.c,v 1.2 2005/07/19 03:45:44 quanstro Exp quanstro $
**		$Source: /home/quanstro/src/elocate/print/RCS/sconv.c,v $
**
** Description:
**
**	new /utf-aware!/ implementation of the %s conversion.
**	width and precision (format->{f1,f2} specified as %W.Ps where
**	W and P are [0-9]*) specify counts of utf-8 characters.
**	FMT_altform (%#s) will count the old way -- in bytes.
**
** Limitations:
**
**	1. calling %#.Ns with N > strlen(s) will core.
**
** History:
**
**	01-jul-2005 quanstro created.
*/

#include <printInternal.h>
#include <rune.h>

#include <string.h>

static inline SIZE_T min(SIZE_T a, SIZE_T b){
	if (a>b){
		return b;
	}
	return a;
}

static fmt_action sconv_bytes(Format *format, const char* s) {
	SIZE_T maxbytes, minwidth, extra;

	if (format->flags & FMT_f2set){
		maxbytes = format->f2;
		if (format->flags & FMT_f1set){
			maxbytes=min(strlen(s), maxbytes);
		}
	} else {
		maxbytes = strlen(s);
	}

	minwidth = (format->flags & FMT_f1set) ? format->f1 : maxbytes;
	extra = (minwidth > maxbytes) ? minwidth - maxbytes : 0;

	if (format->flags & FMT_leftside) {
		fmtappend(format, s, maxbytes);
		fmtpad(format, extra, ' ');
	} else {
		fmtpad(format, extra, ' ');
		fmtappend(format, s, maxbytes);
	}
		
	return FMT_verb;
}

static int fmtcatn(Format* f, const char*s, SIZE_T max_chars, SIZE_T max_bytes){
	SIZE_T bytes;
	unsigned char c;
	Rune r;

	if (~0UL != max_bytes){
		fmtappend(f,s,max_bytes);
		return 1;
	}
	if (~0UL == max_chars){
		fmtcat(f,s);		// f2 is trusted blindly.
		return 1;
	}

	for(bytes=0; max_chars-- && (c=s[bytes]); ){
		if (c<Runeself){
			bytes++;
		} else {
			bytes+= chartorune(&r,s+bytes);		// what about garbage on the end?
		}
	}
	return fmtappend(f,s,bytes);
}

static SIZE_T utfclen(const char*s, SIZE_T chars, SIZE_T* bytes){
	int c;
	long n;
	Rune rune;
	const char* s0;

	n = 0;
	s0 = s;
	for(;chars--;) {
		c = *(unsigned char*)s;
		if(c < Runeself) {
			if(c == 0)
				goto done;
			s++;
		} else
			s += chartorune(&rune, s);
		n++;
	}
done:
	*bytes = s - s0;
	return n;
}

fmt_action sconv(Format *format, int c) {
	SIZE_T maxchars, minwidth, extra;
	SIZE_T bytes;
	const char *s;
	USED(c);

	s = va_arg(format->args, char*);
	if (!s) {
		s = "(null)";
	}

	if (format->flags & FMT_altform) {
		return sconv_bytes(format, s);
	}

	maxchars = ~0;
	extra = 0;
	bytes=~0UL;
	if (format->flags & FMT_f2set && format->flags & FMT_f1set){
		maxchars = utfclen(s, format->f2, &bytes);
		minwidth = format->f1;
		extra = (minwidth > maxchars) ? minwidth - maxchars : 0;
	} else if (format->flags & FMT_f1set){
		maxchars = utfclen(s, ~0UL, &bytes);
		minwidth = format->f1;
		extra = (minwidth > maxchars) ? minwidth - maxchars : 0;
	} else if (format->flags & FMT_f2set){
		maxchars = format->f2;
	}

	if (format->flags & FMT_leftside) {
		fmtcatn(format, s, maxchars, bytes);
		fmtpad(format, extra, ' ');
	} else {
		fmtpad(format, extra, ' ');
		fmtcatn(format, s, maxchars, bytes);
	}
		
	return FMT_verb;
}
