/*
**
**  Classification:	UNCLASSIFIED
**
**  Copyright:		(c) Copyright 1999 Hoovers
**			This program is protected by the copyright law as an
**			unpublished work.
**
**  RCS Info:		$Date: 2005/07/23 19:17:50 $ $Revision: 1.3 $
**
**			$Id: base64_conv.c,v 1.3 2005/07/23 19:17:50 quanstro Exp quanstro $
**
**  Filename:		$Source: /home/quanstro/src/elocate/stdprint/RCS/base64_conv.c,v $
**
**  Originator:		david sward
**			703-438-3500x8124
**			sward@clark.net
**
**  Organization:	Hoovers
**			1761 Business Center Drive Suite 350
**			Reston, VA 20190
**			Ph: 703 438 3500	FAX: 703 438 3550
**
**  Description:
**
**	provide base64 encoding/decoding.
**
**  Limitations:
**
**	1.  none.
**
**  Modification History:
**
** 	29-jan-1999 sward    Created.
**	07-feb-2000 sward    split into {base64} and {unbase64}, and applied sconv
**			     conventions for #*.*
**	24-feb-2000 sward    fixed encoding bug (read_uninit_mem).
**	22-oct-2005 quanstro allow unpadded encodings to support utf-7 (bah!).
**
******************************************************************************/
#include <stdprintInternal.h>
#include <xalloc.h>
#include <memory.h>
#include <string.h>

static const char* encode_tab = \
"ABCDEFGHIJKLMNOPQRSTUVWXYZ" "abcdefghijklmnopqrstuvwxyz" "0123456789" "+/";

static char *_base64_encode(const void *ptr, SIZE_T len, SIZE_T* outlen, const char* et) {
	unsigned long i, j, left;
	uchar* cp;
	char* d;

	cp = (uchar *) ptr;

	/* straight hex encoding expands 1 -> 2, +100% */
	/* base64 encoding expands 3 -> 4,       +33%; much better */
	d = xalloc((len*4)/3 + 4);	/* leave room for possible extra '=' */
	memset(d, 0, (len*4)/3 + 4);

	left = 0;	/* shut up the optimizer */
	for (j = 0, i = 0; i < len; i += 3) {
		left = (len >= i+3) ? 3 : len-i;
		switch (left) {
		case 1:
			d[j++] = et[cp[i+0]>>2];
			d[j++] = et[(cp[i+0]&0x3)<<4];
			break;
		case 2:
			d[j++] = et[cp[i+0]>>2];
			d[j++] = et[((cp[i+0]&0x3)<<4) | (cp[i+1]>>4)];
			d[j++] = et[(cp[i+1]&0xf)<<2];
			break;
		case 3:
			d[j++] = et[cp[i+0]>>2];
			d[j++] = et[((cp[i+0]&0x3)<<4) | (cp[i+1]>>4)];
			d[j++] = et[((cp[i+1]&0xf)<<2) | (cp[i+2]>>6)];
			d[j++] = et[cp[i+2]&0x3f];
			break;
		default:
			break;
		}
	}
	for (; left < 3; left++) {
		d[j++] = '=';
	}
	*outlen = j;
	d[j] = 0;
	return d;
}

static const uchar decode_tab64[] = {
/*  0*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/*  8*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/* 16*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/* 24*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/* 32*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/* 40*/ 0x80, 0x80, 0x80,   62, 0x80, 0x80, 0x80,   63,
/* 48*/   52,   53,   54,   55,   56,   57,   58,   59,
/* 56*/   60,   61, 0x80, 0x80, 0x80,    0, 0x80, 0x80,
/* 64*/ 0x80,    0,    1,    2,    3,    4,    5,    6,
/* 72*/    7,    8,    9,   10,   11,   12,   13,   14,
/* 80*/   15,   16,   17,   18,   19,   20,   21,   22,
/* 88*/   23,   24,   25, 0x80, 0x80, 0x80, 0x80, 0x80,
/* 96*/ 0x80,   26,   27,   28,   29,   30,   31,   32,
/*104*/   33,   34,   35,   36,   37,   38,   39,   40,
/*112*/   41,   42,   43,   44,   45,   46,   47,   48,
/*120*/   49,   50,   51, 0x80, 0x80, 0x80, 0x80, 0x80,
/*128*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/*136*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/*144*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/*152*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/*160*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/*168*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/*176*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/*184*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/*192*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/*200*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/*208*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/*216*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/*224*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/*232*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/*240*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
/*248*/ 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80 };

static uchar* _gen_decode_tab(const uchar* enc_tab) {
	uchar* dec_tab;
	int i;

	dec_tab = xalloc(256);
	memset(dec_tab, 0x80, 256);
	for (i = 0; enc_tab[i] != 0; i++) {
		dec_tab[enc_tab[i]] = i;
	}
	dec_tab['='] = 0;
	return dec_tab;
}


static char* __base64_decode(char* d, const char *str, SIZE_T len, SIZE_T* outlen, const uchar* dt) {
	unsigned long i, j;

	for (i = 0; i < len; i++) {
		if (dt[(int)str[i]] == 0x80) {
			return 0;	/* denied! */
		}
	}

	for (j = 0, i = 0; i < len; i += 4) {
		/* formats: xxxx -> xxx, xxx= -> xx, xx== -> x */
		d[j++] = (dt[(int)str[i+0]]<<2) | (dt[(int)str[i+1]]>>4);
		d[j++] = (dt[(int)str[i+1]]<<4) | (dt[(int)str[i+2]]>>2);
		d[j++] = (dt[(int)str[i+2]]<<6) | (dt[(int)str[i+3]]);
	}
	for (i--; i>0 && str[i] == '='; i--, j--) {}
	if (i == 0 && str[0] == '='){
		return 0;
	}

	*outlen = j;
	d[j] = 0;
	return d;
}

static void *_base64_decode(const char *str, SIZE_T len, SIZE_T* outlen, const uchar* dt) {
	char pad[4];
	int r;
	char* d;
	char* p;
	SIZE_T j;

	r = len%4;
	d = xalloc(((r+len)/4)*3 + 1);
	p = __base64_decode(d, str, len-r, outlen, dt);
	if (!p){
		xfree(d);
		return 0;
	}
	if (!r){
		return d;
	}
	
	str += len-r;
	memset(pad,'=',4);
	memcpy(pad,str,r);
	p = __base64_decode(d + *outlen, pad, r, &j, dt);
	if (p) {
		*outlen += j;
	}

	return d;
}


fmt_action long_base64_conv(Format* f, const char* arg) {
	char* string;
	char* out;
	SIZE_T maxbytes, outlen;
	const char* table;
	USED(arg);

	string = va_arg(f->args, char*);
	if (f->flags & FMT_altform) {
		maxbytes = (f->flags & FMT_f2set) ? f->f2 : strlen(string);
	} else {
		maxbytes = strlen(string);
		if ((f->flags & FMT_f2set) && (f->f2 < maxbytes)){
	    		maxbytes = f->f2;
		}
	}

	if (0 == maxbytes) {
		return FMT_verb;
	}

	table = encode_tab;
	if (f->client_data) {
		table = f->client_data;
	}
	out = _base64_encode(string, maxbytes, &outlen, table);

	if (out) {
		fmtappend(f, out, outlen);
		xfree(out);
	}
	return FMT_verb;
}

fmt_action long_unbase64_conv(Format* f, const char* arg) {
	char* string;
	char* out;
	SIZE_T maxbytes, outlen;
	uchar* dec_tab;
	USED(arg);

	string = va_arg(f->args, char*);
	if (f->flags & FMT_altform) {
		maxbytes = (f->flags & FMT_f2set) ? f->f2 : strlen(string);
	} else {
		maxbytes = strlen(string);
		if ((f->flags & FMT_f2set) && ((SIZE_T)f->f2 < maxbytes)){
	    		maxbytes = f->f2;
		}
	}

	if (0 == maxbytes) {
		return FMT_verb;
	}

	if (f->client_data) {
		dec_tab = _gen_decode_tab((uchar*)f->client_data);
		out = _base64_decode(string, maxbytes, &outlen, dec_tab);
		xfree(dec_tab);
	} else {
		out = _base64_decode(string, maxbytes, &outlen, decode_tab64);
	}

	if (out) {
		fmtappend(f, out, outlen);
		xfree(out);
	}
	return FMT_verb;
}
