#include <base.h>
#include <print.h>
#include <rune.h>

static int utf_cquote_bytes(const char*s, int n, SIZE_T* fmt_chars){
	Rune r;

	if (n<0){
		n = 4;
	}
	if (fullrune(s,n)){
		*fmt_chars=1;
		return chartorune(&r,s);
	}
	*fmt_chars=sizeof("?xxx");	/* ?=\ just put \xxx in output */
	return 1;	
}

static void cquote_size(const char* s, SIZE_T utf_len, SIZE_T maxconv, SIZE_T* conv_len, SIZE_T* conv_bytes){
	SIZE_T utf;
	SIZE_T utf_conv;
	SIZE_T i;
	unsigned int c;
	SIZE_T cchars;

	for(cchars=0, utf_conv=0, utf=0, i=0; utf<utf_len && utf_conv+cchars<=maxconv; utf++){
		utf_conv+=cchars; cchars=0;
		c=(unsigned int)s[i];
		if (c<Runesync){
			switch(c){
			default:
				if (c<' '|| c==0x7f){
					cchars = 3;	/* we're not going to shorten \1 */
				} else {
					cchars = 1;
				}
				break;
			case '"':
			case '\0':
			case '\a':
			case '\b':
			case '\e':
			case '\f':
			case '\n':
			case '\r':
			case '\t':
			case '\v':
				cchars = 2;
				break;
			}
			i++;
		} else {
			i+=utf_cquote_bytes(s+i, maxconv-utf_conv, &cchars);
		}
	}

	*conv_len = utf_conv+cchars;
	//*conv_len = utf;
	*conv_bytes = i;
}

static void cquote_fmt(Format* f, const char* s, SIZE_T utf_len, SIZE_T max_bytes){
	SIZE_T i;
	SIZE_T l;
	SIZE_T b;
	unsigned int c;

	for(i=0, b=0; b<utf_len; ){
		c=(unsigned int)s[i];
		if (c<Runesync){
			switch(c){
			default:
				if (c<' '|| c==0x7f){
					fmtprint(f, "\\%02o", (int)c);
					b+=3;
				} else {
					fmtputc(f, c);
					b++;
				}
				break;
			case '"':
				b+=2; fmtappend(f, "\\\"", 2); break;
			case '\0':
				b+=2; fmtappend(f, "\\0", 2); break;
			case '\a':
				b+=2; fmtappend(f, "\\a", 2); break;
			case '\b':
				b+=2; fmtappend(f, "\\b", 2); break;
			case '\e':
				b+=2; fmtappend(f, "\\e", 2); break;
			case '\f':
				b+=2; fmtappend(f, "\\f", 2); break;
			case '\n':
				b+=2; fmtappend(f, "\\n", 2); break;
			case '\r':
				b+=2; fmtappend(f, "\\r", 2); break;
			case '\t':
				b+=2; fmtappend(f, "\\t", 2); break;
			case '\v':
				b+=2; fmtappend(f, "\\v", 2); break;
			}
			i++;
		} else {
			SIZE_T dummy;

			l=utf_cquote_bytes(s+i, max_bytes-i, &dummy);
			if (4 == l){
				fmtprint(f, "\\%03uo", (unsigned int)s[i]);
				b+=4;
			} else {
				fmtappend(f, s+i, l);
				b++;
			}
			i+=l;
		}
	}
}

fmt_action long_cquote_conv(Format* f, const char* fmt){
	const char* s;
	SIZE_T len;
	SIZE_T fmt_len;
	SIZE_T extra;
	SIZE_T minwidth;
	SIZE_T maxconv;
	SIZE_T maxconv_b;
	USED(fmt);

	s = va_arg(f->args, const char*);
	if (!s){
		s = "(null)";
	}

	if (f->flags&FMT_f2set){
		if (f->flags & FMT_altform){
			len=f->f2;
		} else {
			len = utflen(s);
			if (len>f->f2){
				len=f->f2;
			}
		}
		maxconv=~0UL;
	} else {
		maxconv=~0UL;
		len=utflen(s);
	}

	cquote_size(s, len, maxconv, &fmt_len, &maxconv_b);
	if (~0UL == maxconv){
		maxconv=fmt_len;
	}

	if (f->flags&FMT_f1set){
		minwidth=f->f1;
	} else {
		minwidth=fmt_len;
	}

	if (minwidth>maxconv){
		extra = minwidth-maxconv;
	} else {
		extra=0;
	}

	if (f->flags&FMT_leftside){
		cquote_fmt(f, s, maxconv, maxconv_b);
		fmtpad(f, extra, ' ');
	} else {
		cquote_fmt(f, s, maxconv, maxconv_b);
		fmtpad(f, extra, ' ');
	}

	return FMT_verb;	
}

