/*
**
** Copyright:	Copyright © 2005 erik quanstrom
**
** RCS Info:	$Date: 2005/09/04 03:00:44 $
**		$Id: html_conv.c,v 1.4 2005/09/04 03:00:44 quanstro Exp quanstro $
**		$Source: /home/quanstro/src/elocate/stdprint/RCS/html_conv.c,v $
**
** Description:
**
**	convert to/from html character quoting rules.
**	e.g.:
**		"&<>		↔	&quot; &amp; &lt; &gt;
**		u+00a0		↔	&nbsp;
**		ascii bogons	↔	&#D; or &#xH;
**		unicode char	↔	\uXXXX;
**
** Limitations:
**
**	1. i can't find a good reference. i'm not sure if i'm missing some entities.
**	2. a bad entity aborts the translation.
**	3. input and output are assumed valid utf-8.
**	4. handling of &Dagger; is bogus.
**
** History:
**
**	25-aug-2005 quanstro created.
**	02-sep-2005 quanstro binary search ntab[].
*/

#include <stdprintInternal.h>
#include <rune.h>

#include <memory.h>
#include <ctype.h>

typedef struct {
	Rune a;
	Rune b;
	const char* fmt;
	SIZE_T len;
	boolean ent;
} Tab;

static Tab tab[] = {
	{	0x0000,	0x0008,	"&#%02d;",	5,	true,	},
	{	0x000b,	0x001f,	"&#%02d;",	5,	true,	},
	{	'"',	'"',	"&quot;",	6,	true,	},
	{	'&',	'&',	"&amp;",	5,	true,	},
	{	'<',	'<',	"&lt;",		4,	true,	},
	{	'>',	'>',	"&gt;",		4,	true,	},
	{	0x00a0,	0x00a0,	"&nbsp;",	6,	true,	},
	{	0x7f,	0xffff,	"\\u%04x",	6,	true,	},
	{	0xffff,	0,	"%C",		1,	true,	}
};

#define TAB_COUNT	(sizeof(tab) / sizeof(tab[0]))
#define TAB_LAST	(sizeof(tab) / sizeof(tab[0]) - 1)

// all the entries in tab[] are duplicated here
Tab ntab[] = {
//	{	0x2021,	0x2021,	"&Dagger;",	8,	false,	},
	{	0x00e1,	0x00e1,	"&aacute;",	8,	false,	},
	{	0x00e2,	0x00e2,	"&acirc;",	7,	false,	},
	{	0x00b4,	0x00b4,	"&acute;",	7,	false,	},
	{	0x00e6,	0x00e6,	"&aelig;",	7,	false,	},
	{	0x00e0,	0x00e0,	"&agrave;",	8,	false,	},
	{	'&',	'&',	"&amp;",	5,	true,	},
	{	0x00e5,	0x00e5,	"&aring;",	7,	false,	},
	{	0x00e3,	0x00e3,	"&atilde;",	8,	false,	},
	{	0x00e4,	0x00e4,	"&auml;",	6,	false,	},
	{	0x201e,	0x201e,	"&bdquo;",	7,	false,	},
	{	0x00a6,	0x00a6,	"&brvbar;",	8,	false,	},
	{	0x2022,	0x2022,	"&bull;",		6,	false,	},
	{	0x00e7,	0x00e7,	"&ccedil;",	8,	false,	},
	{	0x00b8,	0x00b8,	"&cedil;",	7,	false,	},
	{	0x00a2,	0x00a2,	"&cent;",	6,	false,	},
	{	0x2663,	0x2663,	"&clubs;",	7,	false,	},
	{	0x00a9,	0x00a9,	"&copy;",	6,	true,	},
	{	0x00a4,	0x00a4,	"&curren;",	8,	false,	},
	{	0x2020,	0x2020,	"&dagger;",	8,	false,	},
	{	0x00b0,	0x00b0,	"&deg;",		5,	false,	},
	{	0x2666,	0x2666,	"&diams;",	7,	false,	},
	{	0x00f7,	0x00f7,	"&divide;",	8,	false,	},
	{	0x00e9,	0x00e9,	"&eacute;",	8,	false,	},
	{	0x00ea,	0x00ea,	"&ecirc;",	7,	false,	},
	{	0x00e8,	0x00e8,	"&egrave;",	8,	false,	},
	{	0x2014,	0x2014,	"&emdash;",	8,	false,	},
	{	0x2013,	0x2013,	"&endash;",	8,	false,	},
	{	0x00f0,	0x00f0,	"&eth;",		5,	false,	},
	{	0x00eb,	0x00eb,	"&euml;",	6,	false,	},
	{	0x20ac,	0x20ac,	"&euro;",	6,	true,	},
	{	0x00bd,	0x00bd,	"&frac12;",	8,	false,	},
	{	0x00bc,	0x00bc,	"&frac14;",	8,	false,	},
	{	0x00be,	0x00be,	"&frac34;",	8,	false,	},
	{	0x2265,	0x2265,	"&ge;",		4,	false,	},
	{	'>',	'>',	"&gt;",		4,	true,	},
	{	0x2665,	0x2665,	"&hearts;",	8,	false,	},
	{	0x2026,	0x2026,	"&hellip;",	8,	false,	},
	{	0x00ed,	0x00ed,	"&iacute;",	8,	false,	},
	{	0x00ee,	0x00ee,	"&icirc;",	7,	false,	},
	{	0x00a1,	0x00a1,	"&iexcl;",	7,	false,	},
	{	0x00ec,	0x00ec,	"&igrave;",	8,	false,	},
	{	0x00bf,	0x00bf,	"&iquest;",	8,	false,	},
	{	0x00ef,	0x00ef,	"&iuml;",	6,	false,	},
	{	0x00ab,	0x00ab,	"&laquo;",	7,	false,	},
	{	0x201c,	0x201c,	"&ldquo;",	7,	false,	},
	{	0x2264,	0x2264,	"&le;",		4,	false,	},
	{	0x2018,	0x2018,	"&lsquo;",	7,	false,	},
	{	'<',	'<',	"&lt;",		4,	true,	},
	{	0x00af,	0x00af,	"&macr;",	6,	false,	},
	{	0x00b5,	0x00b5,	"&micro;",	7,	false,	},
	{	0x00b7,	0x00b7,	"&middot;",	8,	false,	},
	{	0x00a0,	0x00a0,	"&nbsp;",	6,	true,	},
	{	0x2260,	0x2260,	"&ne;",		4,	false,	},
	{	0x00ac,	0x00ac,	"&not;",		5,	false,	},
	{	0x00f1,	0x00f1,	"&ntilde;",	8,	false,	},
	{	0x00f3,	0x00f3,	"&oacute;",	8,	false,	},
	{	0x00f4,	0x00f4,	"&ocirc;",	7,	false,	},
	{	0x00f2,	0x00f2,	"&ograve;",	8,	false,	},
	{	0x00aa,	0x00aa,	"&ordf;",	6,	false,	},
	{	0x00ba,	0x00ba,	"&ordm;",	6,	false,	},
	{	0x00f8,	0x00f8,	"&oslash;",	8,	false,	},
	{	0x00f5,	0x00f5,	"&otilde;",	8,	false,	},
	{	0x00f6,	0x00f6,	"&ouml;",	6,	false,	},
	{	0x00b6,	0x00b6,	"&para;",	6,	false,	},
	{	0x2030,	0x2030,	"&permil;",	8,	false,	},
	{	0x00b1,	0x00b1,	"&plusmn;",	8,	false,	},
	{	0x00a3,	0x00a3,	"&pound;",	7,	false,	},
	{	'"',	'"',	"&quot;",	6,	true,	},
	{	0x00bb,	0x00bb,	"&raquo;",	7,	false,	},
	{	0x201d,	0x201d,	"&rdquo;",	7,	false,	},
	{	0x00ae,	0x00ae,	"&reg;",		5,	true,	},
	{	0x2019,	0x2019,	"&rsquo;",	7,	false,	},
	{	0x00a7,	0x00a7,	"&sect;",	6,	false,	},
	{	0x2660,	0x2660,	"&spades;",	8,	false,	},
	{	0x00b9,	0x00b9,	"&sup1;",	6,	false,	},
	{	0x00b2,	0x00b2,	"&sup2;",	6,	false,	},
	{	0x00b3,	0x00b3,	"&sup3;",	6,	false,	},
	{	0x00df,	0x00df,	"&szlig;",	7,	false,	},
	{	0x00fe,	0x00fe,	"&thorn;",	7,	false,	},
	{	0x00d7,	0x00d7,	"&times;",	7,	false,	},
	{	0x2122,	0x2122,	"&trade;",	7,	true,	},
	{	0x00fa,	0x00fa,	"&uacute;",	8,	false,	},
	{	0x00fb,	0x00fb,	"&ucirc;",	7,	false,	},
	{	0x00f9,	0x00f9,	"&ugrave;",	8,	false,	},
	{	0x00a8,	0x00a8,	"&uml;",	5,	false,	},
	{	0x00fc,	0x00fc,	"&uuml;",	6,	false,	},
	{	0x00fd,	0x00fd,	"&yacute;",	8,	false,	},
	{	0x00a5,	0x00a5,	"&yen;",		5,	false,	},
	{	0x00ff,	0x00ff,	"&yuml;",	6,	false,	},
};

#define NTAB_COUNT	(sizeof(ntab) / sizeof(ntab[0]))

static Tab* bsearch_tab(Tab* t, int n, long k){
	Tab* e;
	int m;

	while(n > 1) {
		m = n/2;
		e = t + m;
		if(k >= e[0].a) {
			t = e;
			n = n-m;
		} else {
			n = m;
		}
	}
	if(n && k <= t[0].b){
		return t;
	}
	return 0;
}

SIZE_T tab_len(Rune r, const char** fmt){
	Tab* t;

	t = bsearch_tab(tab, TAB_COUNT, r);
	if (!t){
		t = tab + TAB_LAST;
	}
	*fmt = t->fmt;
	return t->len;
}

static void quotable_f2(const char* s, SIZE_T len, SIZE_T f2, SIZE_T* fmt_chars){
	SIZE_T i;
	SIZE_T c;
	SIZE_T b;
	SIZE_T cb;
	SIZE_T ci;
	Rune rune;
	unsigned char uc;
	const char* fmt;

	for(i=0,b=0,c=0; c<len; c++){
		uc = (unsigned char)s[i];
		if (Runeself <= uc){
			ci = chartorune(&rune, s+i);
			cb = tab_len(rune, &fmt);
		} else {
			ci = 1; rune = uc;
			cb = tab_len(uc, &fmt);
		}
		if(b+cb>f2){
			break;
		}
		i+=ci;
		b+=cb;
	}
	*fmt_chars=b;
}

static void do_print(Format* f, const char* s, const SIZE_T fmt_chars){
	int i;
	int ci;
	SIZE_T b;
	SIZE_T cb;
	unsigned char uc;
	Rune rune;
	const char* fmt;

	for(i=0, b=0; b < fmt_chars; ){
		uc = (unsigned char)s[i];
		if (Runeself <= uc){
			ci = chartorune(&rune, s+i);
			cb = tab_len(rune, &fmt);
		} else {
			ci = 1; rune = uc;
			cb = tab_len(uc, &fmt);
		}
		fmtprint(f, fmt, rune);
		i+=ci;
		b+=cb;
	}
}

fmt_action long_html_conv(Format *f, const char* fmt){
	const char* s;
	SIZE_T extra;
	SIZE_T maxbytes;
	SIZE_T maxconv;
	SIZE_T minwidth;
	USED(fmt);

	s = va_arg(f->args, const char*);
	if (!s) s = "(null)";

	if (f->flags & FMT_altform) {
		maxbytes = (f->flags & FMT_f2set) ? f->f2 : (SIZE_T)utflen(s);
	} else {
		maxbytes = utflen(s);
		if ((f->flags & FMT_f2set) && (f->f2 < maxbytes)){
			maxbytes = f->f2;
		}
	}

	quotable_f2(s, maxbytes, (f->flags & FMT_f2set) ? f->f2 : ~0UL, &maxconv);
	minwidth = (f->flags & FMT_f1set) ? (SIZE_T)f->f1 : maxbytes;

	if (minwidth>maxconv){
		extra = minwidth-maxconv;
	} else {
		extra=0;
	}

	if (f->flags & FMT_leftside) {
		do_print(f, s, maxconv);
		fmtpad(f, extra, ' ');
	} else {
		fmtpad(f, extra, ' ');
		do_print(f, s, maxconv);
	}

	return FMT_verb;
}

static boolean lower(SIZE_T n, char* t, const char* s){
	char c;
	boolean did_lower;

	did_lower = false;
	while(n--) {
		c = tolower(*s);
		if (c !=*s++){
			did_lower = true;
		}
		*t++ = c;
	}

	return did_lower;
}

static long untab_cmp(const char* b, long lb, const char* k, long lk){
	long n;
	long ln;

	ln=lb;
	if (ln>lk){
		ln=lk;
	}

	n = memcmp(b, k, ln);
	if (n || ln == lk){
		return n;
	}

	return lb>lk ? -1 : 1;
}

static Tab* bsearch_untab(Tab* t, int n, const char* s, SIZE_T l){
	Tab* e;
	int m;
	long k;

	while(n > 1) {
		m = n/2;
		e = t + m;
		k = untab_cmp(s, l, e[0].fmt+1, e[0].len-2);
		if (k == 0){
			return e;
		} else if (k > 0){
			t = e;
			n = n-m;
		} else {
			n = m;
		}
	}
	k = untab_cmp(s, l, t[0].fmt+1, t[0].len-2);
	if (0 == k){
		return t;
	}
	return 0;
}


/*	bogus ad-hoc algorithm. firefix allows &LT and &lt but &THORN works
	and &Thorn doesn't. &thorn is lower case.

	instead of having two entries &Aacute; and &aacute;, if we get an 
	entity in upper case, we'll toupperrune() the rune in ntab.

	this means that &THORN; and &Thorn; → Þ, while &thorn → þ.
	also, this necesitates the wierd case for &Dagger;

	possible solution: use tab.b to hold the "upper case" rune.
	which would be toupperrune() for all entities except for &dagger;.

 */
static Rune find_untab(const char* base, SIZE_T len){
	char low[25];
	boolean have_lower;
	const Tab* T;

	have_lower = false;
	if (lower(len, low, base)){
		have_lower=true;

		/* very odd special case */
		if (6 == len && 0 == memcmp(low, "dagger", 6)){
			return 0x2021;
		}
	}

	T = bsearch_untab(ntab, NTAB_COUNT, low, len);
	if (T) {
		return have_lower ? toupperrune(T[0].a) : T[0].a;
	}
	return 0;
}

static boolean unquotable_f2(boolean doit, Format* f, const char* s, SIZE_T len, SIZE_T f2, SIZE_T* fmt_chars){
	SIZE_T i;
	SIZE_T c;
	SIZE_T b;
	SIZE_T l;
	SIZE_T i0;
	int base;
	Rune rune;
	unsigned char uc;
	const char* rem;

	for(c=0, i=0, b=0; c<len; b++){
		uc = (unsigned char)s[i];
		switch(uc){
		case '\\':
			if(c+2 >= len || tolower(s[i+1]) != 'u'){
				goto the_default;
			}
			c += 2; i += 2;
			if (!strtouln_r(&l, s+i, &rem, len-c, 16) || l > 0xffff || 4 != rem-(s+i)){
				return false;
			}
			c += 4; i += 4;
			if (c<len && ';' == s[i]){
				i++;
				c++;
			}
			if (doit) fmtputr(f, (Rune)l);
			break;
		case '&':
			if (c+1 == len){
				return false;
			}
			c++; i++;
			for(i0=i; c+1<len && ';' != s[i] && !isspace(!s[i]); ){
				i++;
				c++;
			}

			if ('#' == s[i0]){			// this is in the wrong place, but still works.
				i0++;
			}
			if (0==i-i0){
				return false;
			}

			if ('x' == s[i0] || (s[i0] >= '0' && s[i0] <= '9')){
				if ('x' == s[i0]){
					i++; c++;
					base=16;
				} else {
					base=10;
				}
				if (!strtouln_r(&l, s+i0, &rem, len-i0, base) || l >= 0xff || 0 != rem-(s+i)){
					return false;
				}
				if (doit) fmtputr(f, (Rune)l);
			} else {
				// must modify b for &nbsp; ? 
				l = find_untab(s+i0, i-i0);
				if (!l){
					if (doit) {
						fmtputc(f, '&');
						fmtappend(f, s+i0, i-i0);
					}
					goto the_default;
					//return false;		// unknown entity
				}
				if (doit) fmtputr(f, (Rune)l);
			}	
			if (c<len && ';' == s[i]){
				i++;
				c++;
			}
			break;
		default:
		the_default:
			if (Runeself <= uc){
				i += chartorune(&rune, s+i);
				c++;
				if (doit) fmtputr(f, rune);
			} else {
				if (doit) fmtputc(f, s[i]);
				i++;
				c++;
			}
			break;
		}
		if(b==f2){
			break;
		}
	}
	*fmt_chars=b;
	return true;
}

fmt_action long_unhtml_conv(Format *f, const char* fmt){
	const char* s;
	SIZE_T extra;
	SIZE_T maxbytes;
	SIZE_T maxconv;
	SIZE_T minwidth;
	USED(fmt);

	s = va_arg(f->args, const char*);
	if (!s) s = "(null)";

	if (f->flags & FMT_altform) {
		maxbytes = (f->flags & FMT_f2set) ? f->f2 : (SIZE_T)utflen(s);
	} else {
		maxbytes = utflen(s);
		if ((f->flags & FMT_f2set) && (f->f2 < maxbytes)){
			maxbytes = f->f2;
		}
	}

	if (!unquotable_f2(false, f, s, maxbytes, (f->flags & FMT_f2set) ? f->f2 : ~0UL, &maxconv)){
		s="(error)";
		maxconv=7;
	}

	minwidth = (f->flags & FMT_f1set) ? (SIZE_T)f->f1 : maxconv;

	if (minwidth>maxconv){
		extra = minwidth-maxconv;
	} else {
		extra=0;
	}

	if (f->flags & FMT_leftside) {
		unquotable_f2(true, f, s, maxbytes, maxconv, &maxconv);
		fmtpad(f, extra, ' ');
	} else {
		fmtpad(f, extra, ' ');
		unquotable_f2(true, f, s, maxbytes, maxconv, &maxconv);
	}

	return FMT_verb;
}
