/*  
** RCS Info:	$Date: 2005/09/04 03:00:44 $
**		$Id: rcquote_conv.c,v 1.6 2005/09/04 03:00:44 quanstro Exp quanstro $
**		$Source: /home/quanstro/src/elocate/stdprint/RCS/rcquote_conv.c,v $
**
** history:
**
**	01-jul-2005	quanstro	made unnecessiarly utf-compatable
**	03-sep-2005	quanstro	use isspacerune() not private table; fix length bugs.
**
*/
#include <stdprintInternal.h>
#include <string.h>
#include <xalloc.h>
#include <rune.h>

typedef enum {
	bad_okay=0,
	bad_9=1,
	bad_rc=2,
	bad_spaces=3,
	bad_stuff=4
} bad;

static const char tab[Runeself] = {
	[' ']  = bad_9,
	['\t'] = bad_9,
	['\n'] = bad_9,
	['#']  = bad_rc,
	['$']  = bad_rc,
	['&']  = bad_rc,
	['\''] = bad_stuff,
	['(']  = bad_rc,
	[')']  = bad_rc,
	['*']  = bad_rc,
	[';']  = bad_rc,
	['<']  = bad_rc,
	['=']  = bad_rc,
	['>']  = bad_rc,
	['?']  = bad_rc,
	['[']  = bad_rc,
	[']']  = bad_rc,
	['^']  = bad_rc,
	['{']  = bad_rc,
	['}']  = bad_rc
};

static bad quotable_rune(Rune r){
	if (isspacerune(r)){
		return bad_spaces;
	}
	return bad_okay;
}

static boolean quotable_f2(const char*s, SIZE_T len, SIZE_T f2, SIZE_T* fmt_chars){
	SIZE_T i;
	SIZE_T c;
	boolean q;
	SIZE_T b;
	bad r;
	Rune rune;
	unsigned char uc;

	q=false;
	for(i=0,b=0,c=0;c<len && b<f2;c++){
		uc = (unsigned char)s[i];
		if (Runeself <= uc){
			i+=chartorune(&rune, s+i);
			r=quotable_rune(rune);
		} else {
			r=tab[uc];
			i+=1;
		}
		switch(r){
		case bad_stuff:
			if(b+1>=f2){
				goto done;
			}
			b++;
		case bad_9:
		case bad_rc:
		case bad_spaces:
			if (q==false){
				if (b+2 >= f2){
					goto done;
				}
				q=true;
				b+=2;
			}
		default:
			b++;
		}
	}
done:
	*fmt_chars=b;
	return q;
}

static void do_stuff(Format* f, SIZE_T chars, const char* s, Rune q){
	SIZE_T b;
	SIZE_T n;
	bad r;
	Rune rune;
	unsigned char uc;

	b=0;
	if (q){
		fmtputr(f,q);
		b+=2;
	}

	for(;b<chars;s+=n){
		uc = (unsigned char)*s;
		if (Runeself <= uc){
			n=chartorune(&rune, s);
			r=quotable_rune(rune);
		} else {
			r=tab[uc];
			n=1;
		}
		switch(r){
		case bad_stuff:
			b++;
			if (b>=chars){
				goto done;		/* can't leave unstuffed garbage */
			}
			fmtappend(f,s,n);
		default:
			fmtappend(f,s,n);
			b++;
		}
	}
done:
	if(q){
		fmtputr(f,q);
	}
}

fmt_action long_rcquote_conv(Format *f, const char* fmt){
	const char* s;
	SIZE_T extra;
	SIZE_T maxbytes;
	SIZE_T maxconv;
	SIZE_T minwidth;
	boolean quote;
	USED(fmt);

	s = va_arg(f->args, const char*);
	if (!s) s = "(null)";

	if (f->flags & FMT_altform) {
		maxbytes = (f->flags & FMT_f2set) ? f->f2 : (SIZE_T)utflen(s);
	} else {
		maxbytes = utflen(s);
		if ((f->flags & FMT_f2set) && (f->f2 < maxbytes)){
			maxbytes = f->f2;
		}
	}

	quote = quotable_f2(s, maxbytes,(f->flags & FMT_f2set) ? f->f2 : ~0UL, &maxconv);
	minwidth = (f->flags & FMT_f1set) ? (SIZE_T)f->f1 : maxbytes;

	if (minwidth>maxconv){
		extra = minwidth-maxconv;
	} else {
		extra=0;
	}

	if (f->flags & FMT_leftside) {
		do_stuff(f,maxconv, s, quote ? L'\'' : 0);
		fmtpad(f, extra, ' ');
	} else {
		fmtpad(f, extra, ' ');
		do_stuff(f,maxconv, s, quote ? L'\'' : 0);
	}

	return FMT_verb;
}
