/*
**
** Copyright:	Copyright © 2005 erik quanstrom
**
** RCS Info:	$Date: 2005/08/17 01:53:30 $
**		$Id: cfgParser.y,v 1.4 2005/08/17 01:53:30 quanstro Exp quanstro $
**		$Source: /home/quanstro/src/elocate/cfg/RCS/cfgParser.y,v $
**
** Description:
**
**	inline separator definition. e.g. 
**		print("%{sep \n%{list}}\n", l);
**	print a list with newlines between elements and not commas.
**
**	this is especially useful when priting a union element when the extra
**	argument would require special-case code.
**
** Limitations:
**
**	1. should this be called %{./} ?
**
** History:
**
**	18-aug-2005 quanstro created.
*/

#include <base.h>
#include <print.h>
#include <stdprint.h>

#include <xalloc.h>
#include <string.h>
#include <lst.h>

static const char* lastpercent(const char* s){
	int level;
	const char* last;

	level = 0;
	last = 0;
	do {
		switch(*s){
		case '{':
			level++; break;
		case '}':
			level--; break;
		case '%':
			if (0 == level){
				last = s;
			}
			break;
		}
	} while (*++s);
	return last;
}

fmt_action long_sep_conv(Format* f, const char* fmt){
	const char * real_fmt;
	void* client_data;
	fmt_action rv;

	if (0 == fmt){
		return FMT_verb;
	}

	real_fmt = lastpercent(fmt);
	if (0 == real_fmt){
		return fmtengine(f, fmt);
	}
	client_data = f->client_data;
	f->client_data = xcopyn(fmt, real_fmt-fmt);
	rv = fmtengine(f, real_fmt);
	cxfree(f->client_data);
	f->client_data = client_data;

	return rv;
}
