/*
**
**  Classification:	UNCLASSIFIED
**
**  Copyright:		(c) Copyright 1995 IBM Corporation
**			This program is protected by the copyright law as an
**			unpublished work.
**
**  RCS Info:		$Date: 2005/07/23 19:30:59 $ $Revision: 1.3 $
**
**			$Id: url_conv.c,v 1.3 2005/07/23 19:30:59 quanstro Exp quanstro $
**
**  Filename:		$Source: /home/quanstro/src/elocate/stdprint/RCS/url_conv.c,v $
**
**  Originator:		David Sward
**			(703)205-6072
**			sward@dev.infomkt.ibm.com
**
**  Organization:	IBM Corporation / infoMarket
**			3190 Fairview Park Drive
**			Falls Church, VA  22042
**			Ph: (703)205-5600	FAX: (703)205-5706
**
**  Description:
**
**	%{url}	string → url-encoded string
**	%#{url}	url-encoded string → string
**
**  Limitations:
**
**	1.  None
**
**  Modification History:
**
** 	15-Jul-1996 sward Created.
**	02-sep-1998 sward Brought into stdprint.
**
******************************************************************************/
#include <stdprintInternal.h>
#include <string.h>

static int hex(const char ch) {
	if (ch >= '0' && ch <= '9') {
		return ch - '0';
	} else if (ch >= 'a' && ch <= 'f') {
		return ch - 'a' + 10;
	} else if (ch >= 'A' && ch <= 'F') {
		return ch - 'A' + 10;
	}
	return 0;
}

static void url_reverse_conv(Format *f, const char* s) {
	for (; *s; s++) {
		switch (*s) {
		case '+':
			fmtputc(f, ' ');
			break;
		case '%':
			if (s[1] != 0 && s[2] != 0) {
				fmtputc(f, (char)(hex(*(s+1))<<4) + hex(*(s+2)));
				s += 2;
				break;
			}
		default:
			fmtputc(f, *s);
			break;
		}
	}
}

static void url_fwd_conv(Format* f, const unsigned char* s){
	static const char *convchars = "!#$%^&()+=[]{}\\|;:'\"`~<>,/?";
	for (; *s; s++) {
		if (' ' == *s){
			fmtputc(f, '+');
		} else if (*s < ' ' || *s > 127 || strchr(convchars, *s)) {
			fmtprint(f, "%%%2.2x", *s);
		} else {
			fmtputc(f, *s);
		}
	}
}

fmt_action long_url_conv(Format *f, const char* arg) {
	const char *s;
	USED(arg);

	s = va_arg(f->args, char *);
	if (!s){
		return FMT_verb;
	}

	if (f->flags & FMT_altform) {
		url_reverse_conv(f, s);
	} else {
		url_fwd_conv(f, (const unsigned char*)s);
	}
	return FMT_verb;
}
